{$A+,G+,Q-,R-,S-}

USES Crt;

CONST
  twopi = 6.2831853072;
  epsilon = 1e-6;
  maxarraysize = 128;  {doit tre une puissance de 2 !!}
  halfmaxsize = maxarraysize div 2;
  maxfreqsize = halfmaxsize-1;

TYPE
  dataindextype = 1..maxarraysize;
  cmpxindextype = 1..halfmaxsize;
  freqindextype = 1..maxfreqsize;

  complex = record realpart, imagpart : real end;
  real_complex=(r,c);
  dataarraytype = record case real_complex of
                   r: (rp: array [dataindextype] of real);
                   c: (cp: array [cmpxindextype] of complex);
                  end;
  cstermtype = record cosineterm, sineterm : real end;
  fouriertype = record
                 dcterm   : real;
                 noiseterm: real;
                 freqterms: array [freqindextype] of cstermtype;
                end;
  dataarray_fourier=(dtype,ctype);
  mixedtype = record case dataarray_fourier of
               dtype: (dataslot: dataarraytype);
               ctype: (coefslot: fouriertype);
              end;

VAR
  data : dataarraytype;
  didx : dataindextype;
  fidx : freqindextype;
  coefficients : fouriertype;
  mixed : mixedtype;


procedure fftofreal (var mixed : mixedtype; realpoints : integer);
var
 index, minusindex : freqindextype;
 temp1, temp2, temp3, w : complex;
 baseangle : real;

 procedure cadd (a, b : complex; var c : complex);
 begin
  with c do begin
   realpart := a.realpart + b.realpart;
   imagpart := a.imagpart + b.imagpart;
  end;
 end;

 procedure csubtract (a, b : complex; var c : complex);
 begin
  with c do begin
   realpart := a.realpart - b.realpart;
   imagpart := a.imagpart - b.imagpart;
  end;
 end;

 procedure cmultiply (a, b : complex; var c : complex);
 begin
  with c do begin
   realpart := a.realpart*b.realpart - a.imagpart*b.imagpart;
   imagpart := a.realpart*b.imagpart + b.realpart*a.imagpart;
  end;
 end;

 procedure conjugate (a : complex; var b : complex);
 begin
  with b do begin
   realpart := a.realpart;
   imagpart := -a.imagpart;
  end;
 end;

 procedure forwardfft (var data : dataarraytype; complexpoints : integer);

  procedure docomplextransform;
  var partitionsize, halfsize, offset,
      lowindex, highindex : dataindextype;
      baseangle, angle : real;
      bits : integer;
      w, temp : complex;

  begin   { docomplextransform }
   partitionsize := complexpoints;
   with data do
    repeat
     halfsize := partitionsize div 2;
     baseangle := twopi/partitionsize;
     for offset := 1 to halfsize do begin
      angle := baseangle * (offset-1);
      w.realpart := cos(angle);
      w.imagpart := -sin(angle);
      lowindex := offset;
       repeat
        highindex := lowindex + halfsize;
        csubtract (cp[lowindex], cp[highindex], temp);
        cadd (cp[lowindex], cp[highindex], cp[lowindex]);
        cmultiply (temp, w, cp[highindex]);
        lowindex := lowindex + partitionsize
       until lowindex >= complexpoints;
     end;
     partitionsize := partitionsize div 2;
    until partitionsize = 1;
  end;

  procedure shufflecoefficients;
  var lowindex, highindex : dataindextype;
      bits : integer;

   function log2 (index : integer) : integer;
   begin   { log2 }
    if index = 1 then log2 := 0 else log2 := log2(index div 2)+1;
   end;

   function bitreversal (index, bits : integer) : integer;
    function reverse (bits, stib, bitsleft : integer) : integer;
    begin   { reverse }
     if bitsleft = 0 then reverse := stib
     else
      if odd (bits) then reverse := reverse (bits div 2, stib*2+1, bitsleft-1)
                    else reverse := reverse (bits div 2, stib*2, bitsleft-1);
    end;
   begin   { bitreversal }
    bitreversal := reverse(index-1, 0, bits) +1;
   end;

   procedure swap (var a, b : complex);
   var temp : complex;
   begin   { swap }
    temp := a; a := b; b := temp;
   end;

   begin   { shufflecoefficients }
    bits := log2 (complexpoints);
    with data do
     for lowindex := 1 to complexpoints do begin
      highindex := bitreversal(lowindex, bits);
      if highindex > lowindex then swap (cp[lowindex], cp[highindex]);
     end;
   end;

  procedure dividebyn;
  var index : dataindextype;
  begin   { dividebyn }
   with data do
    for index := 1 to complexpoints do with cp[index] do begin
     realpart := realpart/complexpoints;
     imagpart := imagpart/complexpoints;
    end;
  end;

 begin   { forwardfft }
  write('trans... ');
  docomplextransform;
  write('coeffs... ');
  shufflecoefficients;
  write('div... ');
  dividebyn;
 end;

begin   { fftofreal }
 forwardfft (mixed.dataslot, realpoints div 2);
 write('termes... ');
 temp1 := mixed.dataslot.cp[1];
 with mixed.coefslot, temp1 do begin
  dcterm := (realpart + imagpart)/2;
  noiseterm := (realpart - imagpart)/2;
 end;
 baseangle := -twopi/realpoints;

 for index := 1 to realpoints div 4 do begin
  minusindex := (realpoints div 2) - index;
  with mixed.dataslot do begin
   conjugate (cp[minusindex+1], temp2);
   cadd (cp[index+1], temp2, temp1);
   csubtract (cp[index+1], temp2, temp2);
  end;
  w.realpart := sin(index*baseangle);
  w.imagpart := -cos(index*baseangle);
  cmultiply (w, temp2, temp2);
  cadd (temp1, temp2, temp3);
  csubtract (temp1, temp2, temp2);
  conjugate (temp2, temp2);
  with mixed.coefslot.freqterms[index], temp3 do begin
   cosineterm := realpart/2;
   sineterm := -imagpart/2;
  end;
  with mixed.coefslot.freqterms[minusindex], temp2 do begin
   cosineterm := realpart/2;
   sineterm := imagpart/2;
  end;
 end;
end; {fftofreal}


BEGIN
 ClrScr;
 Write('init... ');
 with mixed.dataslot do
 for didx := 1 to maxarraysize do
  rp[didx] :=   5
              + 10 * sin(twopi* 7*(didx-1)/maxarraysize)
              - 15 * cos(twopi*22*(didx-1)/maxarraysize);
 Write('calcul... ');
 fftofreal (mixed, maxarraysize);
 writeln;
 with mixed.coefslot do
 writeln ('DC = ', dcterm:5:2, ' ':5, 'Noise = ', noiseterm);
 writeln;
 writeln('harm  cos  sin');
 for fidx := 1 to maxfreqsize do begin
  write(fidx:2);
  with mixed.coefslot.freqterms[fidx] do begin
   if Abs(cosineterm)<epsilon then write(0:6) else write(cosineterm:6:2);
   if Abs(sineterm)<epsilon then write(0:6) else write(sineterm:6:2);
  end;
  if fidx mod 4 = 0 then writeln else write('  ');
 end;
END.
