{************************************************************************}
{                                CHDTM.PAS                               }
{                                                                        }
{  Loader pour les fichiers DTM (Digital Tracker - Falcon)               }
{                               (et DigiTracker  - PC    )               }
{                                                                        }
{  AUTEUR: Freddy Vtel.                                                }
{                                                                        }
{                                                                        }
{                               Dernires modifications le 26/07/98      }
{************************************************************************}

{$UNDEF Groupe}

{$IFNDEF Groupe}
UNIT ChDTM;
{$ELSE}
UNIT ChDTM_G;
{$ENDIF}

{$A+,G+,Q-,R-,S-}

INTERFACE

{$IFNDEF Groupe}
USES Fichiers,Util,Memoire,Chutil,ModmVar,CRT;
{$ELSE}
USES Fich_grp,Util,Memoire,Chutil_G,ModmVar,CRT;
{$ENDIF}

{==========================================================}

procedure Charge_DTM(chemin: String80; var erreur: WORD);

{==========================================================}

IMPLEMENTATION

TYPE Char4=array[1..4] of CHAR;

CONST
    Max_Samples_DTM=63;

    { Chunk Digital Tracker }

    Chunk_Debut  : Char4='D.T.';
    Chunk_Seq    : Char4='S.Q.';
    Chunk_PATT   : Char4='PATT'; { Aussi dans le format DigiTracker }
    Chunk_INST   : Char4='INST';
    Chunk_DAPT   : Char4='DAPT';
    Chunk_DAIT   : Char4='DAIT';
    Ver204       : Char4='2.04';

    { Chunk DigiTracker     }

    Chunk_Song        : Char4='SONG';
    Chunk_Name        : Char4='NAME'; { Nom du module        }
    Chunk_Info        : Char4='INFO';
    Chunk_Init        : Char4='INIT'; { BPM,Tempo de dpart  }
    Chunk_Sequence    : Char4='PSEQ'; { Squence             }


TYPE
     String5=STRING[5];
     Char22=array[1..22] of CHAR;

     T_Chunk=record
              Header:Char4;
              Taille:LONGINT;      { Taille (format motorola) }
             end;

     T_ChunkDebut=record
                   Type_musique:WORD;
                   Res1:WORD;
                   Res2:WORD;
                   Vitesse:WORD;
                   Tempo:WORD;
                   Bizarre:LONGINT;
                   Nom:array[1..20] of CHAR;
                  end;

     T_Chunk_Seq=record
                  Nb_Pos:WORD;
                  Restart:WORD;
                  Res1:LONGINT;
                 end;

     T_Chunk_PatternInfo=record
                          Nb_Voies:WORD;
                          Nb_Patt:WORD;
                          Version:array[1..4] of CHAR;
                         end;

     T_Chunk_Pattern=record
                      Voies_sauv:LONGINT;
                      Num_Pattern:WORD;
                      Nb_Lignes:WORD;
                     end;

     T_Inf_Instr=record
                  Res1 : LONGINT;
                  Len  : LONGINT;
                  Fine : BYTE;
                  Volume:BYTE;
                  Rep_Start : LONGINT;
                  Rep_Len   : LONGINT;
                  Nom : Char22;
                  TypeInstr:BYTE;
                  Precision:BYTE;
                  Midi1,Midi2:WORD;
                  Freq1,Freq2:WORD;
                 end;

     TInstrInfos=record
                  Nom:Char22;
                  Smp_len : LONGINT;
                  Smp_rep : LONGINT;
                  Smp_rep_Fin : LONGINT;
                  Fine:BYTE;
                  Volume:BYTE;
                  Precision:BYTE;    { 8 ou 16 Bit }
                 end;

     TNoteDTM=record case BYTE of
                 1: (l          : LONGINT);
                 2: (w1,W2      : WORD);
                 3: (b1,b2,b3,b4: BYTE)
              end;

     TInstrDTM=array[1..Max_Samples_DTM] of TInstrInfos;
     InstrDTM=^TInstrDTM;
     TPatternDTM=array[1..2048] of TnoteDTM;  { 32*64 }

VAR Chunk:T_Chunk;
    Taille_Chunk_Actuel:LONGINT;
    Nb_Voies_DTM:BYTE;
    SamplesDTM:^TInstrDTM;
    Version:BYTE;
    Nb_InstrDTM:WORD;

{---------------------------------------------------------}

procedure C_DAPT;

VAR
  PatternDTM:^TPatternDTM;
  Pattern:Ptr_Patt;

  N_Pattern:WORD;
  Nb_Lignes:WORD;
  Taille_Patt:WORD;

  Indice:WORD;
  i:INTEGER;

begin
 New(PatternDTM);
 f_module.DeplacePointeur(4);
 f_module.Lit(N_Pattern,2);
 f_module.Lit(Nb_Lignes,2);

 N_Pattern:=SWAP(N_Pattern)+1;
 Nb_Lignes:=SWAP(Nb_Lignes);
 if Nb_Lignes>64 then begin   { Plus de 64 lignes => Erreur !!! }
    Erreur_de_chargement:=Err_Format;
    Dispose(PatternDTM);
    exit;
 end;
 if Nb_Lignes>64 then Nb_Lignes:=64;
 Taille_Patt:=Nb_Lignes*Nb_Voies_DTM*sizeof(T_Note);

 if not ReserveUMBDOS(Segment_actuel,(Taille_Patt div 16)+1) then begin
   Dispose(PatternDTM);
   Erreur_de_chargement:=Err_Memoire;
   exit;
 end;

 Pointeur_Actuel:=Ptr(Segment_Actuel,0);
 Pattern:=Pointeur_actuel;
 MUS.T^.Pattern[N_Pattern]:=Segment_actuel;
 MUS.T^.Taille_Pattern[N_Pattern]:=Nb_Lignes;
 f_module.Lit(PatternDTM^,Taille_Patt);

 C_Note.Volume:=255;
 for indice:=1 to Nb_Voies_DTM*Nb_Lignes do with PatternDTM^[indice] do begin
   if Version=1 then begin  { Version 2.04 }
      C_Note.Note:=0;
      if b1>0 then
        if (b1 shr 4)>1 then C_Note.Note:=(b1 shr 4)*12+(b1 and $0F)+1;
      C_Note.Instrument:=((b3 shr 4)+(b2 and $03) shl 4);
   end
   else begin { Autre }
      C_Note.Note:=Periode2Note(b2+(WORD(b1 and $7) shl 8));
      C_Note.Instrument:=(b3 shr 4)+(b1 and 16);
   end;
   if C_Note.Instrument>Max_Samples then C_Note.Instrument:=0;

   C_Note.Commande:=(b3 and $0F)+1;
   C_Note.Parametre:=b4;

   if C_Note.Commande=$E+1 then begin {Commande tendue}
     C_Note.Commande:=$11+(b4 shr 4);
     C_Note. Parametre:=(b4 and $0F)
   end
   else
    if (C_Note.Commande=Arpeggio) and (C_Note.Parametre=0) then C_Note.Commande:=Rien;

   if (C_Note.Commande in [VolSlide,PortUp,PortDown]) and (C_Note.Parametre=0) then begin
      C_Note.Commande:=Rien;
      C_Note.Parametre:=0
   end;
   if (C_Note.Commande=SetTempo) then begin
     if (C_Note.Parametre>$20) then C_Note.Commande:=SetBPM;
     if (C_Note.Parametre=0  ) then C_Note.Commande:=Rien;
   end;
   if C_Note.Instrument <>0 then SamplFlags[C_Note.Instrument]:=TRUE;
   Pattern^[Indice]:=C_Note;
 end;

 Dispose(PatternDTM);
end; { Charger_PatternDTM }

{---------------------------------------------------------}

procedure C_DAIT;
begin
 f_module.Lit(NSmp,2);
 NSmp:=SWAP(NSmp);
 inc(NSmp);
 if (NSmp>0) and (SamplesDTM^[NSmp].Precision=8) and (NSmp<=Nb_InstrDTM) then
  with MUS.T^ do begin
    Move(SamplesDTM^[NSmp].Nom,Smp.Nom,22);
    Smp.len:=SamplesDTM^[NSmp].Smp_Len;
    if Smp.len>4 then begin
      inc(Nb_instruments);
      Smp.rep:=SamplesDTM^[NSmp].Smp_rep;
      Smp_rep_fin:=SamplesDTM^[NSmp].Smp_rep_fin;
      Smp.Volume:=SamplesDTM^[NSmp].Volume;
      if SamplesDTM^[NSmp].Fine>15 then Smp.C4Spd:=8363
        else Smp.C4Spd:=FineTuneConv[SamplesDTM^[NSmp].Fine];
      Smp.Info:=0;

      Charger_Sample;
    end;  { Smp_len>4 }
    Move(Smp,MUS.T^.PtrSamples[NSmp]^,sizeof(Smp)); { Copier le sample... }
  end;
end; {C_DAIT}

procedure C_INST;
var InstrDTM:T_Inf_Instr;

begin
 f_module.Lit(Nb_InstrDTM,2);
 Nb_InstrDTM:=SWAP(Nb_InstrDTM);

 { Rserve l'espace DOS pour la dfinition de samples }
 if not Allouer_Samples(Nb_InstrDTM) then exit;

 if Nb_InstrDTM>Max_Samples_DTM then Nb_InstrDTM:=Max_Samples_DTM;
 for NSmp:=1 to Nb_InstrDTM do begin
    f_module.Lit(InstrDTM,sizeof(InstrDTM));
    Nouveau_Sample;

    Smp.Len:=SwapLB(InstrDTM.Len);
    Smp.Rep:=SwapLB(InstrDTM.Rep_Start);
    Smp.Rep_Len:=SwapLB(InstrDTM.Rep_Len);
    {  Ajuster les valeurs du sample  }
    if Smp.Rep_Len<3 then begin
      Smp.rep:=0; Smp_rep_fin:=0 ; Smp.rep_Len:=0;
    end
    else begin
      if Smp.rep_len>Smp.len then Smp.rep_len:=Smp.len;
      if Smp.rep_len+Smp.rep>Smp.len then Smp.rep:=Smp.len-Smp.rep_len;
      Smp_rep_fin:=Smp.rep+Smp.rep_len;
    end;
    SamplesDTM^[NSmp].Nom:=InstrDTM.Nom;
    SamplesDTM^[NSmp].Smp_Len:=Smp.Len;  { Sauver les carracteristiques }
    SamplesDTM^[NSmp].Smp_rep:=Smp.rep;  { du sample                    }
    SamplesDTM^[NSmp].Smp_rep_Fin:=Smp_rep_Fin;
    SamplesDTM^[NSmp].Fine:=InstrDTM.Fine;
    SamplesDTM^[NSmp].Volume:=InstrDTM.Volume;
    SamplesDTM^[NSmp].Precision:=InstrDTM.Precision;
 end;
end; { C_INST }

{---------------------------------------------------------}

procedure Ch_DTM; {Charge un fichier Digital Tracker }
var i:INTEGER;
    Taille_Fich:LONGINT;
    Position_Fichier:LONGINT;
    Ch_Debut:T_ChunkDebut;
    Ch_Seq:T_Chunk_Seq;
    Ch_PattInfo:T_Chunk_PatternInfo;

begin
  New(SamplesDTM);

  Taille_Fich:=f_module.Taille;    { Lire la taille du fichier }
  Erreur_de_chargement:=Ok;

  MUS.H^.Type_Module:=T_DTM;     { type de module }
  MUS.H^.Octave_Min:=0;
  MUS.H^.Octave_Max:=7;
  MUS.H^.Tempo_depart:=6;

  Position_Fichier:=0;
  f_module.PlacePointeur(0);
  Convertir_signe:=FALSE;
  Init_Charge_Sample;

  repeat
   f_module.Lit(Chunk,sizeof(Chunk));
   Taille_Chunk_Actuel:=SwapLB(Chunk.Taille);
   if Chunk.Header=Chunk_debut then begin
     f_module.Lit(Ch_Debut,sizeof(Ch_Debut));
     Ch_Debut.Vitesse:=SWAP(Ch_Debut.Vitesse);
     Ch_Debut.Tempo:=SWAP(Ch_Debut.Tempo);
     Move(Ch_Debut.Nom,MUS.H^.Titre,20);   { Titre du module }
   end;

   if Chunk.Header=Chunk_Seq then begin
     f_module.Lit(Ch_Seq,sizeof(Ch_Seq));
     Ch_Seq.Nb_Pos:=SWAP(Ch_Seq.Nb_Pos);
     Ch_Seq.Restart:=SWAP(Ch_Seq.Restart);
     f_module.Lit(MUS.T^.Sequence,Ch_Seq.Nb_Pos);
     MUS.H^.Sequence_Len:=Ch_Seq.Nb_Pos;
     MUS.H^.Pos_de_restart:=Ch_Seq.Restart;
   end;

   if Chunk.Header=Chunk_PATT then begin
     f_module.Lit(Ch_PattInfo,sizeof(Ch_PattInfo));
     Ch_PattInfo.Nb_Voies:=SWAP(Ch_PattInfo.Nb_Voies);
     Ch_PattInfo.Nb_Patt:=SWAP(Ch_PattInfo.Nb_Patt);

     if Ch_PattInfo.Version=Ver204 then Version:=1 else Version:=0;
     Nb_Voies_DTM:=Ch_PattInfo.Nb_Voies;
     MUS.H^.Nb_de_voies:=Ch_PattInfo.Nb_Voies;
     MUS.H^.Nb_de_voies_Patt:=Ch_PattInfo.Nb_Voies;
   end;

   if Chunk.Header=Chunk_INST then C_INST; { Charge samples info }
   if Chunk.Header=Chunk_DAPT then C_DAPT; { Charge pattern      }
   if Chunk.Header=Chunk_DAIT then C_DAIT; { Charge sample       }

   Position_Fichier:=Position_Fichier+Taille_Chunk_Actuel+8;
   f_module.PlacePointeur(Position_Fichier);
   if Taille_Chunk_Actuel=0 then Erreur_de_Chargement:=Err_Format;
  until (Position_fichier>Taille_fich) or (Erreur_de_Chargement<>Ok);

  Fin_Charge_Sample;
  Dispose(SamplesDTM);

  Fixer_Panning(TRUE);
  FillChar(MUS.T^.Type_Voie,MUS.H^.Nb_de_voies,1); { Voies numriques... }
  Compter_Voies;
end; {Ch_DTM}

{---------------------------------------------------------}

procedure Ch_DT;
begin
 Erreur_de_Chargement:=Err_Non_Gere;
end; {Ch_DT}

{---------------------------------------------------------}

procedure Charge_DTM(chemin: String80; var erreur: WORD);
var
   Nom_fich: String12;

begin  {Charge_DTM}
 erreur:=Ok;
 f_module.Ouvre(lecture,chemin);

 if f_erreur=f_ok then begin
  MUS_Efface;
  Nom_fich:=NomFichierSeul(chemin);

  f_module.Lit(Chunk,sizeof(Chunk));            { Dtecter le type du module }
  if Chunk.Header=Chunk_debut then Ch_DTM       { Digital Tracker }
     else if Chunk.Header=Chunk_Song then Ch_DT { DigiTracker     }
             else Erreur:=Err_Format;

  Verifier_Module(Erreur_de_chargement);
  Erreur:=Erreur_de_chargement;
  f_module.Ferme
  end
 else Erreur:=Err_lecture;  {Impossible d'ouvrir le fichier}

 if Erreur in [Ok,Err_samples] then MUS.Chargee:=TRUE else MUS_Efface;
end; {Charge_DTM}

END.
