{************************************************************************}
{                               DMAUNIT.PAS                              }
{                                                                        }
{  Unit pour le DMA.                                                    }
{                                                                        }
{                                                                        }
{  AUTEURS: Freddy Vtel.                                               }
{                                                                        }
{                               Dbut d'criture le 16/10/95             }
{                               Dernires modifications le 26/07/98      }
{************************************************************************}
UNIT DMAUNIT;

{$A+,G+,Q-,R-,S-}

INTERFACE

procedure Programme_DMA(DMA:BYTE;Tampon:POINTER;Taille:WORD;Autoinit:BOOLEAN);

IMPLEMENTATION

function Lire_Taille_DMA(DMA:BYTE):WORD; assembler;
asm
  xor  AL,AL
  out  0Ch,AL           { Clear flipflop }
  mov  DL,DMA
  xor  DH,DH
  shl  DX,1
  inc  DX               { DX = channel * 2 + 1 = base counter }
  in   AL,DX            { AL = lower BYTE }
  mov  BL,AL
  in   AL,DX            { AL = higher BYTE }
  mov  BH,AL
  mov  AX,BX            { AX = high and low part together ;) - return that }
                        { bytes left to send = ax + 1 }
end;

function Lire_Position_DMA(DMA:BYTE):WORD; assembler;
asm
  xor  AL,AL
  out  0Ch,AL           { clear flipflop }
  mov  DL,DMA
  xor  DH,dh
  shl  DX,1
                        { dx = channel * 2 = Offset }
  in   AL,DX            { al = lower BYTE }
  mov  BL,AL
  in   AL,DX            { al = higher BYTE }
  mov  BH,AL
  mov  AX,BX            { AX = high and low part together ;) - return that }
                        { bytes left to send = ax + 1 }
end;

procedure Programme_DMA(DMA:BYTE;Tampon:POINTER;Taille:WORD;
                        Autoinit:BOOLEAN); assembler;

CONST PAGES : array[0..7] of WORD = ($0087,$0083,$0081,$0082,
                                     $008F,$008B,$0089,$008A);
asm
  cmp  DMA,4
  jae  @DMA16Bit

              { convert POINTER in realadress :
                dmapage*65536+dmaoffset = memsegment*16 + memoffset }
  mov AX,WORD Ptr(Tampon+2)
  rol AX,4
  mov CL,AL
  and AL,0F0h
  and CL,0Fh
  mov DI,AX         { cl:di - realadress ! }

  mov BH,DMA        { bh with dma_channel }
  mov BL,BH
  shl BL,1          { bl with dma_channel*2 }
  mov CH,048h       { ch = 010010xx }
                    {     \| |+- read
                           | +- autoinit flag
                           +- singlemode }
  mov AL,Autoinit
  shl AL,4
  or  CH,AL         { set the autoinit flag }
  add CH,DMA        { prepare for dma_channel }

  mov AL,4
  add AL,BH          { bh = dma_channel }
  out 0Ah,AL         { mask the channel }

  xor AL,AL
  out 0Ch,AL          { clear flipflop }

  mov AL,CH           { ch = DMAmode }
  out 0Bh,AL          { set dmatransfer mode }
  mov AX,DI           { di = DMAbuffer offset }
  push BX
  xor BH,BH
  mov DX,BX           { bx = 2*dma_channel }
  out DX,AL           { lower adress }
  mov AL,AH
  out DX,AL           { higer adress }
  mov DX,WORD ptr (Pages+bx)
  mov AL,CL           { cl = DMAbuffer page }
  out DX,AL           { data page }
  mov DX,BX           { bx = 2*dma_channel }
  mov AX,Taille
  dec AX
  inc DX              { dx = write base count }
  out DX,AL           { write lower length }
  mov AL,AH
  out DX,AL           { write higer length }
  pop BX
  mov AL,DMA
  out 0Ah,AL          { demask channel }
  jmp @Fin

@DMA16Bit:    { jump to here if 16bit DMA setup }

  { convert POINTER in realadress : }
  mov AX,WORD ptr(Tampon+2)      { ax = segment of buffer }
  rol AX,4                  { attention no offset ! }
  mov CL,al
  and AL,0F1h
  and CL,00Eh
  ror AX,1
  mov DI,AX           { cl:di - realadress ! }

  mov BH,DMA          { bh with dma_channel }
  sub BH,4            { channel 4-7 to number 0-3 }
  mov BL,BH
  shl BL,2
  add BL,0C0h         { bl = DMA adressport for current DMAchannel }
                      { bh = DMA16Bitchannel - 4  (0..3) }
  mov CH,048h         { ch = 010010xx }
                      {     \| |+-read
                             | +-autoinitflag
                             +-singlemode }
  mov AL,Autoinit
  shl AL,4
  or  CH,AL           { set the autoinitflag }
  add CH,BH           { ch = command for DMAchannel # }

  mov AL,4
  add AL,BH           { bh = 16bitDMA_channel - 4 }
  out 0D4h,AL         { mask the channel }
  xor AL,AL
  out 0D8h,AL         { clear flipflop }
  mov AL,CH           { ch = dmamode }
  out 0D6h,AL         { set dmatransfer mode }
  mov AX,DI           { lower part of adress }
  push BX
  xor BH,BH           { bx now = c0h/c4h/c8h/cch (0..3) }
  mov DX,BX           { bx = addressport for current channel }
  out DX,AL           { write lower adress }
  mov AL,AH
  out DX,AL           { write higer adress }
  mov DL,DMA
  xor DH,DH
  mov SI,DX           { si = dma16Bitchannel }
  shl SI,1            { si - position in pagetable }
  mov DX,WORD ptr (Pages+si)
  mov AL,CL           { dmabuffer page }
  out DX,AL           { data page }
  mov DX,BX           { old dx value ;) c0h/c4h/c8h/cch }
  mov AX,Taille
  dec AX
  add DX,2            { seperated by 2 -> dx now DMA base count }
  out DX,AL           { write lower length }
  mov AL,AH
  out DX,AL           { write higer length }
  pop BX
  mov AL,BH
  out 0D4h,AL         { demask channel }
@Fin:
end; {Programme_DMA}

END.
