{                                                                            }
{                   ---------------------------------------                  }
{                   Gravis Ultrasound driver - Version 2.11                  }
{                   ---------------------------------------                  }
{                                                                            }
{   Copyright <C> 1993,1994 Ingenuity Software                               }
{   Copyright <C> 1993,1994 Advanced Gravis                                  }
{                                                                            }
{   Modifi par Freddy Vtel                                                }
{                                                                            }

UNIT GUSUnit;

{$A+,G+,Q-,R-,S-}

INTERFACE

TYPE
  PFV          = procedure;
  INT_PROC     = procedure(Voice : INTEGER);
  WORD_PROC    = procedure(MIDI_Send : WORD);
  TWOWORD_PROC = procedure(MIDI_Status : WORD; MIDI_Data : WORD);

  Ultra_CFG = record
    Base_Port      : WORD;
    Dram_DMA_Chan  : WORD;
    ADC_DMA_Chan   : WORD;
    GF1_IRQ_Num    : WORD;
    MIDI_IRQ_Num   : WORD;
   end;

CONST
  Volume_GUS1:array[0..64] of WORD=($0E00,$B000,$B800,$BC00,$BE00,$C000,$C400,$C800,$CC00,
                $D000,$D200,$D400,$D600,$D800,$DA00,$DC00,$DE00,
                $E000,$E100,$E200,$E300,$E400,$E500,$E600,$E700,
                $E800,$E900,$EA00,$EB00,$EC00,$ED00,$EE00,$EF00,
                $F080,$F100,$F180,$F200,$F280,$F300,$F380,$F400,
                $F480,$F500,$F580,$F600,$F680,$F700,$F780,$F800,
                $F880,$F900,$F980,$FA00,$FA80,$FB00,$FB80,$FC00,
                $FC80,$FD00,$FD80,$FE00,$FE80,$FF00,$FF80,$FFF0);

  Init_Voices = 14;    { # of voices to init the card for }

  { Volume Control bit Values (or together TO form Volume control BYTE) }
  Loop_Volume           = 08;
  Bi_Directional_Volume = 16;
  Enable_Volume_Handler = 32;


  { Voice Control bit Values (or together TO form Voice control BYTE) }
  Voice_Data_16Bit        = 04;
  Loop_Voice              = 08;
  Bi_Directional_Voice    = 16;
  Enable_VoiceEnd_Handler = 32;


  { DMA control bit values (or together TO form DMA control BYTE) }
  DMA_Data_16Bit = 64;
  Convert_Data   = 128;


  { ULTRASND environment variable readable or not }
  Ultra_Installed : BOOLEAN = FALSE;


VAR
  Ultra_Config  : Ultra_CFG;
  UltraOk       : BOOLEAN;
  UltraError    : INTEGER;

{ Base routines }

{procedure DisplayMem;}
function  UltraClose : BOOLEAN;

procedure UltraDisableLineIn;
procedure UltraDisableMicIn;
procedure UltraDisableOutput;
procedure UltraEnableLineIn;
procedure UltraEnableMicIn;
procedure UltraEnableOutput;
function  UltraGetLineIn : BOOLEAN;
function  UltraGetMicIn : BOOLEAN;
function  UltraGetOutput : BOOLEAN;
procedure UltraMemInit;
procedure UltraMemClose;
function  UltraMaxAvail : LONGINT;
function  UltraMaxAlloc : LONGINT;
function  UltraMemAvail : LONGINT;
function  UltraMemAlloc(    Size     : LONGINT;
                        VAR Location : LONGINT) : BOOLEAN;
function  UltraMemFree(Size     : LONGINT;
                       Location : LONGINT) : BOOLEAN;
function  UltraOpen(VAR Config : Ultra_CFG;
                        Voices : INTEGER) : BOOLEAN;
function  UltraPeekData(PPort   : INTEGER;
                        Address : LONGINT) : BYTE;
function  UltraPing(PPort : WORD) : BOOLEAN;
procedure UltraPokeData(PPort   : INTEGER;
                        Address : LONGINT;
                        Data    : BYTE);
function  UltraProbe(PPort : WORD) : BOOLEAN;
function  UltraReset(Voices : INTEGER) : BOOLEAN;
procedure UltraSetBalance(Voice : INTEGER;
                          Data  : BYTE);
function  UltraSizeDRAM : INTEGER;

function  DetecteGUS(Affichage:BOOLEAN):BOOLEAN;           {*}

IMPLEMENTATION

USES
  DOS,CRT,Util,VARUnit;

CONST

  { Error codes }
  ULTRA_OK             = 1;
  BAD_NUM_OF_Voices    = 2;
  NO_MEMORY            = 3;
  CORRUPT_MEM          = 4;
  NO_Ultra             = 5;
  DMA_BUSY             = 6;
  BAD_DMA_ADDR         = 7;
  BAD_ENV_VAR          = 8;
  VOICE_NOT_VALID      = 9;
  NO_FREE_VOICES       = 10;
  VOICE_ALREADY_USED   = 11;
  FILE_NOT_FOUND       = 12;
  INSUFFICIENT_GUS_MEM = 13;
  NOT_3D_FILE          = 14;
  BAD_3D_HEADER        = 15;
  BAD_3D_DATA          = 16;
  BAD_3D_FREQ          = 17;
  NOT_BLOCK_DATA       = 18;
  NOT_INTERLEAVED_DATA = 19;

  JOYSTICK_TIMER  = $201;  { 201 }
  JOYSTICK_DATA   = $201;  { 201 }

  GF1_MIDI_CTRL   = $100;  { 3X0 }
  GF1_MIDI_DATA   = $101;  { 3X1 }

  GF1_PAGE        = $102;  { 3X2 }
  GF1_REG_SELECT  = $103;  { 3X3 }
  GF1_DATA_LOW    = $104;  { 3X4 }
  GF1_DATA_HI     = $105;  { 3X5 }
  GF1_IRQ_STAT    = $006;  { 2X6 }
  GF1_DRAM        = $107;  { 3X7 }

  GF1_MIX_CTRL    = $000;  { 2X0 }
  GF1_TIMER_CTRL  = $008;  { 2X8 }
  GF1_TIMER_DATA  = $009;  { 2X9 }
  GF1_IRQ_CTRL    = $00B;  { 2XB }

  { The GF1 Hardware clock. }
  Clock_Rate      = 9878400;

  { Mixer control bits. }
  ENABLE_LINE		= $01;
  ENABLE_DAC		= $02;
  ENABLE_MIC		= $04;

  { interrupt controller 1 }
  CNTRL_8259		= $21;
  OCR_8259		= $20;
  EOI			= $20;
  REARM3		= $2F3;
  REARM5		= $2F5;

  { interrupt controller 2 }
  CNTRL_M_8259	        = $21;
  CNTRL_M2_8259	        = $A1;
  OCR_2_8259		= $A0;

  { DMA }
  DMA_CONTROL		= $41;
  SET_DMA_ADDRESS	= $42;
  SET_DRAM_LOW		= $43;
  SET_DRAM_HIGH		= $44;

  { TIMER }
  TIMER_CONTROL		= $45;
  TIMER1		= $46;
  TIMER2		= $47;

  { SAMPLING }
  SET_SAMPLE_RATE	= $48;
  SAMPLE_CONTROL	= $49;

  { MISC }
  SET_JOYSTICK		= $4B;
  MASTER_RESET		= $4C;

  { Voice register mapping. }
  SET_CONTROL		= $00;
  SET_FREQUENCY		= $01;
  SET_START_HIGH	= $02;
  SET_START_LOW		= $03;
  SET_END_HIGH		= $04;
  SET_END_LOW		= $05;
  SET_VOLUME_RATE	= $06;
  SET_VOLUME_START	= $07;
  SET_VOLUME_END	= $08;
  SET_VOLUME		= $09;
  SET_ACC_HIGH		= $0a;
  SET_ACC_LOW		= $0b;
  SET_BALANCE		= $0c;
  SET_VOLUME_CONTROL 	= $0d;
  SET_VOICES		= $0e;

  GET_CONTROL		= $80;
  GET_FREQUENCY		= $81;
  GET_START_HIGH	= $82;
  GET_START_LOW		= $83;
  GET_END_HIGH		= $84;
  GET_END_LOW		= $85;
  GET_VOLUME_RATE	= $86;
  GET_VOLUME_START	= $87;
  GET_VOLUME_END	= $88;
  GET_VOLUME		= $89;
  GET_ACC_HIGH		= $8a;
  GET_ACC_LOW		= $8b;
  GET_BALANCE		= $8c;
  GET_VOLUME_CONTROL 	= $8d;
  GET_VOICES		= $8e;
  GET_IRQV		= $8f;

  { MIDI }
  MIDI_RESET	        = $03;
  MIDI_ENABLE_XMIT	= $20;
  MIDI_ENABLE_RCV	= $80;
  MIDI_RCV_FULL		= $01;
  MIDI_XMIT_EMPTY	= $02;
  MIDI_FRAME_ERR	= $10;
  MIDI_OVERRUN		= $20;
  MIDI_IRQ_PEND		= $80;

  { JOYSTICK }
  JOY_POSITION		= $0f;
  JOY_BUTTONS		= $f0;

  { GF1_IRQ_STATUS (PORT 3X6) }
  MIDI_TX_IRQ		= $01;
  MIDI_RX_IRQ		= $02;
  GF1_TIMER1_IRQ	= $04;
  GF1_TIMER2_IRQ	= $08;
  WAVETABLE_IRQ		= $20;
  ENVELOPE_IRQ		= $40;
  DMA_TC_IRQ		= $80;

  { GF1_MIX_CTRL (PORT 2X0) }
  ENABLE_LINE_IN	= $01;
  ENABLE_OUTPUT		= $02;
  ENABLE_MIC_IN		= $04;
  ENABLE_GF1_IRQ	= $08;
  GF122			= $10;
  ENABLE_MIDI_LOOP	= $20;
  SELECT_GF1_REG	= $40;

  { DMA control register }
  DMA_ENABLE		= $01;
  DMA_READ		= $02;
  DMA_WIDTH_16		= $04;
  DMA_RATE		= $18;
  DMA_IRQ_ENABLE	= $20;
  DMA_IRQ_PENDING	= $40;
  DMA_DATA_16		= $40;
  DMA_TWOS_COMP		= $80;

  { These are the xfer rate bits ... }
  DMA_R0		= $00;
  DMA_R1		= $08;
  DMA_R2		= $10;
  DMA_R3		= $18;

  { SAMPLE control register }
  ENABLE_ADC		= $01;
  ADC_MODE		= $02;
  ADC_DMA_WIDTH		= $04;
  ADC_IRQ_ENABLE	= $20;
  ADC_IRQ_PENDING	= $40;
  ADC_TWOS_COMP		= $80;

  { RESET control register }
  GF1_MASTER_RESET	= $01;
  GF1_OUTPUT_ENABLE	= $02;
  GF1_MASTER_IRQ	= $04;

  { Voice specific registers }
  VOICE_STOPPED		= $01;
  STOP_VOICE		= $02;
  VC_DATA_TYPE		= $04;
  VC_LOOP_ENABLE	= $08;
  VC_BI_LOOP		= $10;
  VC_WAVE_IRQ		= $20;
  VC_DIRECT		= $40;
  VC_IRQ_PENDING	= $80;

  { Volume specific registers }
  VL_RATE_MANTISSA	= $3f;
  VL_RATE_RANGE		= $C0;
  VL_START_MANT	        = $0F;
  VL_START_EXP		= $F0;
  VL_END_MANT		= $0F;
  VL_END_EXP		= $F0;

  { Volume control register }
  VOLUME_STOPPED	= $01;
  STOP_VOLUME		= $02;
  VC_ROLLOVER           = $04;
  VL_LOOP_ENABLE	= $08;
  VL_BI_LOOP		= $10;
  VL_WAVE_IRQ		= $20;
  VL_DIRECT		= $40;
  VL_IRQ_PENDING	= $80;

  { Voice IRQ }
  VOICE_VOLUME_IRQ	= $40;
  VOICE_WAVE_IRQ	= $80;

  { Memory / Misc }
  DMA_AUTO_INIT   = $01;
  DMA_16          = $40;
  DMA_8           = $00;
  DMA_CVT_2       = $80;
  DMA_NO_CVT      = $00;
  USE_ROLLOVER    = $01;

  ULTRA_PRESENT	  = $0001;
  DRAM_DMA_BUSY   = $0002;
  ADC_DMA_BUSY    = $0004;
  DRAM_DMA_NOWAIT = $0008;
  ADC_DMA_NOWAIT  = $0010;

  READ_DMA        = 1;
  WRITE_DMA       = 2;
  INDEF_READ      = 3;
  INDEF_WRITE     = 4;

  { DMA Controler #1 (8-bit controller) }
  DMA1_STAT	= $08;
  DMA1_WCMD	= $08;
  DMA1_WREQ	= $09;
  DMA1_SNGL	= $0A;
  DMA1_MODE	= $0B;
  DMA1_CLRFF	= $0C;
  DMA1_MCLR	= $0D;
  DMA1_CLRM	= $0E;
  DMA1_WRTALL	= $0F;
  { DMA Controler #2 (16-bit controller) }
  DMA2_STAT	= $D0;
  DMA2_WCMD	= $D0;
  DMA2_WREQ	= $D2;
  DMA2_SNGL	= $D4;
  DMA2_MODE	= $D6;
  DMA2_CLRFF	= $D8;
  DMA2_MCLR	= $DA;
  DMA2_CLRM	= $DC;
  DMA2_WRTALL	= $DE;

  DMA0_ADDR	= $00;
  DMA0_CNT	= $01;
  DMA1_ADDR	= $02;
  DMA1_CNT	= $03;
  DMA2_ADDR	= $04;
  DMA2_CNT	= $05;
  DMA3_ADDR	= $06;
  DMA3_CNT	= $07;
  DMA4_ADDR	= $C0;
  DMA4_CNT	= $C2;
  DMA5_ADDR	= $C4;
  DMA5_CNT	= $C6;
  DMA6_ADDR	= $C8;
  DMA6_CNT	= $CA;
  DMA7_ADDR	= $CC;
  DMA7_CNT	= $CE;
  DMA0_PAGE	= $87;
  DMA1_PAGE	= $83;
  DMA2_PAGE	= $81;
  DMA3_PAGE	= $82;
  DMA4_PAGE	= $8F;
  DMA5_PAGE	= $8B;
  DMA6_PAGE	= $89;
  DMA7_PAGE	= $8A;

  MAX_DMA       = 7;
  DMA_DECREMENT	= $20;

  { Bits for dma flags location }
  DMA_USED	= $0001;
  DMA_PENDING   = $0002;
  TWO_FLAG	= $0004;
  REV_FLAG	= $0008;
  CALIB_COUNT   = $0010;

  { IRQ constants }
  MAX_IRQ       = 16;
  IRQ_UNAVAIL   = $0000;
  IRQ_AVAIL     = $0001;
  IRQ_USED      = $0002;
  OCR1          = $20;
  IMR1          = $21;
  OCR2          = $A0;
  IMR2          = $A1;

  { for memory control scheme }
  MaxNumBanks = 4;
  BlockSizeK  = 256;
  OneK        = 1024;
  UMemInited  : BOOLEAN = FALSE;


TYPE
  ULTRA_DATA = record
     Flags               : WORD;
     Base_Port           : WORD;
     DRAM_DMA_Chan       : WORD;
     ADC_DMA_Chan        : WORD;
     GF1_IRQ_Num         : WORD;
     MIDI_IRQ_Num        : WORD;
     Mix_Image           : BYTE;
     Voices              : BYTE;
     Image_MIDI          : BYTE;
     Used_Voices         : LONGINT;
     Timer_Ctrl          : BYTE;
     Timer_Mask          : BYTE;
     MIDI_Data           : INTEGER;
     MIDI_Control        : INTEGER;
     Voice_Select        : INTEGER;
     Reg_Select          : INTEGER;
     Data_LOW            : INTEGER;
     Data_Hi             : INTEGER;
     IRQ_Status          : INTEGER;
     DRAM_Data           : INTEGER;
     Mix_Control         : INTEGER;
     IRQ_Control         : INTEGER;
     Timer_Control       : INTEGER;
     Timer_Data          : INTEGER;
     Ultra_ErrNo         : INTEGER;
     GF1_SEMA4           : INTEGER;
     IRQ_Pending         : INTEGER;
   end;

  { for new memory control scheme }
  PNode = ^Node;
  Node = record
    StartLoc : LONGINT;
    EndLoc   : LONGINT;
    Prev,
    Next     : PNode;
   end;

  FreeList = record
    BaseOffset : LONGINT;
    List       : PNode;
   end;


CONST
  IMAGE_MIDI : BYTE = 0;

CONST
  FRONT_VOICE = 0;
  NEAR_VOICE  = 1;
  REAR_VOICE  = 2;
  FAR_VOICE   = 3;
  UP_VOICE    = 4;
  DOWN_VOICE  = 5;

  XAXIS_3D    = 0;
  YAXIS_3D    = 1;
  ZAXIS_3D    = 2;

  { flag definitions }
  RIGHT_HEMI = $01; { sound is in right hemisphere }
  FRONT_HEMI = $02; { sound is in front }
  TOP_HEMI   = $04; { sound is above horizontal }

{ ---------------------------------------------------------- }


VAR
  GUSData   : ULTRA_DATA;
  UMemBlock : array[1..MaxNumBanks] of FreeList;
  UMemStruc : LONGINT;
  SaveExit : POINTER;  { Old exit procedure address }

{ ---------------------------------------------------------- }

procedure ClearError;
begin
  UltraOk := TRUE;
  UltraError := ULTRA_OK;
end;

procedure Default_Proc; far;
begin
  { Dummy procedure that does NOThing }
end;

procedure Default_Int_Proc(V : INTEGER); far;
begin
  { Dummy procedure that does NOThing }
end;

procedure Default_WORD_Proc(M : WORD); far;
begin
  { Dummy procedure that does NOThing }
end;

procedure Default_TwoWord_Proc(MS : WORD; MD : WORD); far;
begin
  { Dummy procedure that does NOThing }
end;

{ ---------------------------------------------------------- }

function Make_MS_WORD(x : WORD) : LONGINT;
begin
  Make_MS_WORD := (LONGINT(X) shl 16);
end;

function LSW(x : LONGINT) : WORD;
begin
  LSW := WORD(X);
end;

function MSW(x : LONGINT) : WORD;
begin
  MSW := WORD(X shr 16);
end;

function MSB(x : WORD) : BYTE;
begin
  MSB := BYTE(X shr 8);
end;

function LSB(x : WORD) : BYTE;
begin
  LSB := BYTE(x);
end;

{ Make GF1 address for direct chip i/o. }
function Addr_HIGH(x : LONGINT) : WORD;
begin
  Addr_HIGH := WORD(LONGINT(X shr LONGINT(7)) and LONGINT($1FFF));
end;

function Addr_LOW(x : LONGINT) : WORD;
begin
  Addr_LOW := WORD(LONGINT(X and LONGINT($7F)) shl 9);
end;

function UltraPeekData(PPort   : INTEGER;
                       Address : LONGINT) : BYTE;
begin
  asm
    pushf
    cli
  end;
  PORT[PPort+GF1_REG_Select] := SET_DRAM_LOW;   { 16 bits }
  PORTW[PPort+GF1_DATA_LOW] := WORD(Address);
  PORT[PPort+GF1_REG_Select] := SET_DRAM_HIGH;  { 8 bits }
  PORT[PPort+GF1_DATA_HI] := BYTE(Address shr 16);
  UltraPeekData := PORT[PPort+GF1_DRAM];
  asm
    popf
  end;
  ClearError;
end;


procedure UltraPokeData(PPort   : INTEGER;
                        Address : LONGINT;
                        Data    : BYTE);
begin
  PORT[PPort+GF1_REG_Select] := SET_DRAM_LOW;   { 16 bits }
  PORTW[PPort+GF1_DATA_LOW] := WORD(Address);
  PORT[PPort+GF1_REG_Select] := SET_DRAM_HIGH;  { 8 bits }
  PORT[PPort+GF1_DATA_HI] := BYTE(Address shr 16);
  PORT[PPort+GF1_DRAM] := Data;
  ClearError;
end;

procedure GF1_Delay;
{ This function is used as a 1.6*3 microsecond (or longer) delay.
  This is needed when trying TO change any of the 'self-modifying'
  bits in the voice registers. }
VAR
  I : INTEGER;
  M : BYTE;
begin
  for I := 0 TO 6 do
    M := PORT[GUSData.DRAM_data];
end;

function ParseToHex(VAR FromStr : STRING; VAR ToWord : WORD) : BOOLEAN;
{ Take the first number found.  Disregard ',' and ' ' chars }
begin
  ParseToHex := FALSE;
  while ((fromstr[1] = ' ') or
         (fromstr[1] = ',') or
         (fromstr[1] = '$')) and
         (length(fromstr)<>0) do
    delete(fromstr,1,1);
  if (ord(fromstr[1]) > 47) and
     (ord(fromstr[1]) < 58) and
     (fromstr <> '') then begin
      ParseToHex := TRUE;  { a number was found }
      ToWord := 0;
      while (ord(fromstr[1]) > 47) and
            (ord(fromstr[1]) < 58) and
            (FromStr <> '') do begin
        ToWord := (ToWord * $10) + (Ord(FromStr[1]) - Ord('0'));
        delete(FromStr,1,1);
      end;
      ParseToHex := TRUE;
  end;
end;

function ParseToNum(VAR FromStr : STRING; VAR ToWord : WORD) : BOOLEAN;
{ Take the first number found.  Disregard ',' and ' ' chars }
VAR
  StripStr : STRING[10];
  Code     : INTEGER;
begin
  ParseToNum := FALSE;
  while ((fromstr[1] = ' ') or
         (fromstr[1] = ',')) and
         (length(fromstr)<>0) do
    delete(fromstr,1,1);
  if (ord(fromstr[1]) > 47) and
     (ord(fromstr[1]) < 58) and
     (fromstr <> '') then begin
      ParseToNum := TRUE;  { a number was found }
      StripStr := '';
      while (ord(fromstr[1]) > 47) and
            (ord(fromstr[1]) < 58) and
            (FromStr <> '') do begin
        StripStr := StripStr + FromStr[1];
        delete(FromStr,1,1);
      end;
      Val(StripStr, ToWord, Code);
      if Code=0 then ParseToNum := TRUE;
  end;
end;

function UltraGetCfg(VAR Config : ULTRA_CFG) : BOOLEAN;
VAR
  EnvStr : STRING;
begin
  UltraGetCfg := FALSE;
  UltraOk := FALSE;
  UltraError := BAD_ENV_VAR;
  with Config do begin
    Base_Port     := $220;
    DRAM_DMA_Chan :=   01;
    ADC_DMA_Chan  :=   01;
    GF1_IRQ_Num   :=   11;
    MIDI_IRQ_Num  :=   05;
  end;
  EnvStr := GetEnv('ULTRASND');
  if EnvStr <> '' then
    if ParseToHex(EnvStr, Config.Base_Port) and
       ParseToNum(EnvStr, Config.DRAM_DMA_Chan) and
       ParseToNum(EnvStr, Config.ADC_DMA_Chan) and
       ParseToNum(EnvStr, Config.GF1_IRQ_Num) and
       ParseToNum(EnvStr, Config.MIDI_IRQ_Num) then begin
       UltraGetCfg := TRUE;
       ClearError;
    end;
end;

function UltraPing(PPort : WORD) : BOOLEAN;
VAR
  Val0,
  Val1      : BYTE;
  Save_Val0 : BYTE;
  Save_Val1 : BYTE;
begin
  { Save Current Values }
  Save_Val0 := UltraPeekData(PPort, 0);
  Save_Val1 := UltraPeekData(PPort, 1);

  UltraPokeData(PPort, 0, $AA);
  UltraPokeData(PPort, 1, $55);

  Val0 := UltraPeekData(PPort, 0);
  Val1 := UltraPeekData(PPort, 1);

  { Restore data TO old values }
  UltraPokeData(PPort, 0, Save_Val0);
  UltraPokeData(PPort, 1, Save_Val1);

  if (Val0 = $AA) and
     (Val1 = $55) then begin
    UltraPing := TRUE;
    ClearError;
  end
  else begin
    UltraPing := FALSE;
    UltraOk := FALSE;
    UltraError := No_Ultra;
  end;
end;

function UltraProbe(PPort : WORD) : BOOLEAN;
begin
  GUSData.Base_Port := PPort;
  GUSData.MIDI_Data := PPort + GF1_MIDI_Data;
  GUSData.MIDI_Control := PPort + GF1_MIDI_CTRL;
  GUSData.Voice_Select := PPort + GF1_PAGE;
  GUSData.Reg_Select := PPort + GF1_REG_SELECT;
  GUSData.Data_LOW := PPort + GF1_Data_LOW;
  GUSData.Data_Hi := PPort + GF1_Data_HI;
  GUSData.IRQ_Status := PPort + GF1_IRQ_STAT;
  GUSData.DRAM_Data := PPort + GF1_DRAM;
  GUSData.Mix_Control := PPort + GF1_MIX_CTRL;
  GUSData.Timer_Control := PPort + GF1_TIMER_CTRL;
  GUSData.Timer_Data := PPort+ GF1_TIMER_DATA;
  GUSData.IRQ_Control := PPort + GF1_IRQ_CTRL;

  { Pull a Reset }
  PORT[GUSData.Reg_Select] := MASTER_RESET;
  PORT[GUSData.Data_Hi] := $00;

  GF1_Delay;
  GF1_Delay;

  { Release the reset }
  PORT[GUSData.Reg_Select] := MASTER_RESET;
  PORT[GUSData.Data_Hi] := GF1_MASTER_RESET;

  GF1_Delay;
  GF1_Delay;

  UltraProbe := UltraPing(PPort);
end;


function UltraGetOutput : BOOLEAN;
begin
  UltraGetOutput := (GUSData.Mix_Image and ENABLE_OUTPUT) = 0;
end;

procedure UltraEnableOutput;
begin
  GUSData.Mix_Image := GUSData.Mix_Image and (not ENABLE_OUTPUT);
  PORT[GUSData.Mix_Control] := GUSData.Mix_Image;
end;

procedure UltraDisableOutput;
begin
  GUSData.Mix_Image := GUSData.Mix_Image or ENABLE_OUTPUT;
  PORT[GUSData.Mix_Control] := GUSData.Mix_Image;
end;

function UltraGetLineIn : BOOLEAN;
begin
  UltraGetLineIn := (GUSData.Mix_Image and ENABLE_LINE_IN) = 0;
end;

procedure UltraEnableLineIn;
begin
  GUSData.Mix_Image := GUSData.Mix_Image and (not ENABLE_LINE_IN);
  PORT[GUSData.Mix_Control] := GUSData.Mix_Image;
end;

procedure UltraDisableLineIn;
begin
  GUSData.Mix_Image := GUSData.Mix_Image or ENABLE_LINE_IN;
  PORT[GUSData.Mix_Control] := GUSData.Mix_Image;
end;

function UltraGetMicIn : BOOLEAN;
begin
  UltraGetMicIn := (GUSData.Mix_Image and ENABLE_MIC_IN) <> 0;
end;

procedure UltraEnableMicIn;
begin
  GUSData.Mix_Image := GUSData.Mix_Image or ENABLE_MIC_IN;
  PORT[GUSData.Mix_Control] := GUSData.Mix_Image;
end;

procedure UltraDisableMicIn;
begin
  GUSData.Mix_Image := GUSData.Mix_Image and (not ENABLE_MIC_IN);
  PORT[GUSData.Mix_Control] := GUSData.Mix_Image;
end;

function UltraReset(Voices : INTEGER) : BOOLEAN;
VAR
  V         : BYTE;
  Select,
  Data_LOW,
  Data_Hi : INTEGER;
begin
  if (Voices < 14) or (Voices > 32) then begin
    UltraReset := FALSE;
    UltraOk := FALSE;
    UltraError := Bad_Num_OF_Voices;
    exit;
  end;

  GUSData.Voices := Voices;
  GUSData.Timer_Ctrl := 0;
  GUSData.Timer_Mask := 0;

  Select := GUSData.Reg_Select;
  Data_LOW := GUSData.Data_LOW;
  Data_Hi := GUSData.Data_Hi;

  { set these TO zero so the they don't get summed in for voices that are
    not running. if their volumes are not at zero, whatever value they
    are pointing at, will get summed into the output. By setting that
    location TO 0, that voice will have no contribution TO the output
    (2 locations are done in case voice is set TO 16 bits ... ) }
  UltraPokeData(GUSData.BASE_PORT,0,0);
  UltraPokeData(GUSData.BASE_PORT,1,0);

  asm
    pushf
    cli
  end;
  { Pull a register-level reset on the card. }
  PORT[Select] := MASTER_RESET;
  PORT[Data_Hi] := $00;

  for V := 0 TO 9 do
    GF1_Delay;

  PORT[Select] := MASTER_RESET;
  PORT[Data_Hi] := GF1_MASTER_RESET;

  for V := 0 TO 9 do
    GF1_Delay;

  { Reset the MIDI PORT }
  PORT[GUSData.MIDI_Control] := MIDI_RESET;
  for V := 0 TO 9 do
    GF1_Delay;
  PORT[GUSData.MIDI_Control] := $00;

  { Clear all interrupts }
  PORT[Select] := DMA_Control;
  PORT[Data_Hi] := $00;
  PORT[Select] := TIMER_Control;
  PORT[Data_Hi] := $00;
  PORT[Select] := SAMPLE_Control;
  PORT[Data_Hi] := $00;

  { set the number of active voices }
  PORT[Select] := SET_VOICES;
  PORT[Data_Hi] := Lo((Voices-1) or $C0);

  { Clear interrupts on voices }
  { Reading the status ports will clear the irqs }
  V := PORT[GUSData.IRQ_Status];
  PORT[Select] := DMA_Control;
  V := PORT[Data_Hi];
  PORT[Select] := SAMPLE_Control;
  V := PORT[Data_Hi];
  PORT[select] := GET_IRQV;
  V := PORT[Data_Hi];

  for V := 0 TO Voices-1 do begin
    { Select the proper voice }
    PORT[GUSData.Voice_Select] := V;

    { Stop the voice and volume }
    PORT[Select] := SET_Control;
    PORT[Data_Hi] := (VOICE_STOPPED or STOP_VOICE);
    PORT[Select] := SET_VOLUME_CONTROL;
    PORT[Data_Hi] := (VOLUME_STOPPED or STOP_VOLUME);

    { Wait 4.8 Microseconds or more }
    GF1_Delay;

    { Initialize each voice specific registers.  This is not really
      necessary, but is nice for completeness sake.  Each application
      will set up these TO whatever values it needs }
    PORT[Select] := SET_FREQUENCY;
    PORTW[Data_LOW] := $0400;

    PORT[Select] := SET_START_HIGH;
    PORTW[Data_LOW] := $0;

    PORT[Select] := SET_START_LOW;
    PORTW[Data_LOW] := $0;

    PORT[Select] := SET_END_HIGH;
    PORTW[Data_LOW] := $0;

    PORT[Select] := SET_END_LOW;
    PORTW[Data_LOW] := $0;

    PORT[Select] := SET_VOLUME_RATE;
    PORT[Data_HI] := $01;

    PORT[Select] := SET_VOLUME_START;
    PORT[Data_HI] := $10;

    PORT[Select] := SET_VOLUME_END;
    PORT[Data_HI] := $E0;

    PORT[Select] := SET_VOLUME;
    PORTW[Data_LOW] := $0000;

    PORT[Select] := SET_ACC_HIGH;
    PORTW[Data_LOW] := $0;

    PORT[Select] := SET_ACC_LOW;
    PORTW[Data_LOW] := $0;

    PORT[Select] := SET_BALANCE;
    PORT[Data_Hi] := $07;
  end;

  V := PORT[GUSData.IRQ_STATUS];
  PORT[Select] := DMA_CONTROL;
  V := PORT[Data_Hi];
  PORT[Select] := SAMPLE_CONTROL;
  V := PORT[DATA_Hi];
  PORT[Select] := GET_IRQV;
  V := PORT[Data_Hi];

  { set up GF1 Chip for interrupts and enable DACs }
  PORT[Select] := MASTER_RESET;
  PORT[Data_Hi] := (GF1_MASTER_RESET or GF1_OUTPUT_ENABLE or GF1_MASTER_IRQ);
  { Restore the IRQ state }
  asm
    popf
  end;

  UltraReset := TRUE;
  ClearError;
end;


procedure UltraStopVoice(Voice : INTEGER);
VAR
  Data : BYTE;
begin
  { Save the current IRQ State }
  asm
    pushf
    cli
  end;
  { Select the proper voice }
  PORT[GUSData.Voice_Select] := Lo(Voice);

  { turn off the roll over bit first }
  PORT[GUSData.Reg_Select] := GET_VOLUME_CONTROL;
  Data := PORT[GUSData.Data_Hi];
  Data := Data and (not VC_ROLLOVER);
  PORT[GUSData.Reg_Select] := SET_VOLUME_CONTROL;
  PORT[GUSData.Data_Hi] := Data;
  GF1_Delay;
  PORT[GUSData.Data_Hi] := Data;

  { Now stop the voice }
  PORT[GUSData.Reg_Select] := GET_CONTROL;
  Data := PORT[GUSData.Data_Hi];
  Data := Data and (not VC_WAVE_IRQ); { Disable the IRQs  }
  Data := Data or (VOICE_STOPPED or STOP_VOICE);
  PORT[GUSData.Reg_Select] := SET_Control;
  PORT[GUSData.Data_Hi] := Data;
  GF1_DELAY;
  PORT[GUSData.Data_hi] := Data;
  { Restore the IRQ state }
  asm
    popf
  end;
end;

{ This function returns the # of K found on the card }
function UltraSizeDRAM : INTEGER;
VAR
  I,
  Loc       : LONGINT;
  Val,
  Save0,
  Save1     : BYTE;
  Base_Port : WORD;
  BreakOut  : BOOLEAN;
begin
  UltraSizeDRAM := 0;

  Base_Port := GUSData.Base_Port;

  { Save First Location }
  Save0 := UltraPeekData(Base_Port, 0);

  { See if there is first block there ... }
  UltraPokeData(Base_Port, 0, $AA);
  if (UltraPeekData(Base_Port,0) <> $AA) then exit;

  { Now zero it out so that I can check for mirroring }
  UltraPokeData(Base_Port, 0, $00);

  BreakOut := FALSE;
  I := 0;
  while (not BreakOut) and
        (I < 1023) do begin
    inc(I);

    { Check for mirroring }
    Val := UltraPeekData(Base_Port, 0);
    if Val = 0 then begin
      Loc := LONGINT(I) shl 10;

      { Save Location so its a non-destructive sizing }
      Save1 := UltraPeekData(Base_Port, Loc);

      UltraPokeData(Base_Port, Loc, $AA);
      if (UltraPeekData(Base_Port, Loc) <> $AA) then
        BreakOut := TRUE
      else
        UltraPokeData(Base_Port, Loc, Save1);
    end
    else BreakOut := TRUE;
  end;

  { Now restore location zero ... }
  if I=1023 then I:=1024;
  UltraPokeData(Base_Port,0,Save0);
  UltraSizeDRAM := I;
end;

{ ------------------------------------------------------------------------- }
{ This is the new memory control system by Kurt Kennett                     }
{ ------------------------------------------------------------------------- }

procedure UltraMemInit;
VAR
  Runner : BYTE;
  Maker  : PNode;
begin
  {
  UMemStruc := sizeof(UMemBlock);
  report heap space used only. }
  FillChar(UMemBlock, sizeof(UMemBlock), 0);
  for Runner := 0 TO 3 do
    with UMemBlock[Runner+1] do begin
      BaseOffset := Runner * (BlockSizeK*OneK);
      New(List);
      inc(UMemStruc, sizeof(List^));
      FillChar(List^, sizeof(List^), 0);
      List^.EndLoc := BlockSizeK*OneK-1;
      List^.Next   := nil;
      List^.Prev   := nil;
    end;
  UMemInited := TRUE;
end;

procedure UltraMemClose;
VAR
  Runner : BYTE;
  Killer,
  MList  : PNode;
begin
  for Runner := 1 TO 4 do begin
    MList := UMemBlock[Runner].List;
    while MList <> nil do begin
      Killer := MList;
      MList := MList^.Next;
      dec(UMemStruc, sizeof(Killer^));
      Dispose(Killer);
    end;
  end;
end;

function UltraMaxAvail : LONGINT;
VAR
  Largest : LONGINT;
  MList   : PNode;
  Count   : BYTE;
begin
  Largest := 0;
  for Count := 1 TO 4 do begin
    MList := UMemBlock[Count].List;
    while MList <> nil do begin
      if (MList^.EndLoc-MList^.StartLoc+1) > Largest then
        Largest := (MList^.EndLoc-MList^.StartLoc+1);
      MList := MList^.Next;
    end;
  end;
  UltraMaxAvail := Largest;
end;

function UltraMaxAlloc : LONGINT;
begin
  UltraMaxAlloc := UltraMaxAvail;
end;

function UltraMemAvail : LONGINT;
VAR
  MemCount : LONGINT;
  MList    : PNode;
  Count    : BYTE;
begin
  MemCount := 0;
  for Count := 1 TO 4 do begin
    MList := UMemBlock[Count].List;
    while MList <> nil do begin
      inc(MemCount, (MList^.EndLoc-MList^.StartLoc+1));
      MList := MList^.Next;
    end;
  end;
  UltraMemAvail := MemCount;
end;

function UltraMemAlloc(    Size     : LONGINT;
                       VAR Location : LONGINT) : BOOLEAN;
VAR
  BlockUse : BYTE;
  BlockEnd : LONGINT;
  MList    : PNode;
begin
  UltraMemAlloc := FALSE;
  if not UMemInited then begin
    { Memory structures not initialized }
    Location := -1;
    exit;
  end;
  if (Size mod 32) <> 0 then
    inc(Size, 32 - (Size mod 32)); { bring size up TO a 32 BYTE boundary }
  if Size > BlockSizeK*OneK then begin
    { Size is bigger than allowed sample size }
    Location := -1;
    exit;
  end;
  BlockUse := 1;
  while BlockUse <= MaxNumBanks do begin
    MList := UMemBlock[BlockUse].List;
    { Scan the list for a free space TO use }
    while MList <> nil do begin
      if (MList^.EndLoc-MList^.StartLoc+1) >= Size then begin
        { set the end of the block TO allocate }
        BlockEnd := MList^.StartLoc + Size - 1;

        { Prepare the location parameter with the actual address }
        Location := UMemBlock[BlockUse].BaseOffset + MList^.StartLoc;

        { set the free block's new start position }
        MList^.StartLoc := BlockEnd+1;

        { exit with a TRUE result }
        UltraMemAlloc := TRUE;
        ClearError;
        exit;
      end;
      MList := MList^.Next;
    end;
    inc(BlockUse);
  end;

  { No memory left for block of specified size }
  UltraOk := FALSE;
  UltraError := INSUFFICIENT_GUS_MEM;
end;


function UltraMemFree(    Size : LONGINT;
                      Location : LONGINT) : BOOLEAN;
VAR
  BlockUse : BYTE;
  NextNext,
  MList,
  Temp     : PNode;
begin

  UltraMemFree := FALSE;
  if not UMemInited then begin
    { Memory structures not initialized }
    exit;
  end;
  if (Size mod 32) <> 0 then
    inc(Size, 32 - (Size mod 32)); { bring size up TO a 32 BYTE boundary }
  if Size > BlockSizeK*OneK then begin
    { Size is bigger than allowed sample size }
    exit;
  end;

  { find appropriate block }
  BlockUse := BYTE((Location div (BlockSizeK*OneK))+1);

  { create a temporary block at the specified position }
  New(Temp);
  inc(UMemStruc, sizeof(Temp^));
  FillChar(Temp^, sizeof(Temp^), 0);
  Temp^.StartLoc := Location - UMemBlock[BlockUse].BaseOffset;
  Temp^.EndLoc := Temp^.StartLoc + Size - 1;

  { Insert it at the correct position in the list }
  MList := UMemBlock[BlockUse].List;
  while (MList <> nil) and
        (MList^.StartLoc < Temp^.StartLoc) do MList := MList^.Next;
  if MList= nil then begin
    dec(UMemStruc, sizeof(Temp^));
    Dispose(Temp);
    { Inappropriate location passed TO freemem }
    exit;
  end;
  Temp^.Next := MList;
  Temp^.Prev := MList^.Prev;
  if Temp^.Prev <> nil then
    Temp^.Prev^.Next := Temp;
  MList^.Prev := Temp;
  if Temp^.StartLoc < UMemBlock[BlockUse].List^.StartLoc then
    UMemBlock[BlockUse].List := Temp;  { Mettre le nouveau bloc en premier }

  { merge consecutive blocks }
  MList := UMemBlock[BlockUse].List;
  while MList <> nil do begin
    if MList^.Next <> nil then begin
      if (MList^.Next^.StartLoc-1) = MList^.EndLoc then begin
        MList^.EndLoc := MList^.Next^.EndLoc;
        NextNext := MList^.Next^.Next;
        dec(UMemStruc,sizeof(MList^.Next^));
        Dispose(MList^.Next);
        MList^.Next := NextNext;
        if NextNext <> nil then
          NextNext^.Prev := MList;
      end
      else MList := MList^.Next;
    end
    else MList := MList^.Next;
  end;

  UltraMemFree := TRUE;
end;

{procedure DisplayMem;
var
    BlockUse : BYTE;
    MList    : PNode;
begin
 for BlockUse:=1 to 4 do begin
   Writeln('Bank ',BlockUse);
   MList := UMemBlock[BlockUse].List;
   while MList <> nil do begin
     Writeln('Start: ',MList^.StartLoc,' end: ',MList^.EndLoc);
     MList:=MList^.Next;
   end;
   repeat until Readkey=#13;
 end;
end;}


function UltraOpen(VAR Config : ULTRA_CFG;
                       Voices : INTEGER) : BOOLEAN;
begin
  GUSData.Base_Port := Config.Base_Port;
  GUSData.DRAM_DMA_Chan := Config.DRAM_DMA_Chan;
  GUSData.ADC_DMA_Chan := Config.ADC_DMA_Chan;
  GUSData.GF1_IRQ_Num := Config.GF1_IRQ_Num;
  GUSData.MIDI_IRQ_Num := Config.MIDI_IRQ_Num;
  GUSData.Mix_Image := $0B;
  GUSData.Voices := Voices;

  if not UltraProbe(GUSData.Base_Port) then begin
    UltraOpen := FALSE;
    exit;
  end;

  UltraDisableLineIn;
  UltraDisableMicIn;
  UltraDisableOutput;

  if not UltraReset(Voices) then begin
    UltraOpen := FALSE;
    exit;
  end;

  UltraEnableOutput;

  UltraOpen := TRUE;
  ClearError;
end;


function UltraClose : BOOLEAN;
begin
  UltraDisableOutput;
  UltraDisableLineIn;
  UltraDisableMicIn;
  UltraReset(14);

  UltraClose := TRUE;
  ClearError;
end;


procedure Setup_GUSData;
begin
  FillChar(GUSData, sizeof(GUSData), 0);
  with GUSData do begin
    Flags              := ULTRA_PRESENT;
    Base_Port          := 220;
    DRAM_DMA_Chan      := 1;
    ADC_DMA_Chan       := 1;
    GF1_IRQ_Num        := 11;
    MIDI_IRQ_Num       := 5;
    Mix_Image          := $08;
    Voices             := 32;
    Image_MIDI         := 0;
    Used_Voices        := 0;
  end;
end;


procedure UltraSetBalance(Voice : INTEGER;
                          Data  : BYTE);
VAR
  Temp : BYTE;
begin
  { Make sure balance is between 0 and 15 }
  Temp := Data and $0F;

  asm
    pushf
    cli
  end;
  PORT[GUSData.Voice_Select] := Lo(Voice);
  PORT[GUSData.Reg_Select] := SET_BALANCE;
  PORT[GUSData.Data_Hi] := Temp;
  asm
    popf
  end;
end;

{ ------------------------------------------------------------------------- }

{ exit procedure }
procedure ShutDown; far;
begin
  ExitProc := SaveExit;    { restore old exit procedure }
  UltraMemClose;
  UltraClose;
end;

{ ------------------------------------------------------------------------- }

function DetecteGUS(Affichage:BOOLEAN):BOOLEAN;
begin
 GUS_Presente:=FALSE;

 UltraOk := FALSE;
 Setup_GUSData;

 if UltraGetCfg(Ultra_Config) then { Lire la variable ULTRASND }
    Ultra_Installed := TRUE;

 if (Ultra_Installed and UltraProbe(Ultra_Config.Base_Port) and
    UltraOpen(Ultra_Config,Init_Voices)) then begin
   UltraMemInit;                   { Initialise le gestionnaire de mem GUS }
   SaveExit := ExitProc;
   ExitProc := @ShutDown;
   if UltraReset(Init_Voices) then GUS_Presente:=TRUE;
 end;

 with Ultra_Config do
 if (Affichage and GUS_presente) then Writeln('Gravis UltraSound found at ',EntierHexa(Base_Port,3)+'h',
            ' (IRQ ',GF1_IRQ_Num,', DMA ',Dram_DMA_Chan,'). ');

 DetecteGUS:=GUS_Presente;
end; {DetecteGUS}

{ ------------------------------------------------------------------------- }

END.
