{ͻ}
{                          --== LZ.PAS ==--                        }
{                                                                      }
{ Unit pour compacter/dcompacter des donnes avec l'algorithme de    }
{ Lempel-Ziv, de fichier->fichier ou de mmoire->mmoire               }
{                                                                      }
{ AUTEURS: Fabrice Couteau et Tom Pfau                                 }
{Ķ}
{ Compatibilit: 286           Dbut d'criture 10/04/95 }
{                              Dernires modifications 26/07/98 }
{ͼ}

UNIT LZ;

{$A+,G+,Q-,R-,S-}

INTERFACE

{$I Types}

{==========================================================}

procedure CompLZf  (source,destination: String80; var erreur: WORD);
procedure DecompLZf(source,destination: String80; var erreur: WORD);

procedure CompLZmem  (var source; taille_source: WORD; var destination; var taille_dest: WORD);
function  DecompLZmem(var source; var destination): WORD;

{==========================================================}

IMPLEMENTATION

USES Fichiers;

CONST clear      = 256;    {Clear code}
      eof        = 257;    {end of file marker}
      first_free = 258;    {First free code}
      maxmax     = 4096;   {Max code+1}
      buf_length = 5*1024; {5ko input/output buffers (maximum!)}

TYPE  buffer=array[1..buf_length] of CHAR;

{==========================================================}

procedure CompLZf(source,destination: String80; var erreur: WORD);
{compacte fichier source -> fichier destination}

type hash_rec=record
               first: WORD; {First entry with this value}
               next : WORD; {Next entry along chain}
               CHAR : CHAR; {Suffix CHAR}
              end;

     hash=array[1..maxmax] of hash_rec;

var fin,fout: Fichier;
    input_handle,output_handle: WORD;
    hash_table: ^hash;
    input_data,output_data: ^buffer;
    ptr_hash,ptr_input,ptr_output: POINTER;
    input_offset,input_size,bit_offset: WORD;
    prefix_code,free_code,max_code,nbits: WORD;
    first_use: BOOLEAN;
    k: CHAR;

begin
 fin.Ouvre(lecture,source);
 if f_erreur<>0 then begin erreur:=f_erreur; exit end;

 fout.Cree(not(teste_existe),destination);
 if f_erreur<>0 then begin erreur:=f_erreur; exit end;

 if MemAvail<sizeof(hash)+2*buf_length+64 then
  begin erreur:=8; exit end; {mmoire insuffisante}

 input_handle:=fin.GetHandle; output_handle:=fout.GetHandle;

 New(hash_table);  ptr_hash:=hash_table;
 New(input_data);  ptr_input:=input_data;
 New(output_data); ptr_output:=output_data;

 input_offset:=0; input_size:=0; bit_offset:=0;

 asm
  push DS
  cld

  call @init_table      {Initialize the table and some vars}
  mov AX,clear          {Write a clear code}
  call @write_code
  call @read_char       {Read first CHAR}
 @l4:
  xor AH,AH             {Turn CHAR into code}
 @l4a:
  mov prefix_code,AX    {set prefix code}
  call @read_char       {Read next CHAR}
  jc @l17               {Carry means eof}
  mov k,AL              {Save CHAR in k}
  mov BX,prefix_code    {Get prefix code}
  call @lookup_code     {See if this pair in table}
  jnc @l4a              {nc means yes, new code in AX}
  call @add_code        {Add pair to table}
  mov AX,prefix_code    {Write old prefix code}
  call @write_code
  mov AL,k              {Get last CHAR}
  cmp BX,max_code       {Exceed code size?}
  jl @l4                {less means no}
  cmp nbits,12          {Currently less than 12 bits?}
  jl @l14               {yes}
  mov AX,clear          {Write a clear code}
  call @write_code
  call @init_table      {Reinit table}
  mov AL,k              {get last CHAR}
  jmp @l4               {Start over}
 @l14:
  inc nbits             {Increase number of bits}
  shl max_code,1        {Double max code size}
  jmp @l4               {Get next CHAR}
 @l17:
  mov AX,prefix_code    {Write last code}
  call @write_code
  mov AX,eof            {Write eof code}
  call @write_code
  mov AX,bit_offset     {Make sure buffer is flushed to file}
  or AX,AX
  jz @l18
  mov DX,AX             {convert bits to bytes}
  and DX,7
  shr AX,3
  or DX,DX              {if extra bits, make sure they get}
  jz @l17a              {written}
  inc AX
 @l17a:
  call @flush
 @l18:
  jmp @fin              {done}

 @init_table:
  mov nbits,9           {set code size to 9}
  mov max_code,512      {set max code to 512=2^nbits}
  mov AX,-1             {Unused flag}
  mov CX,256*5/2        {Clear first 256 entries}
  les DI,ptr_hash       {Point to first entry in hash table}
  rep stosw             {Clear it out}
  mov free_code,first_free    {set next code to use}
  retn

 @write_code:
  push AX               {Save code}
  mov AX,bit_offset     {Get bit offset}
  mov CX,nbits          {Adjust bit offset by code size}
  add bit_offset,CX
  mov DX,AX             {Convert bit offset to BYTE offset}
  and DX,7
  shr AX,3
  cmp AX,buf_length-4   {Approaching end of buffer?}
  jl @wc1               {less means no}
  call @flush           {Write the buffer}
  push DX               {DX contains offset within BYTE}
  add DX,nbits          {adjust by code size}
  mov bit_offset,DX     {new bit offset}
  pop DX                {restore DX}
  lds SI,ptr_output
  push DS
  pop ES
  mov DI,SI             {DS:SI=ES:DI -> output_data[0]}
  add SI,AX             {Point to last BYTE}
  movsb                 {move BYTE to first position}
  xor AX,AX             {BYTE offset of zero}
 @wc1:
  les DI,ptr_output
  add DI,AX             {Point into buffer}
  pop AX                {Restore code}
  mov CX,DX             {offset within BYTE}
  xor DX,DX             {DX will catch bits rotated out}
  jcxz @wc3             {if offset in BYTE is zero, skip shift}
 @wc2:
  shl AX,1              {Rotate code}
  rcl DX,1
  loop @wc2
  or AL,ES:[DI]         {Grab bits currently in buffer}
 @wc3:
  stosw                 {Save data}
  mov AL,DL             {Grab extra bits}
  stosb                 {and save}
  retn

 @flush:
  push AX
  push BX
  push CX
  push DX
  mov CX,AX             {AX=number of bytes to write}
  mov AH,40h
  mov BX,output_handle
  lds DX,ptr_output
  int 21h               {write output file}
  pop DX
  pop CX
  pop BX
  pop AX
  retn

 @read_char:
  mov DI,input_offset   {Anything left in buffer?}
  cmp DI,input_size
  jl @rd1               {less means yes}
  push BX
  push CX
  push DX
  mov AH,3Fh
  mov BX,input_handle
  mov CX,buf_length
  lds DX,ptr_input
  int 21h               {Read next chunk of input, AX=number of bytes read}
  pop DX
  pop CX
  pop BX
  or AX,AX              {Anything left?}
  jz @rd2               {equal means no, finished}
  mov input_size,AX     {Save bytes read}
  mov input_offset,0    {Point to beginning of buffer}
  xor DI,DI
 @rd1:
  lds SI,ptr_input
  add SI,DI             {Point at character}
  lodsb                 {Read it in}
  inc input_offset      {Adjust POINTER}
  clc                   {Success}
  retn
 @rd2:
  stc                   {Nothing left}
  retn

 @lookup_code:
  lds SI,ptr_hash       {point to hash table}
  mov DI,BX
  shl DI,2
  add DI,BX             {DI=5*BX}
  add SI,DI             {convert code to address}
  mov first_use,TRUE    {flag}
  cmp WORD ptr [SI],-1  {(hash.first) Has this code been used?}
  je @gc4               {equal means no}
  mov first_use,FALSE   {set flag}
  mov BX,[SI]           {(hash.first) Get first entry}
 @gc2:
  mov SI,BX
  shl SI,2
  add SI,BX             {convert code to address: SI=5*BX}
  cmp [SI+4],AL         {(hash.CHAR) is CHAR the same?}
  jne @gc3              {ne means no}
  clc                   {success}
  mov AX,BX             {put found code in AX}
  retn
 @gc3:
  cmp WORD ptr [SI+2],-1   {(hash.next) More left with this prefix?}
  je @gc4                  {equal means no}
  mov BX,[SI+2]         {(hash.next) get next code}
  jmp @gc2              {try again}
 @gc4:
  stc                   {not found}
  retn

 @add_code:
  mov BX,free_code      {Get code to use}
  mov DI,SI
  lds SI,ptr_hash
  add SI,DI
  cmp first_use,TRUE    {First use of this prefix?}
  je @ac1               {equal means yes}
  mov [SI+2],BX         {(hash.next) point last use to new entry}
  jmp @ac2
 @ac1:
  mov [SI],BX           {(hash.first) Point first use to new entry}
 @ac2:
  cmp BX,maxmax         {Have we reached code limit?}
  je @ac3               {equal means yes, just return}
  mov SI,BX
  shl SI,2
  add SI,BX             {convert code to address: SI=5*BX}
  mov WORD ptr [SI],-1    {(hash.first) initialize pointers}
  mov WORD ptr [SI+2],-1  {(hash.next)}
  mov BYTE ptr [SI+4],AL  {(hash.CHAR) save suffix CHAR}
  inc free_code         {adjust next code}
 @ac3:
  retn

 @fin:
  pop DS
 end;

 Dispose(output_data); Dispose(input_data); Dispose(hash_table);

 fin.Ferme; fout.Ferme;
end; {CompLZf}

{----------------------------------------------------------}

procedure DecompLZf(source,destination: String80; var erreur: WORD);
{dcompacte fichier source -> fichier destination}

type hash_rec=record
               next: WORD; {Prefix code}
               CHAR: CHAR; {Suffix CHAR}
              end;

     hash=array[1..maxmax] of hash_rec;

var fin,fout: Fichier;
    input_handle,output_handle: WORD;
    hash_table: ^hash;
    input_data,output_data: ^buffer;
    ptr_hash,ptr_input,ptr_output: POINTER;
    output_offset,bit_offset: WORD;
    free_code,max_code,nbits: WORD;
    cur_code,old_code,in_code,stack_count: WORD;
    fin_char,k: CHAR;

begin
 fin.Ouvre(lecture,source);
 if erreur<>0 then begin erreur:=f_erreur; exit end;

 fout.Cree(not(teste_existe),destination);
 if erreur<>0 then begin erreur:=f_erreur; exit end;

 if MemAvail<sizeof(hash)+2*buf_length+64 then
  begin erreur:=8; exit end; {mmoire insuffisante}

 input_handle:=fin.GetHandle; output_handle:=fout.GetHandle;

 New(hash_table);  ptr_hash:=hash_table;
 New(input_data);  ptr_input:=input_data;
 New(output_data); ptr_output:=output_data;

 output_offset:=0; bit_offset:=0; stack_count:=0;

 asm
  push DS
  cld

  call @init_table      {Initialize variables}
  mov AH,3Fh
  mov BX,input_handle
  mov CX,buf_length
  lds DX,ptr_input
  int 21h               {Read from input}
 @l1:
  call @read_code       {Get a code}
  cmp AX,eof            {end of file?}
  jne @l2               {no}
  cmp output_offset,0   {Data in output buffer?}
  jz @l1a               {no}
  mov AH,40h
  mov BX,output_handle
  mov CX,output_offset
  lds DX,ptr_output
  int 21h               {Flush buffer}
 @l1a:
  jmp @fin              {done}
 @l2:
  cmp AX,clear          {Clear code?}
  jne @l7               {no}
  call @init_table      {Initialize table}
  call @read_code       {Read next code}
  mov cur_code,AX       {Initialize variables}
  mov old_code,AX
  mov k,AL
  mov fin_char,AL
  call @write_char      {Write character}
  jmp @l1               {Get next code}
 @l7:
  mov cur_code,AX       {Save new code}
  mov in_code,AX
  cmp AX,free_code      {Code in table? (k<w>k<w>k)}
  jl @l11               {yes}
  mov AX,old_code       {get previous code}
  mov cur_code,AX       {make current}
  mov AL,fin_char       {get old last CHAR}
  push AX               {push it}
  inc stack_count
 @l11:
  cmp cur_code,255      {Code or character?}
  jle @l15              {CHAR}
  lds SI,ptr_hash
  mov BX,cur_code       {Convert code to address}
  mov DI,BX
  shl DI,1
  add DI,BX             {DI=3*BX}
  add SI,DI
  lodsw                 {Get prefix code}
  mov cur_code,AX       {Save it}
  lodsb                 {Get suffix CHAR}
  push AX               {Push it}
  inc stack_count
  jmp @l11              {Translate again}
 @l15:
  mov AX,cur_code       {Get code}
  mov fin_char,AL       {Save as final, k}
  mov k,AL
  push AX               {Push it}
  inc stack_count
  mov CX,stack_count    {Pop stack}
  jcxz @l18             {if anything there}
 @l17:
  pop AX
  call @write_char
  loop @l17
 @l18:
  mov stack_count,0     {Clear count on stack}
  call @add_code        {Add new code to table}
  mov AX,in_code        {Save input code}
  mov old_code,AX
  mov BX,free_code      {Hit table limit?}
  cmp BX,max_code
  jl @l23               {Less means no}
  cmp nbits,12          {Still within twelve bits?}
  je @l23               {no (next code should be clear)}
  inc nbits             {Increase code size}
  shl max_code,1        {Double max code}
 @l23:
  jmp @l1               {Get next code}

 @read_code:
  mov AX,bit_offset     {Get bit offset}
  add AX,nbits          {Adjust by code size}
  xchg bit_offset,AX    {Swap}
  mov DX,AX             {Calculate BYTE offset}
  and DX,7
  shr AX,3
  cmp AX,buf_length-4   {Approaching end of buffer?}
  jl @rd0               {no}
  push DX               {Save offset in BYTE}
  add DX,nbits          {Calculate new bit offset}
  mov bit_offset,DX
  mov CX,buf_length     {Length of buffer}
  sub CX,AX             {Calculate bytes left}
  les DI,ptr_input      {Point to beginning of buffer}
  push ES
  pop DS
  mov SI,DI             {DS:SI=ES:DI -> input_data[0]}
  add SI,AX             {Point to CHAR}
  rep movsb             {Move last chars down}
  mov CX,AX             {AX=number of bytes to read}
  mov AH,3Fh
  mov BX,input_handle
  mov DX,DI             {DS:DX -> next CHAR to read}
  int 21h               {Fill rest of buffer}
  xor AX,AX             {Clear AX}
  pop DX                {Restore offset in BYTE}
 @rd0:
  lds SI,ptr_input
  add SI,AX             {Point to CHAR}
  lodsw                 {Get WORD}
  mov BX,AX             {Save in AX}
  lodsb                 {Next BYTE}
  mov CX,DX             {Offset in BYTE}
  jcxz @rd2             {if zero, skip shifts}
 @rd1:
  shr AL,1              {Put code in low (code size) bits of BX}
  rcr BX,1
  loop @rd1
 @rd2:
  mov AX,BX             {put code in AX}
  mov CX,nbits
  sub CL,9
  mov BH,2
  shl BH,CL
  dec BH                {BH=(2 shl (nbits-9))-1}
  and AH,BH             {mask off unwanted high bits of AX -> keep nbits}
  retn

 @init_table:
  mov nbits,9           {Initialize variables}
  mov max_code,512
  mov free_code,first_free
  retn

 @write_char:
  mov DI,output_offset  {Get offset in buffer}
  cmp DI,buf_length     {Full?}
  jl @wc1               {no}
  push AX               {Save registers}
  push CX
  mov AH,40h
  mov BX,output_handle
  mov CX,DI
  lds DX,ptr_output
  int 21h               {Write buffer to file}
  pop CX
  pop AX
  mov output_offset,0   {Point to beginning of buffer}
  xor DI,DI
 @wc1:
  mov SI,DI
  les DI,ptr_output
  add DI,SI             {Point into buffer}
  stosb                 {Store CHAR}
  inc output_offset     {Increment number of chars in buffer}
  retn

 @add_code:
  mov BX,free_code      {Get new code}
  les DI,ptr_hash
  mov SI,BX
  shl SI,1
  add SI,BX             {SI=3*BX}
  add DI,SI
  mov AX,old_code       {get prefix code}
  stosw                 {save it}
  mov AL,k              {get suffix CHAR}
  stosb                 {save it}
  inc free_code         {set next code}
  retn

 @fin:
  pop DS
 end;

 Dispose(output_data); Dispose(input_data); Dispose(hash_table);

 fin.Ferme; fout.Ferme;
end; {DecompLZf}

{----------------------------------------------------------}

procedure CompLZmem(var source; taille_source: WORD; var destination; var taille_dest: WORD);
{compacte mmoire source -> mmoire destination}
{si mmoire insuffisante, taille_dest=0}

type hash_rec=record
               first: WORD; {First entry with this value}
               next : WORD; {Next entry along chain}
               CHAR : CHAR; {Suffix CHAR}
              end;

     hash=array[1..maxmax] of hash_rec;

var hash_table: ^hash;
    ptr_hash: POINTER;
    input_offset,output_offset,bit_offset: WORD;
    prefix_code,free_code,max_code,nbits: WORD;
    first_use: BOOLEAN;
    k: CHAR;

begin
 if MemAvail<sizeof(hash)+64 then begin taille_dest:=0; exit end;

 New(hash_table);  ptr_hash:=hash_table;
 input_offset:=0; output_offset:=0; bit_offset:=0;

 asm
  push DS
  cld

  call @init_table      {Initialize the table and some vars}
  mov AX,clear          {Write a clear code}
  call @write_code
  call @read_char       {Read first CHAR}
 @l4:
  xor AH,AH             {Turn CHAR into code}
 @l4a:
  mov prefix_code,AX    {set prefix code}
  call @read_char       {Read next CHAR}
  jc @l17               {Carry means eof}
  mov k,AL              {Save CHAR in k}
  mov BX,prefix_code    {Get prefix code}
  call @lookup_code     {See if this pair in table}
  jnc @l4a              {nc means yes, new code in AX}
  call @add_code        {Add pair to table}
  mov AX,prefix_code    {Write old prefix code}
  call @write_code
  mov AL,k              {Get last CHAR}
  cmp BX,max_code       {Exceed code size?}
  jl @l4                {less means no}
  cmp nbits,12          {Currently less than 12 bits?}
  jl @l14               {yes}
  mov AX,clear          {Write a clear code}
  call @write_code
  call @init_table      {Reinit table}
  mov AL,k              {get last CHAR}
  jmp @l4               {Start over}
 @l14:
  inc nbits             {Increase number of bits}
  shl max_code,1        {Double max code size}
  jmp @l4               {Get next CHAR}
 @l17:
  mov AX,prefix_code    {Write last code}
  call @write_code
  mov AX,eof            {Write eof code}
  call @write_code
  mov AX,bit_offset
  mov DX,AX             {convert bits to bytes}
  and DX,7
  jz @l18               {test if extra bits}
  inc output_offset     {one more BYTE}
 @l18:
  jmp @fin              {done}

 @init_table:
  mov nbits,9           {set code size to 9}
  mov max_code,512      {set max code to 512=2^nbits}
  mov AX,-1             {Unused flag}
  mov CX,256*5/2        {Clear first 256 entries}
  les DI,ptr_hash       {Point to first entry in hash table}
  rep stosw             {Clear it out}
  mov free_code,first_free    {set next code to use}
  retn

 @write_code:
  push AX               {Save code}
  mov AX,bit_offset     {Get bit offset}
  mov CX,nbits          {Adjust bit offset by code size}
  add bit_offset,CX
  mov DX,AX             {Convert bit offset to BYTE offset}
  and DX,7
  shr AX,3
  les DI,destination
  add DI,AX             {Point into buffer}
  mov output_offset,AX
  pop AX                {Restore code}
  mov CX,DX             {offset within BYTE}
  xor DX,DX             {DX will catch bits rotated out}
  jcxz @wc3             {if offset in BYTE is zero, skip shift}
 @wc2:
  shl AX,1              {Rotate code}
  rcl DX,1
  loop @wc2
  or AL,ES:[DI]         {Grab bits currently in buffer}
 @wc3:
  stosw                 {Save data}
  mov AL,DL             {Grab extra bits}
  stosb                 {and save}
  retn

 @read_char:
  mov DI,input_offset
  cmp DI,taille_source
  jae @rd2              {end}
  lds SI,source
  add SI,DI             {Point at character}
  lodsb                 {Read it in}
  inc input_offset      {Adjust POINTER}
  clc                   {Success}
  retn
 @rd2:
  stc                   {Nothing left}
  retn

 @lookup_code:
  lds SI,ptr_hash       {point to hash table}
  mov DI,BX
  shl DI,2
  add DI,BX             {DI=5*BX}
  add SI,DI             {convert code to address}
  mov first_use,TRUE    {flag}
  cmp WORD ptr [SI],-1  {(hash.first) Has this code been used?}
  je @gc4               {equal means no}
  mov first_use,FALSE   {set flag}
  mov BX,[SI]           {(hash.first) Get first entry}
 @gc2:
  mov SI,BX
  shl SI,2
  add SI,BX             {convert code to address: SI=5*BX}
  cmp [SI+4],AL         {(hash.CHAR) is CHAR the same?}
  jne @gc3              {ne means no}
  clc                   {success}
  mov AX,BX             {put found code in AX}
  retn
 @gc3:
  cmp WORD ptr [SI+2],-1   {(hash.next) More left with this prefix?}
  je @gc4                  {equal means no}
  mov BX,[SI+2]         {(hash.next) get next code}
  jmp @gc2              {try again}
 @gc4:
  stc                   {not found}
  retn

 @add_code:
  mov BX,free_code      {Get code to use}
  mov DI,SI
  lds SI,ptr_hash
  add SI,DI
  cmp first_use,TRUE    {First use of this prefix?}
  je @ac1               {equal means yes}
  mov [SI+2],BX         {(hash.next) point last use to new entry}
  jmp @ac2
 @ac1:
  mov [SI],BX           {(hash.first) Point first use to new entry}
 @ac2:
  cmp BX,maxmax         {Have we reached code limit?}
  je @ac3               {equal means yes, just return}
  mov SI,BX
  shl SI,2
  add SI,BX             {convert code to address: SI=5*BX}
  mov WORD ptr [SI],-1    {(hash.first) initialize pointers}
  mov WORD ptr [SI+2],-1  {(hash.next)}
  mov BYTE ptr [SI+4],AL  {(hash.CHAR) save suffix CHAR}
  inc free_code         {adjust next code}
 @ac3:
  retn

 @fin:
  pop DS
 end;

 Dispose(hash_table); taille_dest:=output_offset+1;
end; {CompLZmem}

{----------------------------------------------------------}

function DecompLZmem(var source; var destination): WORD;
{dcompacte mmoire source -> mmoire destination}
{renvoie la taille initiale (non compacte), ou 0 si mmoire insuffisante}
{Remarque: la taille compacte est inutile car l'algorithme s'arrte
           lorsqu'il rencontre le code 'eof'}

type hash_rec=record
               next: WORD; {Prefix code}
               CHAR: CHAR; {Suffix CHAR}
              end;

     hash=array[1..maxmax] of hash_rec;

var hash_table: ^hash;
    ptr_hash: POINTER;
    output_offset,bit_offset: WORD;
    free_code,max_code,nbits: WORD;
    cur_code,old_code,in_code,stack_count: WORD;
    fin_char,k: CHAR;

begin
 if MemAvail<sizeof(hash)+64 then begin DecompLZmem:=0; exit end;

 New(hash_table);  ptr_hash:=hash_table;
 output_offset:=0; bit_offset:=0; stack_count:=0;

 asm
  push DS
  cld

  call @init_table      {Initialize variables}
 @l1:
  call @read_code       {Get a code}
  cmp AX,eof            {end of file?}
  jne @l2               {no}
  jmp @fin              {done}
 @l2:
  cmp AX,clear          {Clear code?}
  jne @l7               {no}
  call @init_table      {Initialize table}
  call @read_code       {Read next code}
  mov cur_code,AX       {Initialize variables}
  mov old_code,AX
  mov k,AL
  mov fin_char,AL
  call @write_char      {Write character}
  jmp @l1               {Get next code}
 @l7:
  mov cur_code,AX       {Save new code}
  mov in_code,AX
  cmp AX,free_code      {Code in table? (k<w>k<w>k)}
  jl @l11               {yes}
  mov AX,old_code       {get previous code}
  mov cur_code,AX       {make current}
  mov AL,fin_char       {get old last CHAR}
  push AX               {push it}
  inc stack_count
 @l11:
  cmp cur_code,255      {Code or character?}
  jle @l15              {CHAR}
  lds SI,ptr_hash
  mov BX,cur_code       {Convert code to address}
  mov DI,BX
  shl DI,1
  add DI,BX             {DI=3*BX}
  add SI,DI
  lodsw                 {Get prefix code}
  mov cur_code,AX       {Save it}
  lodsb                 {Get suffix CHAR}
  push AX               {Push it}
  inc stack_count
  jmp @l11              {Translate again}
 @l15:
  mov AX,cur_code       {Get code}
  mov fin_char,AL       {Save as final, k}
  mov k,AL
  push AX               {Push it}
  inc stack_count
  mov CX,stack_count    {Pop stack}
  jcxz @l18             {if anything there}
 @l17:
  pop AX
  call @write_char
  loop @l17
 @l18:
  mov stack_count,0     {Clear count on stack}
  call @add_code        {Add new code to table}
  mov AX,in_code        {Save input code}
  mov old_code,AX
  mov BX,free_code      {Hit table limit?}
  cmp BX,max_code
  jl @l23               {Less means no}
  cmp nbits,12          {Still within twelve bits?}
  je @l23               {no (next code should be clear)}
  inc nbits             {Increase code size}
  shl max_code,1        {Double max code}
 @l23:
  jmp @l1               {Get next code}

 @read_code:
  mov AX,bit_offset     {Get bit offset}
  add AX,nbits          {Adjust by code size}
  xchg bit_offset,AX    {Swap}
  mov DX,AX             {Calculate BYTE offset}
  and DX,7
  shr AX,3
  lds SI,source
  add SI,AX             {Point to CHAR}
  lodsw                 {Get WORD}
  mov BX,AX             {Save in AX}
  lodsb                 {Next BYTE}
  mov CX,DX             {Offset in BYTE}
  jcxz @rd2             {if zero, skip shifts}
 @rd1:
  shr AL,1              {Put code in low (code size) bits of BX}
  rcr BX,1
  loop @rd1
 @rd2:
  mov AX,BX             {put code in AX}
  mov CX,nbits
  sub CL,9
  mov BH,2
  shl BH,CL
  dec BH                {BH=(2 shl (nbits-9))-1}
  and AH,BH             {mask off unwanted high bits of AX -> keep nbits}
  retn

 @init_table:
  mov nbits,9           {Initialize variables}
  mov max_code,512
  mov free_code,first_free
  retn

 @write_char:
  les DI,destination
  add DI,output_offset  {Point into buffer}
  stosb                 {Store CHAR}
  inc output_offset     {Increment number of chars in buffer}
  retn

 @add_code:
  mov BX,free_code      {Get new code}
  les DI,ptr_hash
  mov SI,BX
  shl SI,1
  add SI,BX             {SI=3*BX}
  add DI,SI
  mov AX,old_code       {get prefix code}
  stosw                 {save it}
  mov AL,k              {get suffix CHAR}
  stosb                 {save it}
  inc free_code         {set next code}
  retn

 @fin:
  pop DS
 end;

 Dispose(hash_table); DecompLZmem:=output_offset;
end; {DecompLZmem}

{==========================================================}

END.
