{ͻ}
{                       --== MEMLIN.PAS ==--                       }
{                                                                      }
{ Unit pour grer la mmoire DOS avec adressage linaire sur 20 bits  }
{ avec allocation/libration automatiques                              }
{                                                                      }
{   IMPORTANT: Les allocations se font avec le DOS, donc le programme  }
{              doit limiter le tas au strict ncessaire !!             }
{                                                                      }
{ AUTEUR: Fabrice Couteau                                              }
{Ķ}
{ Compatibilit: 386           Dbut d'criture 28/05/95 }
{                              Dernires modifications 26/07/98 }
{ͼ}

UNIT MemLin;

{$A+,G+,Q-,R-,S-}

{$DEFINE TestAddr}    {pour tester les dbordements d'adresses}

INTERFACE

{==========================================================}

function  MemDOSLibre: LONGINT;
function  NbrPagesDisp: WORD;
function  NbrPagesNec(taille: LONGINT): WORD;

function  LitOctet(adresse: LONGINT): BYTE;
procedure EcritOctet(adresse: LONGINT; octet: BYTE);

procedure LitBloc(adresse_debut: LONGINT; taille: WORD; var buffer);
procedure EcritBloc(adresse_debut: LONGINT; taille: WORD; var buffer);

procedure RemplitBloc(adresse_debut,nombre: LONGINT; octet: BYTE);

procedure CopieBloc(adresse_source,adresse_destination,taille: LONGINT);

{==========================================================}

IMPLEMENTATION

CONST nbr_pages_max=40; {nombre de pages de 16ko}
      adresse_max=nbr_pages_max*16384-1;

      {codes d'erreur}
      err_debordement=1;
      err_mem_insuff =2;
      err_liberation =3;

VAR   TablePages: array[0..nbr_pages_max-1] of
                   record
                    segment_page: WORD; {0 si page non alloue}

                    {rserv pour usage futur (EMS...)}
                   end;

      anc_ExitProc: POINTER;

{==========================================================}

procedure AfficheErreur(code_erreur: BYTE);
begin
 asm
  mov AX,03h
  int 10h
 end;
 Write('MEMLIN: ');
 case code_erreur of
  err_debordement: WriteLn('Dbordement d''adresse');
  err_mem_insuff : WriteLn('Mmoire insuffisante');
  err_liberation : WriteLn('Libration mmoire a chou');
 end;
 halt(code_erreur);
end; {AfficheErreur}

{----------------------------------------------------------}

function MemDOSLibre: LONGINT; assembler;
{renvoie le nombre d'octets libres en mmoire DOS}
asm
 mov AH,48h
 mov BX,65535
 int 21h       {BX=nombre de paragraphes libres}
 rol BX,4
 mov DX,BX
 and DX,0000000000001111b
 mov AX,BX
 and AX,1111111111110000b
end; {MemDOSLibre}

{----------------------------------------------------------}

function NbrPagesDisp: WORD;
{renvoie le nombre total de pages disponibles}
begin
 NbrPagesDisp:=MemDOSLibre div (16384+16); {une page de 16ko + 16 octets MCB}
end; {NbrPagesDisp}

{----------------------------------------------------------}

function NbrPagesNec(taille: LONGINT): WORD;
{renvoie le nombre de pages ncessaires pour crire 'taille' octets
 -> VERIFIER AVANT TOUT ACCES EN ECRITURE QUE NbrPagesNec<=NbrPagesDisp <-}
begin
 if taille=0 then NbrPagesNec:=0
             else NbrPagesNec:=2+((taille-1) div 16384); {rserve une marge d'une page en cas de recouvrement}
end; {NbrPagesNec}

{----------------------------------------------------------}

procedure InitTablePages;
var i: WORD;
begin
 for i:=0 to nbr_pages_max-1 do
  TablePages[i].segment_page:=0; {page non alloue par dfaut}
 if NbrPagesDisp=0 then AfficheErreur(err_mem_insuff);
end; {InitTablePages}

{----------------------------------------------------------}

procedure LiberePages;
var i,segment: WORD;
    erreur: BOOLEAN;
begin
 erreur:=FALSE;
 for i:=0 to nbr_pages_max-1 do with TablePages[i] do begin
  segment:=segment_page; segment_page:=0;
  if segment<>0 then begin
   asm
    mov AH,49h
    mov ES,segment
    int 21h        {libre la page}
    jnc @ok
    mov erreur,TRUE
   @ok:
   end;
   if erreur then AfficheErreur(err_liberation);
  end;
 end;
end; {LiberePages}

{----------------------------------------------------------}

function AllouePage16ko: WORD; assembler;
{si mmoire suffisante, renvoie le segment de la page
 sinon renvoie 0}
asm
 mov AH,48h
 mov BX,16384/16  {nombre de paragraphes}
 int 21h          {alloue 16ko}
 jnc @ok
 xor AX,AX        {mmoire insuffisante}
@ok:
end; {AllouePage16ko}

{----------------------------------------------------------}

function LitOctet(adresse: LONGINT): BYTE;
var numero_page,offset_page: WORD;
begin
 {$IFDEF TestAddr}
 if (adresse<0) or (adresse>adresse_max) then AfficheErreur(err_debordement);
 {$ENDIF}
 asm
  les AX,adresse
  mov DX,ES
  mov BX,AX
  and AX,0011111111111111b
  mov offset_page,AX      {offset_page=adresse mod 16384}
  shr BX,14
  shl DX,2
  or  BX,DX
  mov numero_page,BX      {numero_page=adresse div 16384}
 end;
 LitOctet:=Mem[TablePages[numero_page].segment_page:offset_page];
end; {LitOctet}

{----------------------------------------------------------}

procedure EcritOctet(adresse: LONGINT; octet: BYTE);
var numero_page,offset_page: WORD;
begin
 {$IFDEF TestAddr}
 if (adresse<0) or (adresse>adresse_max) then AfficheErreur(err_debordement);
 {$ENDIF}
 asm
  les AX,adresse
  mov DX,ES
  mov BX,AX
  and AX,0011111111111111b
  mov offset_page,AX      {offset_page=adresse mod 16384}
  shr BX,14
  shl DX,2
  or  BX,DX
  mov numero_page,BX      {numero_page=adresse div 16384}
 end;
 with TablePages[numero_page] do begin
  if segment_page=0 then segment_page:=AllouePage16ko;
  if segment_page=0 then AfficheErreur(err_mem_insuff);
  Mem[segment_page:offset_page]:=octet;
 end;
end; {EcritOctet}

{----------------------------------------------------------}

procedure LitBloc(adresse_debut: LONGINT; taille: WORD; var buffer);
var adresse_fin: LONGINT;
    numero_page_debut,numero_page_fin,
    offset_page_debut,offset_page_fin,
    segment_source,index_buffer,numero_page: WORD;
begin
 if taille=0 then exit;

 adresse_fin:=adresse_debut+taille-1;
 {$IFDEF TestAddr}
 if (adresse_debut<0) or (adresse_fin>adresse_max) then AfficheErreur(err_debordement);
 {$ENDIF}
 asm
  les AX,adresse_debut
  mov DX,ES
  mov BX,AX
  and AX,0011111111111111b
  mov offset_page_debut,AX
  shr BX,14
  shl DX,2
  or  BX,DX
  mov numero_page_debut,BX

  les AX,adresse_fin
  mov DX,ES
  mov BX,AX
  and AX,0011111111111111b
  mov offset_page_fin,AX
  shr BX,14
  shl DX,2
  or  BX,DX
  mov numero_page_fin,BX
 end;

 segment_source:=TablePages[numero_page_debut].segment_page;

 if numero_page_fin=numero_page_debut then
  asm
   push DS
   cld
   mov DS,segment_source
   mov SI,offset_page_debut
   les DI,buffer
   mov CX,taille
   shr CX,1
   jnc @pair
   movsb
  @pair:
   rep movsw
   pop DS
  end

 else begin  {numero_page_fin>numero_page_debut}

  asm
   push DS
   cld
   mov DS,segment_source
   mov SI,offset_page_debut
   les DI,buffer
   mov CX,16384
   sub CX,offset_page_debut
   mov index_buffer,CX  {pour la boucle for...}
   shr CX,1
   jnc @pair
   movsb
  @pair:
   rep movsw
   pop DS
  end;
  for numero_page:=numero_page_debut+1 to numero_page_fin-1 do
  {boucle non excute si numero_page_fin=numero_page_debut+1}
   begin
    segment_source:=TablePages[numero_page].segment_page;
    asm
     push DS
     cld
     mov DS,segment_source
     xor SI,SI
     les DI,buffer
     add DI,index_buffer
     mov CX,16384/4
     DB 66h; rep movsw  {rep movsd}
     pop DS
    end;
    inc(index_buffer,16384);
   end;
  segment_source:=TablePages[numero_page_fin].segment_page;
  asm
   push DS
   cld
   mov DS,segment_source
   xor SI,SI
   les DI,buffer
   add DI,index_buffer
   mov CX,offset_page_fin
   inc CX
   shr CX,1
   jnc @pair
   movsb
  @pair:
   rep movsw
   pop DS
  end;
 end;
end; {LitBloc}

{----------------------------------------------------------}

procedure EcritBloc(adresse_debut: LONGINT; taille: WORD; var buffer);
var adresse_fin: LONGINT;
    numero_page_debut,numero_page_fin,
    offset_page_debut,offset_page_fin,
    segment_destination,index_buffer,numero_page: WORD;
begin
 if taille=0 then exit;

 adresse_fin:=adresse_debut+taille-1;
 {$IFDEF TestAddr}
 if (adresse_debut<0) or (adresse_fin>adresse_max) then AfficheErreur(err_debordement);
 {$ENDIF}
 asm
  les AX,adresse_debut
  mov DX,ES
  mov BX,AX
  and AX,0011111111111111b
  mov offset_page_debut,AX
  shr BX,14
  shl DX,2
  or  BX,DX
  mov numero_page_debut,BX

  les AX,adresse_fin
  mov DX,ES
  mov BX,AX
  and AX,0011111111111111b
  mov offset_page_fin,AX
  shr BX,14
  shl DX,2
  or  BX,DX
  mov numero_page_fin,BX
 end;

 with TablePages[numero_page_debut] do begin
  if segment_page=0 then segment_page:=AllouePage16ko;
  if segment_page=0 then AfficheErreur(err_mem_insuff);
  segment_destination:=segment_page;
 end;

 if numero_page_fin=numero_page_debut then
  asm
   push DS
   cld
   mov ES,segment_destination
   mov DI,offset_page_debut
   lds SI,buffer
   mov CX,taille
   shr CX,1
   jnc @pair
   movsb
  @pair:
   rep movsw
   pop DS
  end

 else begin {numero_page_fin>numero_page_debut}

  asm
   push DS
   cld
   mov ES,segment_destination
   mov DI,offset_page_debut
   lds SI,buffer
   mov CX,16384
   sub CX,offset_page_debut
   mov index_buffer,CX  {pour la boucle for...}
   shr CX,1
   jnc @pair
   movsb
  @pair:
   rep movsw
   pop DS
  end;
  for numero_page:=numero_page_debut+1 to numero_page_fin-1 do
  {boucle non excute si numero_page_fin=numero_page_debut+1}
   begin
    with TablePages[numero_page] do
     begin
      if segment_page=0 then segment_page:=AllouePage16ko;
      if segment_page=0 then AfficheErreur(err_mem_insuff);
      segment_destination:=segment_page;
     end;
    asm
     push DS
     cld
     mov ES,segment_destination
     xor DI,DI
     lds SI,buffer
     add SI,index_buffer
     mov CX,16384/4
     DB 66h; rep movsw   {rep movsd}
     pop DS
    end;
    inc(index_buffer,16384);
   end;
  with TablePages[numero_page_fin] do
   begin
    if segment_page=0 then segment_page:=AllouePage16ko;
    if segment_page=0 then AfficheErreur(err_mem_insuff);
    segment_destination:=segment_page;
   end;
  asm
   push DS
   cld
   mov ES,segment_destination
   xor DI,DI
   lds SI,buffer
   add SI,index_buffer
   mov CX,offset_page_fin
   inc CX
   shr CX,1
   jnc @pair
   movsb
  @pair:
   rep movsw
   pop DS
  end;
 end;
end; {EcritBloc}

{----------------------------------------------------------}

procedure RemplitBloc(adresse_debut,nombre: LONGINT; octet: BYTE);
var adresse_fin: LONGINT;
    numero_page_debut,numero_page_fin,
    offset_page_debut,offset_page_fin,
    segment_destination,numero_page: WORD;
begin
 if nombre=0 then exit;

 adresse_fin:=adresse_debut+nombre-1;
 {$IFDEF TestAddr}
 if (adresse_debut<0) or (adresse_fin>adresse_max) then AfficheErreur(err_debordement);
 {$ENDIF}
 asm
  les AX,adresse_debut
  mov DX,ES
  mov BX,AX
  and AX,0011111111111111b
  mov offset_page_debut,AX
  shr BX,14
  shl DX,2
  or  BX,DX
  mov numero_page_debut,BX

  les AX,adresse_fin
  mov DX,ES
  mov BX,AX
  and AX,0011111111111111b
  mov offset_page_fin,AX
  shr BX,14
  shl DX,2
  or  BX,DX
  mov numero_page_fin,BX
 end;

 with TablePages[numero_page_debut] do begin
  if segment_page=0 then segment_page:=AllouePage16ko;
  if segment_page=0 then AfficheErreur(err_mem_insuff);
  segment_destination:=segment_page;
 end;

 if numero_page_fin=numero_page_debut then
  asm
   cld
   mov ES,segment_destination
   mov DI,offset_page_debut
   mov AL,octet
   mov AH,AL
   mov CX,WORD(nombre)
   shr CX,1
   jnc @pair
   stosb
  @pair:
   rep stosw
  end

 else begin  {numero_page_fin>numero_page_debut}

  asm
   cld
   mov ES,segment_destination
   mov DI,offset_page_debut
   mov AL,octet
   mov AH,AL
   mov CX,16384
   sub CX,offset_page_debut
   shr CX,1
   jnc @pair
   stosb
  @pair:
   rep stosw
  end;
  for numero_page:=numero_page_debut+1 to numero_page_fin-1 do
  {boucle non excute si numero_page_fin=numero_page_debut+1}
   begin
    with TablePages[numero_page] do
     begin
      if segment_page=0 then segment_page:=AllouePage16ko;
      if segment_page=0 then AfficheErreur(err_mem_insuff);
      segment_destination:=segment_page;
     end;
    asm
     cld
     mov ES,segment_destination
     xor DI,DI
     mov AL,octet
     mov AH,AL
     mov CX,16384/4
     DB 66h; rep stosw   {rep stosd}
    end;
   end;
  with TablePages[numero_page_fin] do
   begin
    if segment_page=0 then segment_page:=AllouePage16ko;
    if segment_page=0 then AfficheErreur(err_mem_insuff);
    segment_destination:=segment_page;
   end;
  asm
   cld
   mov ES,segment_destination
   xor DI,DI
   mov AL,octet
   mov AH,AL
   mov CX,offset_page_fin
   inc CX
   shr CX,1
   jnc @pair
   stosb
  @pair:
   rep stosw
  end;
 end;
end; {RemplitBloc}

{----------------------------------------------------------}

procedure CopieBloc(adresse_source,adresse_destination,taille: LONGINT);
var numero_page_source,numero_page_destination,
    segment_source,segment_destination,taille_a_copier: WORD;
    offset_page_source,offset_page_destination: INTEGER; {peuvent tre <0 !}
    taille_copiee: LONGINT;
begin
 if (taille=0) or (adresse_source=adresse_destination) then exit;

 {$IFDEF TestAddr}
 if (adresse_source<0) or
    (adresse_destination<0) or
    (adresse_source+taille-1>adresse_max) or
    (adresse_destination+taille-1>adresse_max) then AfficheErreur(err_debordement);
 {$ENDIF}

 if (adresse_destination<adresse_source) or
    (adresse_destination>=adresse_source+taille) then begin
 {pas de recouvrement, ou recouvrement dbut source/fin dest. -> copie en sens croissant}
  asm
   les AX,adresse_source
   mov DX,ES
   mov BX,AX
   and AX,0011111111111111b
   mov offset_page_source,AX
   shr BX,14
   shl DX,2
   or  BX,DX
   mov numero_page_source,BX

   les AX,adresse_destination
   mov DX,ES
   mov BX,AX
   and AX,0011111111111111b
   mov offset_page_destination,AX
   shr BX,14
   shl DX,2
   or  BX,DX
   mov numero_page_destination,BX
  end;

  taille_copiee:=0;
  repeat
   segment_source:=TablePages[numero_page_source].segment_page;
   with TablePages[numero_page_destination] do begin
    if segment_page=0 then segment_page:=AllouePage16ko;
    if segment_page=0 then AfficheErreur(err_mem_insuff);
    segment_destination:=segment_page;
   end;
   if offset_page_source>=offset_page_destination
    then taille_a_copier:=16384-offset_page_source
    else taille_a_copier:=16384-offset_page_destination;
   if taille_copiee+taille_a_copier>taille then taille_a_copier:=taille-taille_copiee;
   asm
    push DS
    cld
    mov DS,segment_source
    mov SI,offset_page_source
    mov ES,segment_destination
    mov DI,offset_page_destination
    mov CX,taille_a_copier
    shr CX,1
    jnc @pair
    movsb
   @pair:
    rep movsw
    pop DS
   end;
   inc(taille_copiee,taille_a_copier);
   inc(offset_page_source,taille_a_copier);
   if offset_page_source>=16384 then
    begin inc(numero_page_source); offset_page_source:=0 end;
   inc(offset_page_destination,taille_a_copier);
   if offset_page_destination>=16384 then
    begin inc(numero_page_destination); offset_page_destination:=0 end;
  until taille_copiee=taille;
 end

 else begin  {recouvrement fin source/dbut dest. -> copie en sens dcroissant}

  inc(adresse_source,taille-1);
  inc(adresse_destination,taille-1); {commence par la fin}

  asm
   les AX,adresse_source
   mov DX,ES
   mov BX,AX
   and AX,0011111111111111b
   mov offset_page_source,AX
   shr BX,14
   shl DX,2
   or  BX,DX
   mov numero_page_source,BX

   les AX,adresse_destination
   mov DX,ES
   mov BX,AX
   and AX,0011111111111111b
   mov offset_page_destination,AX
   shr BX,14
   shl DX,2
   or  BX,DX
   mov numero_page_destination,BX
  end;

  taille_copiee:=0;
  repeat
   segment_source:=TablePages[numero_page_source].segment_page;
   with TablePages[numero_page_destination] do begin
    if segment_page=0 then segment_page:=AllouePage16ko;
    if segment_page=0 then AfficheErreur(err_mem_insuff);
    segment_destination:=segment_page;
   end;
   if offset_page_source>=offset_page_destination
    then taille_a_copier:=offset_page_destination+1
    else taille_a_copier:=offset_page_source+1;
   if taille_copiee+taille_a_copier>taille then taille_a_copier:=taille-taille_copiee;
   asm
    push DS
    std
    mov DS,segment_source
    mov SI,offset_page_source
    mov ES,segment_destination
    mov DI,offset_page_destination
    mov CX,taille_a_copier
    rep movsb  {!! ne pas utiliser movsw car sens dcroissant !!}
   @fin:
    pop DS
   end;
   inc(taille_copiee,taille_a_copier);
   dec(offset_page_source,taille_a_copier);
   if offset_page_source<0 then
    begin dec(numero_page_source); offset_page_source:=16384-1 end;
   dec(offset_page_destination,taille_a_copier);
   if offset_page_destination<0 then
    begin dec(numero_page_destination); offset_page_destination:=16384-1 end;
  until taille_copiee=taille;
 end;
end; {CopieBloc}

{----------------------------------------------------------}

{$F+}
procedure ExitMemLin;
begin
 ExitProc:=anc_ExitProc;
 LiberePages;
end; {ExitMemLin}
{$F-}

{==========================================================}

{initialisation de l'Unit}
begin
 InitTablePages;
 anc_ExitProc:=ExitProc;
 ExitProc:=@ExitMemLin; {installe routine de sortie}
END.
