{************************************************************************}
{                               MEMOIRE.PAS                              }
{                                                                        }
{  Gestion de la mmoire (DOS, UMB et EMS)                               }
{                                                                        }
{  AUTEUR: Freddy Vtel.                                                }
{                                                                        }
{  Rq: Les allocations en mmoire DOS et UMB sont en paragraphes !!!     }
{                                                                        }
{                               Dbut d'criture le 27/03/95             }
{                               Dernires modifications le 26/07/98      }
{************************************************************************}

UNIT Memoire;
{$A+,G+,Q-,R-,S-}

INTERFACE

USES Crt,VarUnit;

CONST {Numros des fonctions du DOS}
      GET_MEM     =$48;    { Rserve de la mmoire vive              }
      FREE_MEM    =$49;    { Libre de la mmoire vive               }
      CHANGE_MEM  =$4A;    { Modifie la taille d'une zone de mmoire }
      GET_STRATEGY=$5800;  { Lit la stratgie d'allocation           }
      SET_STRATEGY=$5801;  { Fixe la stratgie d'allocation          }
      GET_UMB     =$5802;  { Lit l'tat d'inclusion des blocs UMB    }
      SET_UMB     =$5803;  { Fixe l'inclusion des blocs UMB          }

      {Stratgies de recherche de mmoire}
      CHERCHE_EN_BAS  =$00; { premier bloc de mmoire libre          }
      CHERCHE_MEILLEUR=$01; { meilleur bloc de mmoire libre         }
      CHERCHE_EN_HAUT =$02; { dernier bloc de mmoire libre          }
      DABORD_UMB      =$80; { Chercher dans la zone UMB              }

      EMS_INT  =$67;        { N d'interruption pour l'accs  l'EMM }
      EMS_ERR  =WORD(-1);   { Est renvoy en cas d'erreur }

      MCB_DETRUIT     =7;
      PAS_ASSEZ_DE_MEM=8;

{      MAX_Handle_EMS=50;
      MAX_Bloc_DOS=256;
      MAX_Bloc_UMB=256;}

      MEMStatus: BYTE=0;   { code d'erreur mmoire DOS et UMB (0 si OK)}
      EMSStatus: BYTE=0;   { code d'erreur EMM (0 si OK)}

      Segment_EMS: WORD=$0000;    { segment du Page Frame EMS }
      Nb_pages_EMS_Libres: WORD=0;

      Memoire_Initialisee: BOOLEAN=FALSE;

{==========================================================}

{**** Mmoire DOS ****}
function  MemoireDOSLibre: WORD;
function  ReserveDOS(var segment: WORD; Taille: WORD): WORD;
function  ModifieDOS(segment: WORD; Taille: WORD): WORD;
procedure LibereDOS(segment: WORD);
procedure FixeRecherche(strategie: BYTE);

{**** Mmoire UMB ****}
function  ReserveUMB(var segment: WORD; Taille: WORD): WORD;
procedure LibereUMB(segment: WORD);

{****   UMB/DOS   ****}
function  ReserveUMBDOS(var segment: WORD; Taille: WORD): BOOLEAN;
procedure LibereUMBDOS(segment: WORD);

{**** Mmoire EMS ****}
function  EMSInst: BOOLEAN;
function  EMSVersion : BYTE;
function  PagesEMSLibres: WORD;
function  EMSFrameSeg: WORD;
function  ReserveEMS(Pages: WORD): WORD;
function  LibereEMS(handle: WORD): BOOLEAN;
function  EMSMap(handle,PLog: WORD; PPhys: BYTE): BOOLEAN;
function  SaveMapEMS(handle: WORD): BOOLEAN;

function  ReserveBloc(var Bloc: T_BlocMem; Taille: WORD; UseUMB,UseEMS: BOOLEAN): BOOLEAN;
function  AccedeBloc(Bloc: T_BlocMem):BOOLEAN; { Accder au bloc mmoire }
procedure LibereBloc(Bloc: T_BlocMem);

procedure DetecteMemoire(affiche: BOOLEAN);

{==========================================================}

IMPLEMENTATION

{==========================================================}

function MemoireDOSLibre: WORD; assembler;
asm
 mov AH,Get_Mem
 mov BX,0FFFFh             {Allouer toute la mmoire     }
 int 21h                   {=> le Dos nous donne la mmoire disponible}
 mov AX,BX                 {Mmoire libre en paragraphes}
end; {MemoireDOSLibre}

{----------------------------------------------------------}

function ReserveDOS(var segment: WORD; Taille: WORD): WORD;
var Taille_Libre: WORD;
begin
 asm
  mov MEMStatus,0
  mov BX,Taille
  mov AH,GET_MEM
  int 21H
  jnc @Ok
  mov MEMStatus,AL
  mov AX,0
 @Ok:
  les DI,segment
  mov ES:[DI],AX
  mov Taille_Libre,BX
 end;
 if MEMStatus=MCB_DETRUIT then ReserveDOS:=0
                          else ReserveDOS:=Taille_Libre;
end; {ReserveDOS}

{----------------------------------------------------------}

function ModifieDOS(segment: WORD; Taille: WORD): WORD; assembler;
asm
 mov MEMStatus,0
 mov AX,segment
 mov ES,AX             { ES <- segment de la zone   }
 mov BX,Taille         { Nouvelle taille de la zone }
 mov AH,CHANGE_MEM
 int 21h
 jnc @Ok
 mov MEMStatus,AL      { Code d'erreur }
@Ok:
 mov AX,BX             { Taille libre si mmoire insuffisante }
end; {ModifieDOS}

{----------------------------------------------------------}

procedure LibereDOS(segment: WORD); assembler;
{ Rq: La zone a un offset nul !!! }
asm
 mov MEMStatus,0
 mov AX,segment
 mov ES,AX         { ES <- Segment de la Zone }
 mov AH,FREE_MEM
 int 21h
 jnc @Ok
 mov MEMStatus,AL
@Ok:
end; {LibereDOS}

{----------------------------------------------------------}

procedure FixeRecherche(strategie: BYTE); assembler;
asm
 mov AX,SET_STRATEGY
 xor BH,BH
 mov BL,strategie
 int 21h
end; {FixeRecherche}

{----------------------------------------------------------}

function LierUMB(Lier: BOOLEAN): BOOLEAN; assembler;
asm
 mov AX,GET_UMB
 int 21h
 mov AH,0
 jc @Fin
 mov AH,AL
 push AX
 mov AX,SET_UMB
 xor BH,BH
 mov BL,Lier
 int 21h
 pop AX
@Fin:
 mov AL,AH
end; {LierUMB}

{----------------------------------------------------------}

function ReserveUMB(var segment: WORD; Taille: WORD): WORD;
var Memoire_Allouee: WORD;
begin
 LierUMB(TRUE);
 FixeRecherche(DABORD_UMB);
 Memoire_Allouee:=ReserveDOS(segment,Taille);
 if Memoire_Allouee=Taille then
  if segment<$A000 then begin { Zone alloue en mmoire DOS ! }
    Memoire_Allouee:=0;
    LibereDOS(segment);
    MEMStatus:=PAS_ASSEZ_DE_MEM;
  end;
 FixeRecherche(CHERCHE_EN_BAS);
 LierUMB(FALSE);
 ReserveUMB:=Memoire_Allouee
end;

{----------------------------------------------------------}

procedure LibereUMB(segment: WORD);
var SaveLink: BOOLEAN;
begin
 SaveLink:=LierUMB(TRUE);
 LibereDOS(segment);
 LierUMB(SaveLink)
end;  {LibereUMB}

{----------------------------------------------------------}

function ReserveUMBDOS(var segment: WORD; Taille: WORD): BOOLEAN;
var Resultat: BOOLEAN;
begin
 Resultat:=FALSE;

 if UMB_Presente  then Resultat:=(ReserveUMB(segment,Taille)=Taille);
 if not(Resultat) then Resultat:=(ReserveDOS(segment,Taille)=Taille);

 ReserveUMBDOS:=Resultat;
end;

{----------------------------------------------------------}

procedure LibereUMBDOS(segment: WORD);
begin
 if segment<>0 then
  if (segment<$A000) then LibereDOS(segment)
                     else LibereUMB(segment);
end;

{----------------------------------------------------------}

function EMSInst: BOOLEAN;
{renvoie TRUE si un EMM est install}
type Char8=array[1..8] of CHAR;
     Char8Ptr=^Char8;
const EMMid: Char8='EMMXXXX0';
begin
 EMSInst:=(Char8Ptr(Ptr(MemW[0:4*EMS_Int+2],10))^=EMMid)
end; {EMSInst}

{----------------------------------------------------------}

function EMSVersion: BYTE; assembler;
asm
 mov EMSStatus,0
 mov AH,46h           { Dterminer la version de l'EMS       }
 int EMS_INT
 cmp AH,0
 jz  @Ok
 mov EMSStatus,AH     { Erreur => Renvoyer le code d'erreur  }
 mov AX,EMS_ERR       { EMSVersion=EMS_ERR                   }
@Ok:
end;

{----------------------------------------------------------}

function PagesEMSLibres: WORD; assembler;
asm
 mov EMSStatus,0
 mov AH,42h           { Dterminer le nombre de pages        }
 int EMS_INT
 cmp AH,0
 jz  @Ok
 mov EMSStatus,AH     { Erreur => Renvoyer le code d'erreur  }
 mov AX,EMS_ERR
 jmp @Fin
@Ok:
 mov AX,DX            { Renvoyer le nb de pages libres       }
@Fin:
end; {PagesEMSLibres}

{----------------------------------------------------------}

function EMSFrameSeg: WORD; assembler;
asm
 mov EMSStatus,0
 mov AH,41h           { N fct.: Rech segment de Page Frame  }
 int EMS_INT
 cmp AH,0
 jz  @Ok
 mov EMSStatus,AH     { Erreur => Renvoyer le code d'erreur  }
 mov AX,EMS_ERR
 jmp @Fin
@Ok:
 mov AX,BX            { Renvoyer segment de Page Frame       }
@Fin:
end; {EMSFrameSeg}

{----------------------------------------------------------}

function ReserveEMS(Pages: WORD): WORD; assembler;
asm
 mov EMSStatus,0
 mov AH,43h           { N fct.: Allouer pages                  }
 mov BX,Pages         { Nb de pages  allouer                   }
 int EMS_INT          { Appeler EMM (ou QEMM ...)               }
 cmp AH,0
 jz  @Ok
 mov EMSStatus,AH     { Erreur => Renvoyer le code d'erreur     }
 mov AX,EMS_ERR
 jmp @Fin
@Ok:
 mov AX,DX            { Renvoyer le handle                      }
@Fin:
end; {ReserveEMS}

{----------------------------------------------------------}

function LibereEMS(handle: WORD): BOOLEAN; assembler;
asm
 mov AH,45h           { N fct.: Librer pages   }
 mov DX,handle        { Fixer handle EMS         }
 int EMS_INT          { Appeler EMM              }
 mov EMSStatus,AH     { Ranger le code d'erreur  }
 cmp AH,0
 je  @Ok
 xor AX,AX            { Erreur => FALSE          }
 jmp @Fin
@Ok:
 mov AX,1             { Ok => TRUE               }
@Fin:
end;  {LibereEMS}

{----------------------------------------------------------}

function EMSMap(handle,PLog: WORD; PPhys: BYTE): BOOLEAN; assembler;
asm
 mov AH,44h           { N fct.: Fixer Mapping   }
 mov AL,PPhys         { Fixer la page physique   }
 mov BX,PLog          { Fixer la page logique    }
 mov DX,handle        { Fixer handle EMS         }
 int EMS_INT          { Appeler EMM              }
 mov EMSStatus,AH     { Ranger le code d'erreur  }
 cmp AH,0
 je  @Ok
 xor AX,AX            { Erreur => FALSE          }
 jmp @Fin
@Ok:
 mov AX,1             { Ok => TRUE               }
@Fin:
end; {EMSMap}

{----------------------------------------------------------}

function SaveMapEMS(handle: WORD): BOOLEAN; assembler;
asm
 mov AH,47h           { N fct.: Sauver Mapping  }
 mov DX,handle        { Fixer handle EMS         }
 int EMS_INT          { Appeler EMM              }
 mov EMSStatus,AH     { Ranger le code d'erreur  }
 cmp AH,0
 je  @Ok
 xor AX,AX            { Erreur => FALSE          }
 jmp @Fin
@Ok:
 mov AX,1             { Ok => TRUE               }
@Fin:
end; {SaveMapEMS}

{---------------------------------------------------------}

function ReserveB_EMS(var Bloc: T_BlocMem; Taille: WORD): BOOLEAN;
var i: BYTE;
    Resultat: BOOLEAN;
begin
 Bloc.Segment:=0;
 Bloc.Nb_Pages_EMS:=(Taille div 1024)+1;
 Bloc.Handle_EMS:=ReserveEMS(Bloc.Nb_Pages_EMS);
 Resultat:=(Bloc.Handle_EMS<>EMS_ERR);

 if Resultat then begin
   for i:=0 to Bloc.Nb_Pages_EMS-1 do EMSMap(Bloc.Handle_EMS,i,i);
   Bloc.Segment:=Segment_EMS;
 end;

 ReserveB_EMS:=Resultat;
end;  { ReserveB_EMS }

{---------------------------------------------------------}

function ReserveB_UMB(var Bloc: T_BlocMem; Taille: WORD): BOOLEAN;
var segment: WORD;
    Resultat: BOOLEAN;
begin
 Resultat:=(ReserveUMB(segment,Taille)=Taille);

 if Resultat then Bloc.Segment:=segment
 else
  if MEMStatus=MCB_DETRUIT then begin
    Writeln('Bad UMB block! () ');
    repeat until ReadKey=#13;
  end;

 ReserveB_UMB:=Resultat;
end;  { ReserveB_UMB }

{---------------------------------------------------------}

function ReserveB_DOS(var Bloc: T_BlocMem; Taille: WORD): BOOLEAN;
var segment: WORD;
    Resultat: BOOLEAN;
begin
 Resultat:=(ReserveDOS(segment,Taille)=Taille);

 if Resultat then Bloc.Segment:=segment
 else
  if MEMStatus=MCB_DETRUIT then begin
    Writeln ('Bad DOS block! () ');
    repeat until ReadKey=#13;
  end;

 ReserveB_DOS:=Resultat;
end;  {ReserveB_DOS}

{---------------------------------------------------------}

function ReserveBloc(var Bloc: T_BlocMem; Taille: WORD; UseUMB,UseEMS: BOOLEAN): BOOLEAN;
{ Allouer de la mmoire dans les UMB, l'EMS ou la mmoire conventionelle }
{ La taille mmoire  allouer est en paragraphes (Blocs de 16 octets)    }

var Resultat: BOOLEAN;

begin
 Resultat:=FALSE;

 Bloc.Segment:=0;
 Bloc.Handle_EMS:=EMS_ERR;
 Bloc.Nb_Pages_EMS:=0;

 { priorits : EMS, UMB, DOS}

 { Allouer en EMS }
 if (UseEMS and EMS_Presente) then Resultat:=ReserveB_EMS(Bloc,Taille);

 { Allouer en memoire UMB }

 if (UseUMB and UMB_Presente) and not(Resultat) then Resultat:=ReserveB_UMB(Bloc,Taille);

 { Allouer en mmoire DOS }
 if not(Resultat) then Resultat:=ReserveB_DOS(Bloc,Taille);

 ReserveBloc:=Resultat;
end;  {ReserveBloc}

{----------------------------------------------------------}

function AccedeBloc(Bloc: T_BlocMem): BOOLEAN;
var i: BYTE;
begin
 if Bloc.Segment<>0 then begin
   if Bloc.Handle_EMS<>EMS_ERR then
    for i:=0 to Bloc.Nb_Pages_EMS-1 do EMSMap(Bloc.Handle_EMS,i,i);
   AccedeBloc:=TRUE;
 end
 else AccedeBloc:=FALSE;
end; {AccedeBloc}

{----------------------------------------------------------}

procedure LibereBloc(Bloc: T_BlocMem);
begin
 if Bloc.Handle_EMS<>EMS_ERR then LibereEMS(Bloc.Handle_EMS)
                             else LibereUMBDOS(Bloc.Segment);
end;

{----------------------------------------------------------}

procedure DetecteDOS(affiche: BOOLEAN);
begin
 if affiche then Write(MemoireDOSLibre shr 6,'k DOS. ');
end; {DetecteDOS}

{----------------------------------------------------------}

procedure DetecteUMB(affiche: BOOLEAN);
var segment: WORD;
begin
 UMB_Presente:=(ReserveUMB(segment,64)=64);
 if UMB_Presente then begin
   LibereUMB(segment);
   if affiche then Write('UMB found. ');
 end
 else
  if MEMStatus=MCB_DETRUIT then begin
    Write('Bad UMB block! () ');
    repeat until ReadKey=#13;
  end;
end; {DetecteUMB}

{----------------------------------------------------------}

procedure DetecteEMS(affiche: BOOLEAN);
begin
 EMS_Presente:=EMSInst;
 if EMS_Presente then begin
   Nb_pages_EMS_Libres:=PagesEMSLibres;
   if Nb_pages_EMS_Libres<>EMS_ERR then begin
     Segment_EMS:=EMSFrameSeg;
     if affiche then Write(Nb_pages_EMS_Libres*16,'k EMS. ');
   end;
 end;
end; {DetecteEMS}

{----------------------------------------------------------}

procedure DetecteMemoire(affiche: BOOLEAN);
begin
 if affiche then Write('Memory: ');
 DetecteDOS(affiche); DetecteUMB(affiche); DetecteEMS(affiche);
 if affiche then Writeln;
end; {DetecteMemoire}

{==========================================================}

END.
