{************************************************************************}
{                                MM_PLAY.PAS                             }
{                                                                        }
{  Gestion de Mod Master pendant l'coute (Affichages, Shell...)         }
{                                                                        }
{  AUTEURS: Freddy Vtel et Fabrice Couteau                             }
{                                                                        }
{                               Dbut d'criture le 30/08/95             }
{                               Dernires modifications le 26/07/98      }
{************************************************************************}

UNIT MM_PLAY;

{$A+,G+,Q-,R-,S-}

INTERFACE

{==========================================================}

procedure ModuleSuivant;
procedure GestionAffMus;

{==========================================================}

IMPLEMENTATION

USES Crt,Dos,Clavier,Texte,Util,SourisSm,
     Memoire,ModmVar,ModmUnit,ModmCmd,FFT_U,
     MM_VAR,MM_AFF,MM_DIV;

{$I Types}

CONST {types d'affichage de la fentre infrieure}
      aff_aide =1; {page d'aide}
      aff_instr=2; {liste des instruments}
      aff_voies=3; {liste des voies}
      aff_part =4; {partition}
      aff_clav =5; {claviers}
      aff_spect=6; {spectre}

      type_aff: BYTE=aff_instr; {type d'affichage courant}
      anc_type_aff: BYTE=aff_instr;

      {modes de raffichage des donnes (par ordre de priorit dcroissante)}
      reaff_tout  =4; {tout l'cran: fentres sup. & inf. + titre, ...}
      reaff_fen   =3; {fentre infrieure: fond+titre+contenu}
      reaff_titre =2; {fentre infrieure: titre seulement}
      reaff_lignes=1; {fentre infrieure: lignes (contenu)}
      reaff_rien  =0; {rien}

      reaffiche: BYTE=reaff_tout; {mode de raffichage de la fentre infrieure}

      {tats pour liste_instr.etat}
      smp_pasjoue =$00; {l'instrument n'a jamais t jou}
      smp_dejajoue=$01; {l'instrument a dj t jou, mais n'est pas actif}
      smp_actif   =$02; {l'instrument est actif sur la voie, sans tre jou}
      smp_joue    =$04; {l'instrument est en train d'tre jou sur la voie}
      smp_stoppe  =$08; {l'instrument est stopp}
      smp_charge  =$10; {l'instrument est charg (c'est une copie du flag Sample.Info initial)}


      {constantes pour FFT}
      log2_taille_tampon_FFT = log2_nbr_points_max;
      taille_tampon_FFT      = 1 shl log2_taille_tampon_FFT;
      intervalle_freq_FFT    = 50; {=1/T, en Hz}
      pas_frequence          = 10*intervalle_freq_FFT; {pas de frquence pour l'affichage de l'chelle}
      bande_spectre_FFT      = (taille_tampon_FFT*intervalle_freq_FFT) div 2; {largeur de bande en Hz}
      mode_FFT = FFT_directe+FFT_trapeze+FFT_module;

      {types de curseurs pour T_CursOsc}
      curs_dernier  =0; curs_aucun=curs_dernier;
      curs_vide     =1;
      curs_continu  =2;
      curs_pointille=3;


TYPE {type de curseur pour AfficheOscillo}
     T_CursOsc=record
                type_   : BYTE; {continu, pointille, vide, aucun/dernier}
                position: BYTE; {0  position  largeur_osc-1}
               end;


VAR {structure d'affichage pour aff_instr}
    liste_instr: record
                  mode_oscillo: BYTE; {oscillo son (canal de sortie) ou instrument}

                  nombre : BYTE; {nombre total d'instruments (chargs ou non)}
                  premier: BYTE; {n de l'instrument affich en haut}
                  courant: BYTE; {n de l'instrument en surbrillance}

                  etat: Max_Samples_Byte; {tat (charg,actif,...)}

                  zoom: BYTE; {facteur de zoom pour l'onde de l'instrument courant}
                  debut_onde: WORD; {offset dans l'onde chantillonne affiche  gauche}
                  echant_onde: BOOLEAN; {commande de r-chantillonnage de l'onde instrument}
                 end;

    {structure d'affichage pour aff_voies}
    liste_voies: record
                  nombre  : BYTE; {nombre total de voies}
                  courante: BYTE; {n de la voie en surbrillance}

                  type_barres: BYTE; {pitch mono, pitch stro, priode}
                  sample,
                  volume,
                  pitch,pitch_g,pitch_d: Max_Voies_Byte;
                  vitesse_pitch: BYTE; {vitesse des barres de pitch (dpend du tempo)}
                 end;

    {structure d'affichage pour aff_part}
    partition: record
                nombre_voies_vis: BYTE; {nombre de voies visibles}
                premiere_voie   : BYTE; {n de la voie affiche  gauche}
                voie_courante   : BYTE; {n de la voie en surbrillance}

                anc_pos_note: BYTE; {position note lors du dernier affichage}
                anc_pattern : BYTE; {n pattern    lors du dernier affichage}

                ligne_aff: array[0..8*8] of BYTE; {flags ligne dj affiche}
               end;

    {structure d'affichage pour aff_clav}
    claviers: record
               premiere_octave: BYTE; {n de l'octave affiche  gauche}

               nombre_instr : BYTE; {nombre d'instruments chargs}
               premier_instr: BYTE; {n de l'instrument affich en haut}
               instr_courant: BYTE; {n de l'instrument en surbrillance}

               num_instr: Max_Samples_Byte; {n des instruments chargs}
              end;

    {structure d'affichage pour aff_spect}
    spectre: record
              mode_spectre: BYTE; {mono/stro}
              premiere_frequence: WORD; {frquence affiche  gauche (en Hz)}

              amplification: BYTE; {coefficient multiplicateur avant FFT}
             end;

{==========================================================}

(*
procedure NewInt2F; external;
procedure Int2FOn;  external;
procedure Int2FOff; external;
{$L MM_DETEC}
*)

{----------------------------------------------------------}

{$F+}

procedure EchantTampon(var source,destination; long_source,long_dest: WORD); external;
{chantillonnage 8 bits: source (long_source) -> destination (long_dest)}

procedure EchantTamponStereo(var source,dest_gauche,dest_droite; long_source,long_dest: WORD); external;
{chantillonnage 8 bits: source (long_source) -> dest_gauche,dest_droite (long_dest chacun)
 !! ATTENTION !! dest_gauche et dest_droite doivent tre dans le mme segment !!
 format du tampon source: offset pair  =valeur gauche
                          offset impair=valeur droite}

procedure ConvSigne(var tampon; taille: WORD); external;
{conversion du tampon en 8 bits signs}

procedure Extension16(var source,destination; taille: WORD; amplification: BYTE); external;
{extension: source 8 bits signs -> destination 16 bits signs, avec amplification}

{$F-}

{$L Echant}

{----------------------------------------------------------}

procedure Shell;

var Mem_DOS_Libre_ko: WORD;
    anc_prompt,nouv_prompt: String80;

begin
 TextMode(LastMode);
 Curseur(plat);
 Mem_DOS_Libre_ko:=MemoireDOSLibre shr 6;

 if Mem_DOS_Libre_ko>=64 then begin
   Writeln('Free memory: ',Mem_DOS_Libre_ko-2,
           'k - type exit to return to Mod Master...',#13#10);
   anc_prompt:=GetEnv('PROMPT');
   nouv_prompt:='[Mod Master] ';
   if anc_prompt<>'' then nouv_prompt:=nouv_prompt+anc_prompt
                     else nouv_prompt:=nouv_prompt+'$p$g';

   if mode_interactif then FinSourisSmooth;

   SwapVectors;
{   Int2FOn;      { Active l'IRQ de dtection    }
   Exec(Getenv('COMSPEC'),'/K PROMPT '+nouv_prompt);
{   Int2FOff;     { Dsactive l'IRQ de dtection }
   SwapVectors;

   if mode_interactif then InitSourisSmooth(#166,#167,#169,#170);

   if DosError<>0 then begin
     if DosError=8 then Writeln('not enough memory!')
                   else WriteLn('DOS error #',DosError);
     Write('Press any key...');
     ReadKey;
   end
 end
 else begin
   Write('not enough memory (',Mem_DOS_Libre_ko,'k). Press any key...');
   ReadKey;
 end;

 Curseur(off);
end; {Shell}

{----------------------------------------------------------}

procedure ModuleSuivant;
{passe au module prcdent/suivant dans le programme courant actif}
begin
 with programme_courant do
  case actif of
   FALSE: lecture_module:=FALSE;

   TRUE : if (Numero_module+sens>=1) and
             (Numero_module+sens<=Nombre_modules) then inc(Numero_module,sens)
                                                  else lecture_module:=FALSE;
  end;
end; {ModuleSuivant}

{----------------------------------------------------------}

procedure AffChrono;
var minutes,secondes: SHORTINT;
begin
 AdrTexte :=PtrVRAM(0,x_chrono-1,y_chrono);
 AttrTexte:=16*noir+cyan;

 if Chrono then begin
   minutes :=Chrono_min;
   secondes:=Chrono_sec;
   AffCar(' ');
 end
 else begin
   minutes :=Duree_min-Chrono_min;
   secondes:=Duree_sec-Chrono_sec;
   if (secondes<0) and (minutes>0) then begin dec(minutes); inc(secondes,60); end;
   if Duree_50<Chrono_50 then dec(secondes);
   if secondes<0 then
    if minutes>0 then begin dec(minutes); inc(secondes,60); end
                 else secondes:=0;
   AffCar('-');
 end;

 AffCh(EntierTexte2(minutes  ,2+et_zeros)+':'+EntierTexte2(secondes ,2+et_zeros)+'/'+
       EntierTexte2(Duree_min,2+et_zeros)+':'+EntierTexte2(Duree_sec,2+et_zeros));
end; {AffChrono}

{----------------------------------------------------------}

procedure AffFonctions;
begin
 with Info^,Variables^ do begin
   {affiche le nom des fonctions}
   AffChXY(0,x_txt_Pause-1     ,y_txt_Pause     ,' '+XString(txt_Pause    ,' ',10),coul_fonc_c[Pause]);
   AffChXY(0,x_txt_Repeat-1    ,y_txt_Repeat    ,' '+XString(txt_Repeat   ,' ',10),coul_fonc[Boucler_MOD]);
   AffChXY(0,x_txt_Jump-1      ,y_txt_Jump      ,' '+XString(txt_Jump     ,' ',10),coul_fonc[Jump]);
   AffChXY(0,x_txt_Interpol2-1 ,y_txt_Interpol2 ,' '+XString(txt_Interpol2,' ',10),coul_fonc[Utiliser_Interpolation]);

   if not(Shuffle) then
    AffChXY(0,x_txt_Introscan2-1,y_txt_Introscan2,' '+XString(txt_Introscan,' ',10),coul_fonc[Introscan]);
   if not(Introscan) then
    AffChXY(0,x_txt_Shuffle-1   ,y_txt_Shuffle   ,' '+XString(txt_Shuffle  ,' ',10),coul_fonc[Shuffle]);

   AffChXY(0,x_txt_Fadeout-1   ,y_txt_Fadeout   ,' '+XString(txt_Fadeout  ,' ', 5),coul_fonc[Fading.actif]);
   AffChXY(0,x_txt_Mute-1      ,y_txt_Mute      ,' '+XString(txt_Mute     ,' ', 5),coul_fonc[Mute]);

   {affiche les symboles des fonctions}
   AdrTexte:=PtrVRAM(0,x_symb_Play,y_symb_Play);
                  AttrTexte:=coul_symb_Play     [not(Pause)   ]; AffCar(car_symb_Play  );
   inc(AdrTexte); AttrTexte:=coul_symb_Pause    [Pause        ]; AffCar(car_symb_Pause );
   inc(AdrTexte); AttrTexte:=coul_symb_Repeat   [Boucler_MOD  ]; AffCar(car_symb_Repeat);
   inc(AdrTexte); AttrTexte:=coul_symb_Jump     [Jump         ]; AffCar(car_symb_Jump  );
   inc(AdrTexte); AttrTexte:=coul_symb_Interpol                ; AffCh (ch_symb_Interpol[Utiliser_Interpolation]);
   inc(AdrTexte); AttrTexte:=coul_symb_Introscan[Introscan]    ; AffCar(car_symb_Introscan);
   inc(AdrTexte); AttrTexte:=coul_symb_Shuffle  [Shuffle]      ; AffCar(car_symb_Shuffle);
   inc(AdrTexte); AttrTexte:=coul_symb_Fadeout  [Fading.actif] ; AffCar(car_symb_Fadeout);
   inc(AdrTexte); AttrTexte:=coul_symb_Mute     [Mute]         ; AffCar(car_symb_Mute);
  end;
end; {AffFonctions}

{----------------------------------------------------------}

function AdrSample(n_smp: WORD): POINTER;
{renvoie l'adresse du sample en mmoire}
var i: BYTE;
begin
 AdrSample:=nil;

 with MUS.T^.PtrSamples[n_smp]^ do
  if (((Info and M_Adlib)=0) and
      ((liste_instr.etat[n_smp] and smp_charge)=smp_charge)) then
   with Mem do
    if segment<>$0000 then
     if handle_EMS<>EMS_ERR then begin {sample en EMS}
       for i:=0 to Nb_Pages_EMS-1 do EMSMap(handle_EMS,i,i);  {rend le sample accessible}
       AdrSample:=Ptr(Segment_EMS,0);
     end
     else AdrSample:=Ptr(segment,0);  {sample en mmoire DOS}
end; {AdrSample}

{----------------------------------------------------------}

procedure AfficheOscillo(xg,yh,larg_osc_car: BYTE; var onde: array of SHORTINT;
                         var table_car: array of CHAR; var curseur: array of T_CursOsc);

var oscillo: record
              nombre_car: BYTE;
              ligne: array[0..hauteur_osc_car-1] of STRING[largeur_osc_car];
              car_utilise: array[CHAR] of BOOLEAN;
              motif: array[CHAR] of array[0..7] of BYTE;
             end;

    c,c_debut: CHAR;
    larg_onde,i,y1,y2,ym,nombre_car_cons: BYTE;


function CalcYonde(y: SHORTINT): BYTE; assembler;
asm
 mov AL,y
 mov BH,hauteur_osc-1
 mul BH
 sub BH,AH
 mov AL,BH      {AL=(hauteur_osc-1)-((hauteur_osc-1)*y) div 256}
end;


procedure AffPointOsc(x,y: BYTE);
{0  x  largeur_osc-1, 0  y  hauteur_osc-1}
var ofs_nombre_car,ofs_ligne,ofs_car_utilise,ofs_motif: WORD;
    ptr_table: POINTER;
begin
 with oscillo do begin
   ofs_nombre_car:=Ofs(nombre_car);
   ofs_ligne:=Ofs(ligne);
   ofs_car_utilise:=Ofs(car_utilise);
   ofs_motif:=Ofs(motif);
   ptr_table:=@table_car;
 end;

 asm
  push DS
  push SS
  pop DS             {DS=SS (segment variables locales)}

  mov SI,ofs_nombre_car     {DS:SI -> nombre_car}
  mov DI,ofs_ligne          {DS:DI -> ligne[][]}
  mov BX,ofs_motif          {DS:BX -> motif[]}
  mov CX,07h                {pour masques x et y  la fin...}

  mov AH,y
  shr AH,3
  mov AL,largeur_osc_car+1
  mul AH
  add DI,AX
  xor AH,AH
  mov AL,x
  shr AL,3
  add DI,AX
  inc DI             {DS:DI -> ligne[y shr 3][(x shr 3)+1]}

  xor AX,AX
  or  AL,[DI]        {AL=car = ligne[][]}
  jnz @suite

  mov AL,[SI]
  cmp AL,nombre_car_limite
  jae @fin           {si nombre_car>=nombre_car_limite, plus de car. libres}

  inc BYTE([SI])     {nombre_car++}
  les SI,ptr_table   {ES:SI -> table_car[0]}
  add SI,AX
  mov AL,ES:[SI]
  mov [DI],AL        {car=ligne[][] = table_car[nombre_car-1]}
  mov DI,ofs_car_utilise
  add DI,AX
  mov BYTE([DI]),TRUE{car_utilise[car] = TRUE}

 @suite:
  shl AX,3
  add BX,AX
  mov AL,y
  and AX,CX
  add BX,AX          {DS:BX=masque = @motif[car][y and $07]}

  and CL,x
  mov AL,10000000b
  shr AL,CL
  or  [BX],AL        {masque^ = masque^ or ($80 shr (x and $07))}

 @fin:
  pop DS
 end;
end; {AffPointOsc}


procedure AffLigneOsc(x,y1,y2: BYTE);
{0  x  largeur_osc-1, 0  y1,y2  hauteur_osc-1}
var s: SHORTINT;
begin
 s:=Sgn(y2-y1);
 repeat AffPointOsc(x,y1); inc(y1,s); until y1=y2;
end; {AffLigneOsc}


procedure AffLignePointilleOsc(x: BYTE);
{0  x  largeur_osc-1}
var y: BYTE;
begin
 y:=0;
 repeat AffPointOsc(x,y); inc(y,3); until y>hauteur_osc-1;
end; {AffLignePointilleOsc}


begin  {AfficheOscillo}
 larg_onde:=8*larg_osc_car;
 FillChar(oscillo,sizeof(oscillo),0);

 with oscillo do begin
   y1:=CalcYonde(onde[0]);
   y2:=CalcYonde(onde[1]);
   ym:=(y1+y2) div 2;
   AffLigneOsc(0,y1,ym);
   for i:=1 to (larg_onde-1)-1 do begin
     y1:=y2;
     AffLigneOsc(i,ym,y1);
     y2:=CalcYonde(onde[i]);
     ym:=(y1+y2) div 2;
     AffLigneOsc(i,y1,ym);
   end;
   AffLigneOsc(larg_onde-1,ym,y1);

   i:=0;
   repeat
    with curseur[i] do
     case type_ of
      curs_dernier  : break;
      curs_vide     :;
      curs_continu  : if position<=larg_onde-1 then AffLigneOsc(position,0,hauteur_osc-1);
      curs_pointille: if position<=larg_onde-1 then AffLignePointilleOsc(position);
     end;
    inc(i);
   until FALSE;

   AttendVBL;

   c_debut:=#0;
   nombre_car_cons:=0;
   for c:=#0 to #255 do begin
     if car_utilise[c] then begin
       if c_debut=#0 then c_debut:=c;
       inc(nombre_car_cons);
     end;
     if (not(car_utilise[c]) or (c=#255)) then
      if nombre_car_cons<>0 then begin
        RedefCar(0,c_debut,nombre_car_cons,motif[c_debut],binaire);
        c_debut:=#0;
        nombre_car_cons:=0;
      end;
   end;

   for i:=0 to hauteur_osc_car-1 do begin
     ligne[i][0]:=CHAR(larg_osc_car);
     AffChXY(0,xg,yh+i,ligne[i],16*noir+vert);
   end;
 end;
end; {AfficheOscillo}

{----------------------------------------------------------}

procedure AffAscenseur(aff_fleche_haut,aff_fleche_bas: BOOLEAN);
{ascenseur pour la fentre infrieure}
var i: BYTE;

begin
 for i:=1 to 31 do
  AffCarXY(0,xd_fen_infos_mus-1,yh_fen_infos_mus+2+(i-1),' ',16*or_clair+blanc);

 if aff_fleche_haut then begin
   AffCarXY(0,xd_fen_infos_mus-1,yh_fen_infos_mus+2,'',16*or_clair+blanc);
   AffCarXY(0,xd_fen_infos_mus-1,yh_fen_infos_mus+3,'',16*or_clair+blanc);
 end;

 if aff_fleche_bas then begin
   AffCarXY(0,xd_fen_infos_mus-1,yb_fen_infos_mus-1,'',16*or_clair+blanc);
   AffCarXY(0,xd_fen_infos_mus-1,yb_fen_infos_mus-2,'',16*or_clair+blanc);
 end;
end; {AffAscenseur}

{----------------------------------------------------------}

procedure VoieSolo(n: BYTE);
{met la voie n en solo / ractive toutes les voies}
var v: BYTE;
    solo: BOOLEAN;

begin
 with MUS.H^ do begin
   solo:=FALSE;
   for v:=1 to Nb_de_voies_patt do
    if not(Voie_Active(v)) then if solo then break else solo:=TRUE;

   if not(solo) or not(Voie_Active(n)) then begin {met la voie en solo}
     Activer_Voie(n);
     for v:=1 to Nb_de_voies_patt do if v<>n then Desactiver_Voie(v);
   end
   else for v:=1 to Nb_de_voies_patt do Activer_Voie(v);
 end;
end; {VoieSolo}

{----------------------------------------------------------}

procedure ChangeLangueNote;
begin
 case LangueNote of
  Anglais : LangueNote:=Francais;
  Francais: LangueNote:=Anglais;
 end;
end; {ChangeLangueNote}

{----------------------------------------------------------}

function ActiveDesactiveInstr(n: BYTE): BOOLEAN;
{renvoie TRUE si l'instrument n a t activ ou dsactiv}
begin
 with liste_instr do
  if (etat[n] and smp_charge)=smp_charge then begin
    etat[n]:=etat[n] xor smp_stoppe;
    with MUS.T^.PtrSamples[n]^ do Info:=Info xor M_Charge;
    ActiveDesactiveInstr:=TRUE;
  end
  else ActiveDesactiveInstr:=FALSE;
end; {ActiveDesactiveInstr}

{----------------------------------------------------------}
{ GESTION DE L'AIDE                                        }
{----------------------------------------------------------}

procedure InitStrucAide; far;
begin
  {rien!}
end; {InitStrucAide}

{----------------------------------------------------------}

procedure InitAffAide; far;

const S_VIDE: STRING[1]='';

      {textes anglais}
      S_A_01: STRING[24+2*1]='^C^  Chronometer/Countdown';
      S_A_02: STRING[12+2*1]='^D^  DOS shell';
      S_A_03: STRING[37+2*1]='^I^  Enable/Disable 8-bit interpolation';
      S_A_04: STRING[28+2*1]='^J^  Allow/Skip backward jumps';
      S_A_05: STRING[22+2*1]='^M^  Enable/Disable mute';
      S_A_06: STRING[23+2*1]='^P^  Enable/Disable pause';
      S_A_07: STRING[24+2*1]='^R^  Enable/Disable repeat';
      S_A_08: STRING[20+2*1]='^S^  Disable introscan';
      S_A_09: STRING[34+2*2]='^-^/^+^  Decrease/Increase main volume';
      S_A_10: STRING[34+2*2]='^0^..^9^  Enable/Disable channel 1..10';
      S_A_11: STRING[38+2*4]='^Ctrl^+^'#27'^/^Ctrl^+^'#26'^  Fast rewind/forward'; {<- car. spciaux}
      S_A_12: STRING[67+2*4]='^Ctrl^+^Page^/^Ctrl^+^Page^  Previous/Next module within current playlist';
      S_A_13: STRING[54+2*1]='^Space^  Previous or next module within current playlist';
      S_A_14: STRING[42+2*1]='^F2^  Samples list (name, length, loop, ...)';
      S_A_15: STRING[42+2*1]='^F3^  Channels list (volume, sample, effect)';
      S_A_16: STRING[63+2*1]='^F4^  Partition (sample, note, octave, volume, effect, parameter)';
      S_A_17: STRING[33+2*1]='^F5^  Keyboards and chords detector';
      S_A_18: STRING[27+2*1]='^F6^  Sound spectrum analyzer';
      S_A_19: STRING[44+2*1]='^Esc^ Stop playing the current module/playlist';

      S_A_20: STRING[50+2*6]='^^/^^ ^Page^/^Page^ ^Home^/^end^  Move the sample selector';
      S_A_21: STRING[39+2*1]='^^  Enable/Disable the selected sample';
      S_A_22: STRING[61+2*1]='^Tab^  Swap the scope display: sound output/current sample wave';
      S_A_23: STRING[35+2*2]='^'#27'^/^'#26'^  Scroll the current sample wave'; {<- car. spciaux}
      S_A_24: STRING[38+2*2]='^/^/^*^  Decrease/Increase the zoom factor';

      S_A_25: STRING[51+2*6]='^^/^^ ^Page^/^Page^ ^Home^/^end^  Move the channel selector';
      S_A_26: STRING[40+2*1]='^^  Enable/Disable the selected channel';
      S_A_27: STRING[58+2*2]='^Ctrl^+^^  Enable/Disable the selected channel playing solo';
      S_A_28: STRING[74+2*3]='^Tab^/^Shift^+^Tab^  Change the display: mono/stereo pitch bars+volume/note bars';
      S_A_29: STRING[47+2*1]='^N^  Swap the type of note (A B C.../Do R Mi...)';

      S_A_30: STRING[51+2*6]='^'#27'^/^'#26'^ ^Page^/^Page^ ^Home^/^end^  Move the channel selector'; {<- car. spciaux}
      S_A_31: STRING[63+2*3]='^Tab^/^Shift^+^Tab^  Increase/Decrease the number of visible channels';

      S_A_32: STRING[59+2*6]='^^/^^ ^Page^/^Page^ ^Home^/^end^  Move the sample/keyboard selector';
      S_A_33: STRING[44+2*2]='^'#27'^/^'#26'^  Scroll the keyboards one octave down/up'; {<- car. spciaux}

      S_A_34: STRING[54+2*6]='^'#27'^/^'#26'^ ^Page^/^Page^ ^Home^/^end^  Scroll the frequencies scale'; {<- car. spciaux}
      S_A_35: STRING[67+2*2]='^^/^^  Increase/Decrease the signal amplify before the FFT processing';
      S_A_36: STRING[43+2*1]='^Tab^  Swap the spectrum display: mono/stereo';


      {textes franais}
      S_F_01: STRING[31+2*1]='^C^  Chronomtre/Compte  rebours';
      S_F_02: STRING[16+2*1]='^D^  Commandes DOS';
      S_F_03: STRING[40+2*1]='^I^  Active/Dsactive interpolation 8 bits';
      S_F_04: STRING[41+2*1]='^J^  Autorise/Ignore sauts arrires (Jump)';
      S_F_05: STRING[35+2*1]='^M^  Active/Dsactive sourdine (Mute)';
      S_F_06: STRING[25+2*1]='^P^  Active/Dsactive pause';
      S_F_07: STRING[39+2*1]='^R^  Active/Dsactive rptition (repeat)';
      S_F_08: STRING[22+2*1]='^S^  Dsactive introscan';
      S_F_09: STRING[41+2*2]='^-^/^+^  Diminue/Augmente le volume principal';
      S_F_10: STRING[33+2*2]='^0^..^9^  Active/Dsactive voie 1..10';
      S_F_11: STRING[42+2*4]='^Ctrl^+^'#27'^/^Ctrl^+^'#26'^  Avance/Retour rapide module'; {<- car. spciaux}
      S_F_12: STRING[68+2*4]='^Ctrl^+^Page^/^Ctrl^+^Page^  Module prcdent/suivant du programme courant';
      S_F_13: STRING[56+2*1]='^Espace^  Module prcdent ou suivant du programme courant';
      S_F_14: STRING[55+2*1]='^F2^  Liste des intruments (nom, longueur, bouclage, ...)';
      S_F_15: STRING[47+2*1]='^F3^  Liste des voies (volume, instrument, effet)';
      S_F_16: STRING[66+2*1]='^F4^  Partition (instrument, note, octave, volume, effet, paramtre)';
      S_F_17: STRING[35+2*1]='^F5^  Claviers et dtecteur d''accords'; {<- guillemets}
      S_F_18: STRING[31+2*1]='^F6^  Analyseur de spectre sonore';
      S_F_19: STRING[53+2*1]='^Echap^  Termine la lecture du module/programme courant';

      S_F_20: STRING[60+2*6]='^^/^^ ^Page^/^Page^ ^Dbut^/^Fin^  Dplace le slecteur d''instrument'; {<- guillemets}
      S_F_21: STRING[46+2*1]='^^  Active/Dsactive l''instrument slectionn'; {<- guillemets}
      S_F_22: STRING[73+2*1]='^Tab^  Change l''affichage de l''oscilloscope: canal de sortie/instr. courant'; {<- guillemets}
      S_F_23: STRING[48+2*2]='^'#27'^/^'#26'^  Fait dfiler l''onde de l''instrument courant'; {<- guillemets+car. spciaux}
      S_F_24: STRING[40+2*2]='^/^/^*^  Diminue/Augmente le facteur de zoom';

      S_F_25: STRING[55+2*6]='^^/^^ ^Page^/^Page^ ^Dbut^/^Fin^  Dplace le slecteur de voie';
      S_F_26: STRING[42+2*1]='^^  Active/Dsactive la voie slectionne';
      S_F_27: STRING[55+2*2]='^Ctrl^+^^  Active/Dsactive la voie slectionne en solo';
      S_F_28: STRING[75+2*3]='^Tab^/^Maj^+^Tab^  Change l''affichage: barres pitch mono/stro+vol./barres notes'; {<- guill.}
      S_F_29: STRING[48+2*1]='^N^  Change le type de note (A B C.../Do R Mi...)';

      S_F_30: STRING[55+2*6]='^'#27'^/^'#26'^ ^Page^/^Page^ ^Dbut^/^Fin^  Dplace le slecteur de voie'; {<- car. spciaux}
      S_F_31: STRING[57+2*3]='^Tab^/^Maj^+^Tab^  Augmente/Diminue le nombre de voies visibles';

      S_F_32: STRING[68+2*6]='^^/^^ ^Page^/^Page^ ^Dbut^/^Fin^  Dplace le slecteur d''instrument/clavier'; {<- guill.}
      S_F_33: STRING[57+2*2]='^'#27'^/^'#26'^  Fait dfiler les claviers d''une octave plus bas/haut'; {<- guill.+car. spc.}

      S_F_34: STRING[63+2*6]='^'#27'^/^'#26'^ ^Page^/^Page^ ^Dbut^/^Fin^  Fait dfiler l''chelle de frquences'; {gui+sp}
      S_F_35: STRING[70+2*2]='^^/^^  Augmente/Diminue l''amplification du signal avant calcul de la FFT'; {<- guillemets}
      S_F_36: STRING[47+2*1]='^Tab^  Change l''affichage du spectre: mono/stro'; {<- guillemets}


      nbr_lig_aide_gene=20;
      Aide_gene: array[langues,1..nbr_lig_aide_gene] of ^STRING
                 =((@S_A_01,@S_A_02,@S_A_03,@S_A_04,@S_A_05,
                    @S_A_06,@S_A_07,@S_A_08,@S_A_09,@S_A_10,
                    @S_A_11,@S_A_12,@S_A_13,@S_VIDE,@S_A_14,
                    @S_A_15,@S_A_16,@S_A_17,@S_A_18,@S_A_19),

                   (@S_F_01,@S_F_02,@S_F_03,@S_F_04,@S_F_05,
                    @S_F_06,@S_F_07,@S_F_08,@S_F_09,@S_F_10,
                    @S_F_11,@S_F_12,@S_F_13,@S_VIDE,@S_F_14,
                    @S_F_15,@S_F_16,@S_F_17,@S_F_18,@S_F_19)
                  );

      nbr_lig_aide_loc=6;
      Aide_loc: array[aff_instr..aff_spect] of
                 record
                  nbr: BYTE;
                  txt: array[langues,1..nbr_lig_aide_loc] of ^STRING;
                 end

                =({liste instruments}
                  (nbr:6; txt:((@S_A_20,@S_A_21,@S_A_22,@S_VIDE,@S_A_23,@S_A_24),
                               (@S_F_20,@S_F_21,@S_F_22,@S_VIDE,@S_F_23,@S_F_24))),

                  {liste voies}
                  (nbr:5; txt:((@S_A_25,@S_A_26,@S_A_27,@S_A_28,@S_A_29,nil    ),
                               (@S_F_25,@S_F_26,@S_F_27,@S_F_28,@S_F_29,nil    ))),

                  {partition}
                  (nbr:5; txt:((@S_A_30,@S_A_26,@S_A_27,@S_A_31,@S_A_29,nil    ),
                               (@S_F_30,@S_F_26,@S_F_27,@S_F_31,@S_F_29,nil    ))),

                  {claviers}
                  (nbr:4; txt:((@S_A_32,@S_A_33,@S_A_21,@S_A_29,nil    ,nil    ),
                               (@S_F_32,@S_F_33,@S_F_21,@S_F_29,nil    ,nil    ))),

                  {spectre}
                  (nbr:3; txt:((@S_A_34,@S_A_35,@S_A_36,nil    ,nil    ,nil    ),
                               (@S_F_34,@S_F_35,@S_F_36,nil    ,nil    ,nil    )))
                 );

var i: BYTE;

begin
 {cadre aide Touches Gnrales}
 Cadre(0,1,xg_cadre_aide_gene,yh_cadre_aide_gene,xd_cadre_aide_gene,yb_cadre_aide_gene,16*or_fond+gris_fonce,'',0,'');
 AffChXY(0,xg_cadre_aide_gene+6,yh_cadre_aide_gene,txt_aide_gene[Langue],16*or_fond+gris);

 {cadre aide Touches Locales}
 Cadre(0,1,xg_cadre_aide_loc,yh_cadre_aide_loc,xd_cadre_aide_loc,yb_cadre_aide_loc,16*or_fond+gris_fonce,'',0,'');
 AffChXY(0,xg_cadre_aide_loc+6,yh_cadre_aide_loc,txt_aide_loc[Langue],16*or_fond+gris);

 for i:=1 to nbr_lig_aide_gene do
  AffChXY2(0,xg_cadre_aide_gene+1+1,yh_cadre_aide_gene+i+1,Aide_gene[Langue,i]^,16*or_fond+gris_fonce,16*or_fond+or_clair);

 with Aide_loc[anc_type_aff] do
  for i:=1 to nbr do
   AffChXY2(0,xg_cadre_aide_loc+1+1,yh_cadre_aide_loc+i+1,txt[Langue,i]^,16*or_fond+gris_fonce,16*or_fond+or_clair);

end; {InitAffAide}

{----------------------------------------------------------}

procedure AffAide; far;
begin
  {rien!}
end; {AffAide}

{----------------------------------------------------------}

procedure TesteToucheAide(touche: Key); far;
begin
 {rtablit l'ancien cran quelle que soit la touche}
 type_aff:=anc_type_aff;
 reaffiche:=reaff_fen;
end; {TesteToucheAide}

{----------------------------------------------------------}
{ GESTION DE LA LISTE DES INSTRUMENTS                      }
{----------------------------------------------------------}

procedure InitStrucInstr; far;

var s: BYTE;

begin
 with liste_instr,MUS do begin
   mode_oscillo:=oscillo_son;

   nombre:=H^.Nb_Total_Samples;
   premier:=1;
   courant:=1;

   FillChar(etat,sizeof(etat),smp_pasjoue);
   {recopie le flag M_Charge pour contrler si l'instrument peut tre stopp}
   for s:=1 to nombre do with T^.PtrSamples[s]^ do
    if (Info and M_Charge)=M_Charge then etat[s]:=etat[s] or smp_charge;

   zoom:=zoom_smp_min;
   debut_onde:=0;
   echant_onde:=TRUE;
 end;
end; {InitStrucInstr}

{----------------------------------------------------------}

procedure InitAffInstr; far;
begin
 AffChXY(0,xg_fen_infos_mus+1,yh_fen_infos_mus+1,
           '    '+XCString(txt_instr[Langue],' ',32)+'type '+
           txt_long[Langue]+'loop'#15'loop'#12'vol C4  fine ',
           16*or_sombre+or_clair);
end; {InitAffInstr}

{----------------------------------------------------------}

procedure AffLigneInstr(n_smp: BYTE; y: BYTE; aff_tout: BOOLEAN; coul: type_coul);

var type_smp,long_smp,loop_deb,loop_len,C4_smp: STRING[5];
    vol_smp: STRING[3];
    fine_smp: CHAR;
    i,coul_fond,coul_clignote,coul_note: BYTE;

begin
 coul_fond:=16*couleur_fond[coul];
 coul_clignote:=0;

 with liste_instr,MUS.T^.PtrSamples[n_smp]^ do begin
   if (etat[n_smp] and smp_joue)=smp_joue then coul_note:=jaune
   else
   if (etat[n_smp] and smp_actif)=smp_actif then coul_note:=gris
   else
   if (etat[n_smp] and smp_dejajoue)=smp_dejajoue then coul_note:=or_sombre
   else coul_note:=couleur_fond[coul];

   if aff_tout then begin
     if ((etat[n_smp] and smp_charge)=smp_charge) or (Len<>0) then begin {instrument charg ou non vide}
       if (etat[n_smp] and smp_charge)=0 then coul_clignote:=128; {instrument non vide mais pas charg}
       case (Info and M_Adlib) of
        0      : begin
                  case (Info and M_16bit) of
                   0      : type_smp:='SMP8';
                   M_16bit: type_smp:='SMP16';
                  end;

                  long_smp:=EntierTexte2(Len,5);
                  if Rep_Len=0 then begin
                    loop_deb:='';
                    loop_len:='';
                  end
                  else begin
                    loop_deb:=EntierTexte2(Rep,5);
                    loop_len:=EntierTexte2(Rep_Len,5);
                  end;
                  vol_smp :=EntierHexa(Volume,2)+'h';
                  C4_smp  :=EntierTexte2(C4Spd,5);
                  fine_smp:=CHAR(BYTE('0')+BYTE(Fine_Tune1));
                 end;

        M_Adlib: begin
                  type_smp:='ADLIB';
                  long_smp:='-----';  {\                        }
                  loop_deb:=long_smp; { > n'existe pas en Adlib }
                  loop_len:=loop_deb; {/                        }
                  vol_smp :=EntierHexa(AVolume,2)+'h';
                  C4_smp  :=EntierTexte2(AC4Spd,5);
                  fine_smp:=' ';
                 end;
       end;
     end
     else begin  {instrument vide}
       type_smp:='';
       long_smp:='';
       loop_deb:='';
       loop_len:='';
       vol_smp :='';
       C4_smp  :='';
       fine_smp:=' ';
     end;

     AdrTexte :=PtrVRAM(0,xg_fen_infos_mus+1,y);

     AttrTexte:=coul_fond+or_sombre; AffCar('');
     AttrTexte:=coul_fond+couleur_curs[coul]; AffCh(EntierTexte2(n_smp,2));
     AttrTexte:=coul_fond+couleur_LED_vr[(etat[n_smp] and smp_charge)=smp_charge,
                                         (etat[n_smp] and smp_stoppe)=smp_stoppe]; AffCar(car_LED);
     AttrTexte:=coul_fond+coul_note; AffCar('');
     AttrTexte:=coul_fond+or_sombre; AffCar('');
     AttrTexte:=coul_fond+noir     ; AffChE(Nom,' ',32);
     AttrTexte:=coul_fond+or_sombre; AffCar('');
     AttrTexte:=coul_fond+gris+coul_clignote; AffChE(type_smp,' ',5);
     AttrTexte:=coul_fond+or_sombre; AffCar('');
     AttrTexte:=coul_fond+gris     ; AffChE(long_smp,' ',5);
     AttrTexte:=coul_fond+or_sombre; AffCar('');
     AttrTexte:=coul_fond+gris     ; AffChE(loop_deb,' ',5);
     AttrTexte:=coul_fond+or_sombre; AffCar('');
     AttrTexte:=coul_fond+gris     ; AffChE(loop_len,' ',5);
     AttrTexte:=coul_fond+or_sombre; AffCar('');
     AttrTexte:=coul_fond+gris     ; AffChE(vol_smp,' ',3);
     AttrTexte:=coul_fond+or_sombre; AffCar('');
     AttrTexte:=coul_fond+gris     ; AffChE(C4_smp,' ',5);
     AttrTexte:=coul_fond+or_sombre; AffCar('');
     AttrTexte:=coul_fond+gris     ; AffChE(' '+fine_smp,' ',1+1+1);
     AttrTexte:=coul_fond+or_sombre; AffCar('');
   end
   else AffCarXY(0,xg_fen_infos_mus+1+4,y,'',coul_fond+coul_note);
 end;
end; {AffLigneInstr}

{----------------------------------------------------------}

procedure AffInstr; far;

var i: BYTE;

begin
 with liste_instr do begin
   for i:=premier to premier+31-1 do if i<=nombre then
    AffLigneInstr(i,yh_fen_infos_mus+2+(i-premier),reaffiche>=reaff_lignes,type_coul(i=courant));

   if nombre>31 then AffAscenseur(premier>1,premier+31-1<nombre);
 end;
end; {AffInstr}

{----------------------------------------------------------}

procedure TesteToucheInstr(touche: Key); far;

var anc_courant: BYTE;

begin
 with liste_instr do begin
   anc_courant:=courant;

   case touche of
    Up  : if courant>1 then begin
            dec(courant);
            if courant<premier then dec(premier);
            reaffiche:=reaff_lignes;
          end;

    Down: if courant<nombre then begin
            inc(courant);
            if courant>premier+31-1 then inc(premier);
            reaffiche:=reaff_lignes;
          end;

    PgUp: begin
           if courant>16 then begin
             dec(courant,16);
             if courant<premier then premier:=courant;
           end
           else begin
             courant:=1;
             premier:=1;
           end;
           reaffiche:=reaff_lignes;
          end;

    PgDn: begin
           inc(courant,16);
           if courant>nombre then courant:=nombre;
           if nombre>31 then
            if premier<courant-31+1 then premier:=courant-31+1;
           reaffiche:=reaff_lignes;
          end;

    Home: if courant>1 then begin
            courant:=1;
            premier:=1;
            reaffiche:=reaff_lignes;
          end;

    End_: if courant<nombre then begin
            courant:=nombre;
            if nombre>31 then premier:=nombre-31+1;
            reaffiche:=reaff_lignes;
          end;

    {active/dsactive l'instrument courant}
    Enter: if ActiveDesactiveInstr(courant) then reaffiche:=reaff_lignes;

    {inverse le mode d'affichage de l'oscillo: son (canal de sortie) <-> instrument courant}
    Tab: begin
          case mode_oscillo of
           oscillo_son  : mode_oscillo:=oscillo_instr;
           oscillo_instr: mode_oscillo:=oscillo_son;
          end;
          {efface le fond de l'oscillo}
          EffaceZone(0,x_oscillo-2,y_oscillo,x_oscillo+largeur_osc_car,y_oscillo+hauteur_osc_car-1,16*or_fond);
          echant_onde:=TRUE;
          reaffiche:=reaff_rien;
         end;
   end;

   {cas particuliers en mode oscillo instrument}
   if (mode_oscillo=oscillo_instr) and ((etat[courant] and smp_charge)=smp_charge) then begin
     case touche of
      Left : if debut_onde>=pas_scroll_smp then dec(debut_onde,pas_scroll_smp)
                                           else debut_onde:=0;
      Right: inc(debut_onde,pas_scroll_smp);

      Key('*'): if MUS.T^.PtrSamples[courant]^.Len>=(largeur_osc shl (zoom+1)) then
                 if zoom<zoom_smp_max then begin
                   inc(zoom);
                   debut_onde:=debut_onde*2;
                   echant_onde:=TRUE;
                 end;

      Key('/'): if zoom>zoom_smp_min then begin
                  dec(zoom);
                  debut_onde:=debut_onde div 2;
                  echant_onde:=TRUE;
                end;
     end;

     if debut_onde+largeur_osc>(largeur_osc shl zoom) then debut_onde:=(largeur_osc shl zoom)-largeur_osc;
   end;

   {rinit. affichage onde si l'instrument courant change}
   if courant<>anc_courant then begin
     zoom:=zoom_smp_min;
     debut_onde:=0;
     echant_onde:=TRUE;
   end;
 end;
end; {TesteToucheInstr}

{----------------------------------------------------------}

procedure MAJ_EtatSmp;

var v,s: BYTE;
    etat_change: array[1..Max_Samples] of BOOLEAN;

begin
 FillChar(etat_change,sizeof(etat_change),FALSE);

 with Info^,liste_instr do begin
   {modifie l'tat de chaque instrument actif sur au moins une voie}
   for v:=1 to MUS.H^.Nb_de_voies_patt do begin
     s:=VSample[v]+1; {n d'instrument actif sur la voie}
     if s in [1..Max_Samples] then begin
       etat[s]:=etat[s] or (smp_actif or smp_dejajoue);
       if (VControle[v] and M_Pas_Active)=M_Pas_Active then etat[s]:=etat[s] and not(smp_joue)
                                                       else etat[s]:=etat[s] or smp_joue;
       etat_change[s]:=TRUE;
     end;
   end;

   {modifie l'tat des instruments non actifs sur aucune voie}
   for s:=1 to nombre do if not(etat_change[s]) then
    etat[s]:=etat[s] and not(smp_actif or smp_joue);
 end;
end; {MAJ_EtatSmp}

{----------------------------------------------------------}
{ GESTION DE LA LISTE DES VOIES                            }
{----------------------------------------------------------}

procedure InitStrucVoies; far;
begin
 with liste_voies do begin
   nombre:=MUS.H^.Nb_de_voies;
   courante:=1;

   type_barres:=type_barres_;

   {contrle de validit}
   if (type_barres=barres_pitch_stereo)
      and not(Variables^.Utilise_Stereo) then type_barres:=barres_pitch_mono;

   FillChar(sample,sizeof(sample),0);
   FillChar(volume,sizeof(volume),0);
   FillChar(pitch  ,sizeof(pitch),0);
   FillChar(pitch_g,sizeof(pitch_g),0);
   FillChar(pitch_d,sizeof(pitch_d),0);

   vitesse_pitch:=2;
 end;
end; {InitStrucVoies}

{----------------------------------------------------------}

procedure InitAffVoies; far;
begin
  {rien!}
end; {InitAffVoies}

{----------------------------------------------------------}

procedure AffBarrePitch(x,y: BYTE; longueur_max: BYTE; valeur,valeur_max: BYTE; sens: SHORTINT);

var longueur,couleur,i,seuil_orange,seuil_rouge: BYTE;

begin
 longueur:=MinEntier((WORD(longueur_max)*(valeur+(valeur_max div longueur_max))) div valeur_max,longueur_max);

 seuil_orange:=(13*longueur_max) div 22;
 seuil_rouge :=(19*longueur_max) div 22;

 for i:=1 to longueur do begin
   if i>=seuil_rouge then couleur:=rouge
   else
    if i>=seuil_orange then couleur:=orange else couleur:=vert;
   AffCarXY(0,x,y,car_curs,16*noir+couleur);
   inc(x,sens);
 end;

 for i:=longueur+1 to longueur_max do begin
   AffCarXY(0,x,y,car_curs,16*noir+gris_fonce);
   inc(x,sens);
 end;
end; {AffBarrePitch}

{----------------------------------------------------------}

procedure AffBarrePeriode(x,y: BYTE; longueur_max: BYTE; valeur: WORD);

const couleur: array[BOOLEAN] of BYTE=(16*noir+gris_fonce,16*noir+cyan);

var p,i: BYTE;

begin
 p:=(longueur_max+1)-LimiteEntier(valeur shr 6,1,longueur_max);

 for i:=1 to longueur_max do begin
   AffCarXY(0,x,y,car_barre_curs,couleur[i=p]);
   inc(x);
 end;
end; {AffBarrePeriode}

{----------------------------------------------------------}

procedure AffLigneVoie(n_voie: BYTE; y: BYTE; coul: type_coul);

const couleur_numsmp: array[BOOLEAN] of BYTE=(gris,vert_pastel);

var coul_fond,coul_numsmp: BYTE;
    n_smp,num_comm,i: BYTE;
    p: ^T_Note;

    valeur  : STRING[4];
    num_smp : STRING[2];
    nom_smp : STRING[22];
    nom_comm: STRING[20];

begin
 coul_fond:=16*couleur_fond[coul];

 with MUS,Info^,liste_voies do begin
   n_smp:=VSample[n_voie]+1;
   if n_smp in [1..liste_instr.nombre] then begin
     num_smp:=EntierTexte2(n_smp,2);
     nom_smp:=T^.PtrSamples[n_smp]^.Nom;
   end
   else begin
     num_smp:='';
     nom_smp:='';
   end;

   coul_numsmp:=couleur_numsmp[(VControle[n_voie] and M_Pas_Active)=0];

   p:=Ptr(Segment_Pattern,sizeof(T_Note)*((n_voie-1)+H^.Nb_de_voies_patt*Pos_Note));
   num_comm:=p^.Commande;
   if num_comm in [Arpeggio..Release] then nom_comm:=NomCommande[num_comm]
                                      else nom_comm:='';

   AdrTexte :=PtrVRAM(0,xg_fen_infos_mus+1,y);

   AttrTexte:=coul_fond+couleur_curs[coul]; AffCh(EntierTexte2(n_voie,2));
   AttrTexte:=coul_fond+couleur_LED_vr[n_voie<=H^.Nb_de_voies_patt,not(Voie_Active(n_voie))]; AffCar(car_LED);
   AttrTexte:=16*noir+or_clair; AffCar(car_HP[Volume_Gauche[n_voie]<Volume_Droite[n_voie]]);

   case type_barres of
    barres_pitch_mono  : begin
                          AffBarrePitch(xg_fen_infos_mus+4+1,y,23,pitch[n_voie],$80,+1);

                          valeur:=EntierHexa(VVolume[n_voie],2)+'h';
                         end;

    barres_pitch_stereo: begin
                          AffBarrePitch(xg_fen_infos_mus+4+11  ,y,11,pitch_g[n_voie],$60,-1);
                          AffBarrePitch(xg_fen_infos_mus+4+11+2,y,11,pitch_d[n_voie],$60,+1);
                          AffCarXY   (0,xg_fen_infos_mus+4+11+1,y,'',16*noir+or_fond);

                          valeur:=EntierHexa(VVolume[n_voie],2)+'h';
                         end;

    barres_periode     : begin
                          AffBarrePeriode(xg_fen_infos_mus+4+1,y,23,VPeriode_re[n_voie]);

                          with p^ do
                           case Note of
                            1..Max_Notes: valeur:=NomNote[LangueNote,(Note-1) mod 12]+CHAR(BYTE('0')+((Note-1) div 12));
                            Note_KeyOff : valeur:=Aff_KeyOff;
                            else          valeur:='';
                           end;
                         end;
   end;

   while Length(valeur)<4 do valeur:=' '+valeur;

   AdrTexte :=PtrVRAM(0,xg_fen_infos_mus+1+4+1+22,y);

   AttrTexte:=coul_fond+gris       ; AffChE(valeur  ,' ',4+1);
   AttrTexte:=coul_fond+coul_numsmp; AffChE(num_smp ,' ',2+1);
   AttrTexte:=coul_fond+noir       ; AffChE(nom_smp ,' ',22+1);
   AttrTexte:=coul_fond+orange     ; AffChE(nom_comm,' ',20);
 end;
end; {AffLigneVoie}

{----------------------------------------------------------}

procedure AffVoies; far;

var v: BYTE;

begin
 with liste_voies do begin
   for v:=1 to MUS.H^.Nb_de_voies_patt do
    AffLigneVoie(v,yh_fen_infos_mus+1+(v-1),type_coul(v=courante));
 end;
end; {AffVoies}

{----------------------------------------------------------}

procedure TesteToucheVoies(touche: Key); far;

var v: BYTE;
    solo: BOOLEAN;

begin
 with liste_voies,Info^ do
  case touche of
   Up  : if courante>1 then begin
           dec(courante);
           reaffiche:=reaff_lignes;
         end;

   Down: if courante<nombre then begin
           inc(courante);
           reaffiche:=reaff_lignes;
         end;

   PgUp: begin
          if courante>8 then dec(courante,8) else courante:=1;
          reaffiche:=reaff_lignes;
         end;

   PgDn: begin
          inc(courante,8);
          if courante>nombre then courante:=nombre;
          reaffiche:=reaff_lignes;
         end;

   Home: if courante>1 then begin
           courante:=1;
           reaffiche:=reaff_lignes;
         end;

   End_: if courante<nombre then begin
           courante:=nombre;
           reaffiche:=reaff_lignes;
         end;

   {active/dsactive la voie courante}
   Enter: begin
           if Voie_Active(courante) then Desactiver_Voie(courante) else Activer_Voie(courante);
           reaffiche:=reaff_titre;
          end;

   {voie courante en solo/active toutes les voies}
   CtrlEnter: begin
               VoieSolo(courante);
               reaffiche:=reaff_titre;
              end;

   Tab: begin
         case type_barres of
          barres_pitch_mono  : if Variables^.Utilise_Stereo then type_barres:=barres_pitch_stereo
                                                            else type_barres:=barres_periode;
          barres_pitch_stereo: type_barres:=barres_periode;
          barres_periode     : type_barres:=barres_pitch_mono;
         end;
         reaffiche:=reaff_rien;
        end;

   ShiftTab: begin
              case type_barres of
               barres_pitch_mono  : type_barres:=barres_periode;
               barres_pitch_stereo: type_barres:=barres_pitch_mono;
               barres_periode     : if Variables^.Utilise_Stereo then type_barres:=barres_pitch_stereo
                                                                 else type_barres:=barres_pitch_mono;
              end;
              reaffiche:=reaff_rien;
             end;

   Key('N'): if type_barres=barres_periode then ChangeLangueNote;
  end;
end; {TesteToucheVoies}

{----------------------------------------------------------}

procedure MAJ_EtatVoies;

var p: ^T_Note;
    n_voies,v: BYTE;

begin
 with Info^,liste_voies do begin
   n_voies:=MUS.H^.Nb_de_voies_patt;
   p:=Ptr(Segment_Pattern,sizeof(T_Note)*((1-1)+n_voies*Pos_Note));

   for v:=1 to n_voies do begin
     {ne met  jour que les voies actives (instrument est encore jou), non stoppes, et non 'keyoff'}
     if (VControle[v] and (M_Pas_Active+M_Stoppee+M_Keyoff))=0 then begin
       {rajuste le pitch si changement ou rattaque de l'instrument (note/volume)}
       if (p^.Instrument<>0) or (p^.Note<>0) or (p^.Volume<=$80) or
          (VSample[v]<>sample[v]) or (VVolume[v]<>volume[v]) then begin
         volume[v] :=VVolume[v];
         sample[v] :=VSample[v];
         pitch[v]  :=volume[v];
         pitch_g[v]:=Volume_Gauche[v];
         pitch_d[v]:=Volume_Droite[v];
       end;
     end;

     if Tempo<=8 then vitesse_pitch:=11-Tempo else vitesse_pitch:=2;
     if pitch[v]  >vitesse_pitch then dec(pitch[v]  ,vitesse_pitch) else pitch[v]  :=0;
     if pitch_g[v]>vitesse_pitch then dec(pitch_g[v],vitesse_pitch) else pitch_g[v]:=0;
     if pitch_d[v]>vitesse_pitch then dec(pitch_d[v],vitesse_pitch) else pitch_d[v]:=0;

     inc(p);
   end;
 end;
end; {MAJ_EtatVoies}

{----------------------------------------------------------}
{ GESTION DE LA PARTITION                                  }
{----------------------------------------------------------}

procedure InitStrucPart; far;
begin
 with partition do begin
   {adapte automatiquement le nombre de voies visibles selon le nombre de voies}
   case MUS.H^.Nb_de_voies of
    0..4: nombre_voies_vis:=4;
    5   : nombre_voies_vis:=5;
    6   : nombre_voies_vis:=6;
    7,8 : nombre_voies_vis:=8;
    9..Max_Voies: nombre_voies_vis:=9;
   end;
   premiere_voie:=1;
   voie_courante:=1;

   anc_pos_note :=$FF; {pour valider changement dans AffPart}
   anc_pattern  :=$FF;

   FillChar(ligne_aff,sizeof(ligne_aff),$00);
 end;
end; {InitStrucPart}

{----------------------------------------------------------}

procedure InitAffPart; far;

const couleur_voie: array[BOOLEAN] of BYTE=(or_clair,blanc);

var largeur,n_voies,v: BYTE;

begin
 with partition,Info^ do begin
   case nombre_voies_vis of
    4: largeur:=17;
    5: largeur:=14;
    6: largeur:=11;
    8: largeur:=8;
    9: largeur:=7;
   end;

   n_voies:=MUS.H^.Nb_de_voies_patt;

   AdrTexte :=PtrVRAM(0,xg_fen_infos_mus+1,yh_fen_infos_mus+1);
   AttrTexte:=16*or_sombre+or_clair;

   if nombre_voies_vis<>5 then AffCh('   ') else AffCh('   ');

   for v:=1 to nombre_voies_vis do begin
     AttrTexte:=16*or_sombre+or_clair; AffChE('',' ',1+((largeur-3) div 2));
     AttrTexte:=16*or_sombre+couleur_LED_vr[premiere_voie+(v-1)<=n_voies,not(Voie_Active(premiere_voie+(v-1)))];
     AffCar(car_LED);
     AttrTexte:=16*or_sombre+couleur_voie[premiere_voie+(v-1)=voie_courante];
     AffChE(EntierTexte2(premiere_voie+(v-1),2),' ',largeur-1-((largeur-3) div 2));
   end;

   if nombre_voies_vis<>5 then
    begin AttrTexte:=16*or_sombre+or_clair; AffCh(' '); end;

   if premiere_voie>1 then
    AffChXY(0,xg_fen_infos_mus+1+5,yh_fen_infos_mus+1,'',16*or_sombre+blanc);
   if premiere_voie+nombre_voies_vis-1<n_voies then
    AffChXY(0,xd_fen_infos_mus-4  ,yh_fen_infos_mus+1,'',16*or_sombre+blanc);
 end;
end; {InitAffPart}

{----------------------------------------------------------}

procedure AffLignePartition(p_note,y: BYTE; coul: type_coul);

var flag,v,long_ligne: BYTE;
    p: ^T_Note;
    pv: WordPtr;

    instr   : STRING[2];
    note_oct: STRING[4];
    vol     : STRING[2];
    comm    : STRING[6];
    param   : STRING[2];

begin
 with MUS,Info^,partition do begin
   flag:=1 shl (p_note and $07);
   if (ligne_aff[p_note shr 3] and flag)=0 then begin {teste si ligne dja affiche}
     {utilise la page 1 comme tampon}
     AdrTexte:=PtrVRAM(1,1,p_note+1);

     if nombre_voies_vis<>5 then
      begin AttrTexte:=16*or_fond+or_sombre; AffCar(''); end;

     AttrTexte:=16*or_fond+vert_pastel;
     if p_note=0 then AffCh('   ')
     else begin
       AffCh(EntierTexte2(p_note,3));
       p:=Ptr(Segment_Pattern,sizeof(T_Note)*((premiere_voie-1)+H^.Nb_de_voies_patt*(p_note-1)));
     end;

     for v:=1 to nombre_voies_vis do begin
       if (p_note<>0) and (v<=H^.Nb_de_voies_patt) then with p^ do begin
         if Instrument<>0 then instr:=EntierTexte2(Instrument,2) else instr:='';

         case Note of
          1..Max_Notes: note_oct:=NomNote[LangueNote,(Note-1) mod 12]+CHAR(BYTE('0')+((Note-1) div 12));
          Note_KeyOff : note_oct:=Aff_KeyOff;
          else          note_oct:='';
         end;

         if Volume<=$80 then vol:=EntierHexa(Volume,2) else vol:='';

         if Commande in [Arpeggio..Release] then begin
           case nombre_voies_vis of
            4,5,8: comm:=NomCommande6[Commande];
            6,9  : comm:=NomCommande5[Commande];
           end;
           param:=EntierHexa(Parametre,2);
         end
         else begin
           comm :='';
           param:='';
         end;
       end
       else begin
         instr   :='';
         note_oct:='';
         vol     :='';
         comm    :='';
         param   :='';
       end;

       AttrTexte:=16*or_fond+or_sombre; AffCar('');
       AttrTexte:=16*or_fond+or_clair;

       case nombre_voies_vis of
        4: begin
                                          AffChE(instr   ,' ',2);
            AttrTexte:=16*or_fond+noir  ; AffChE(note_oct,' ',4+1);
            AttrTexte:=16*or_fond+cyan  ; AffChE(vol     ,' ',2);
            AttrTexte:=16*or_fond+orange; AffChE(comm    ,' ',6);
            AttrTexte:=16*or_fond+gris  ; AffChE(param   ,' ',2);
           end;

        5: begin
                                          AffChE(instr   ,' ',2);
            AttrTexte:=16*or_fond+noir  ; AffChE(note_oct,' ',4);
            AttrTexte:=16*or_fond+orange; AffChE(comm    ,' ',6);
            AttrTexte:=16*or_fond+gris  ; AffChE(param   ,' ',2);
           end;

        6: begin
                                          AffChE(instr   ,' ',2);
            AttrTexte:=16*or_fond+noir  ; AffChE(note_oct,' ',4);
            AttrTexte:=16*or_fond+orange; AffChE(comm    ,' ',5);
           end;

        8: if comm='' then begin
                                           AffChE(instr   ,' ',2+1);
             AttrTexte:=16*or_fond+noir  ; AffChE(note_oct,' ',4+1);
           end
           else begin
             AttrTexte:=16*or_fond+orange; AffChE(comm    ,' ',6);
             AttrTexte:=16*or_fond+gris  ; AffChE(param   ,' ',2);
           end;

        9: if comm='' then begin
                                           AffChE(instr   ,' ',2+1);
             AttrTexte:=16*or_fond+noir  ; AffChE(note_oct,' ',4);
           end
           else begin
             AttrTexte:=16*or_fond+orange; AffChE(comm    ,' ',5);
             AttrTexte:=16*or_fond+gris  ; AffChE(param   ,' ',2);
           end;
       end;

       inc(p);
     end;

     if nombre_voies_vis<>5 then
      begin AttrTexte:=16*or_fond+or_sombre; AffCar(''); end;

     ligne_aff[p_note shr 3]:=ligne_aff[p_note shr 3] or flag; {met le flag  1}
   end;

   {copie la ligne de la page 1 -> page 0}
   if partition.nombre_voies_vis<>5 then long_ligne:=77 else long_ligne:=78;
   pv:=WordPtr(PtrVRAM(0,xg_fen_infos_mus+1,y));
   Move(PtrVRAM(1,1,p_note+1)^,pv^,2*long_ligne);

   {change la couleur de fond de la ligne active (position note courante)}
   if coul=inverse then
    asm
     les DI,pv
     mov BL,long_ligne
    @change_fond:
     and BYTE(ES:[DI+1]),00001111b
     or  BYTE(ES:[DI+1]),16*or_moyen
     add DI,2
     dec BL
     jnz @change_fond
    end;
 end;
end; {AffLignePartition}

{----------------------------------------------------------}

procedure AffPart; far;

var i,pos_note_max: BYTE;
    p_note: SHORTINT;

begin
 with Info^,partition do begin
   if (anc_pos_note<>Pos_Note) or (anc_pattern<>Num_Pattern) or (reaffiche>=reaff_titre) then begin
     anc_pos_note:=Pos_Note;
     if (anc_pattern<>Num_Pattern) or (reaffiche>=reaff_titre) then begin
       anc_pattern:=Num_Pattern;
       FillChar(ligne_aff,sizeof(ligne_aff),$00);
     end;

     pos_note_max:=MUS.T^.Taille_Pattern[anc_pattern+1];

     for i:=1 to 31 do begin
       p_note:=(i-16)+(anc_pos_note+1);
       if not(p_note in [1..pos_note_max]) then p_note:=0; {ligne vide}

       AffLignePartition(p_note,yh_fen_infos_mus+2+(i-1),type_coul(i=16));
     end;

     if nombre_voies_vis<>5 then
      AffAscenseur((1-16)+(anc_pos_note+1)>1,(31-16)+(anc_pos_note+1)<pos_note_max);
   end;
 end;
end; {AffPart}

{----------------------------------------------------------}

procedure TesteTouchePart(touche: Key); far;

var n_voies,v: BYTE;

begin
 n_voies:=MUS.H^.Nb_de_voies_patt;

 with partition,Info^ do
  case touche of
   Left : if voie_courante>1 then begin
            dec(voie_courante);
            if voie_courante<premiere_voie then dec(premiere_voie);
            reaffiche:=reaff_titre;
          end;

   Right: if voie_courante<n_voies then begin
            inc(voie_courante);
            if voie_courante>premiere_voie+nombre_voies_vis-1 then inc(premiere_voie);
            reaffiche:=reaff_titre;
          end;

   PgUp: begin
          if voie_courante>nombre_voies_vis then begin
            dec(voie_courante,nombre_voies_vis);
            if voie_courante<premiere_voie then premiere_voie:=voie_courante;
          end
          else begin
            voie_courante:=1;
            premiere_voie:=1;
          end;
          reaffiche:=reaff_titre;
         end;

   PgDn: begin
          inc(voie_courante,nombre_voies_vis);
          if voie_courante>n_voies then voie_courante:=n_voies;
          if n_voies>nombre_voies_vis then
           if premiere_voie<voie_courante-nombre_voies_vis+1 then premiere_voie:=voie_courante-nombre_voies_vis+1;
          reaffiche:=reaff_titre;
         end;

   Home: if voie_courante>1 then begin
           voie_courante:=1;
           premiere_voie:=1;
           reaffiche:=reaff_titre;
         end;

   End_: if voie_courante<n_voies then begin
           voie_courante:=n_voies;
           if n_voies>nombre_voies_vis then premiere_voie:=n_voies-nombre_voies_vis+1;
           reaffiche:=reaff_titre;
         end;

   Tab: begin
         case nombre_voies_vis of
          4: nombre_voies_vis:=5;
          5: nombre_voies_vis:=6;
          6: nombre_voies_vis:=8;
          8: nombre_voies_vis:=9;
          9: nombre_voies_vis:=4;
         end;
         voie_courante:=1;
         premiere_voie:=1;
         reaffiche:=reaff_titre;
        end;

   ShiftTab: begin
              case nombre_voies_vis of
               4: nombre_voies_vis:=9;
               5: nombre_voies_vis:=4;
               6: nombre_voies_vis:=5;
               8: nombre_voies_vis:=6;
               9: nombre_voies_vis:=8;
              end;
              voie_courante:=1;
              premiere_voie:=1;
              reaffiche:=reaff_titre;
             end;

   {active/dsactive la voie courante}
   Enter: begin
           if Voie_Active(voie_courante) then Desactiver_Voie(voie_courante) else Activer_Voie(voie_courante);
           reaffiche:=reaff_titre;
          end;

   {voie courante en solo/active toutes les voies}
   CtrlEnter: begin
               VoieSolo(voie_courante);
               reaffiche:=reaff_titre;
              end;

   Key('N'): begin
              ChangeLangueNote;
              reaffiche:=reaff_titre;
             end;
  end;
end; {TesteTouchePart}

{----------------------------------------------------------}
{ GESTION DES CLAVIERS                                     }
{----------------------------------------------------------}

procedure InitStrucClav; far;

var s: BYTE;

begin
 with claviers do begin
   premiere_octave:=3; {octave de base des .mod}

   nombre_instr:=0;
   FillChar(num_instr,sizeof(num_instr),1);

   for s:=1 to liste_instr.nombre do
    if (liste_instr.etat[s] and smp_charge)=smp_charge then begin
      inc(nombre_instr);
      num_instr[nombre_instr]:=s;
    end;

   premier_instr:=1;
   instr_courant:=1;
 end;
end; {InitStrucClav}

{----------------------------------------------------------}

procedure InitAffClav; far;

var o: BYTE;

begin
 Cadre(0,1,xg_fen_infos_mus+1,yh_fen_infos_mus+1,xd_fen_infos_mus-1,yb_fen_infos_mus-1,
       16*or_fond+or_sombre,'',0,'');

 AdrTexte :=PtrVRAM(0,xg_fen_infos_mus+1,yh_fen_infos_mus+1);
 AttrTexte:=16*or_sombre+or_clair; AffCh('     '); AffCar(car_barre_d);

 with claviers do begin
   for o:=1 to 5 do begin
     AffChE(car_barre_g,' ',7);
     AffChE(CHAR(BYTE('0')+premiere_octave+(o-1)),' ',6);
     AffCar(car_barre_d);
   end;
   AffChE(car_barre_g,' ',3);

   if premiere_octave>0 then
    AffChXY(0,xg_fen_infos_mus+1+7,yh_fen_infos_mus+1,'',16*or_sombre+blanc);
   if premiere_octave+5-1<Max_Octaves-1 then
    AffChXY(0,xd_fen_infos_mus-5  ,yh_fen_infos_mus+1,'',16*or_sombre+blanc);
 end;
end; {InitAffClav}

{----------------------------------------------------------}

procedure AffOctaveClavier(x,y: BYTE; masque_notes: WORD);
{ _____________
        
        
        
 ;
 DoRMiFaSoLaSi  }

var touche: array[Do_..Si] of BOOLEAN;
    n,coul_fond,coul_txt: BYTE;
    p: WordPtr;

begin
 for n:=Do_ to Si do touche[n]:=(masque_notes and (1 shl n))<>0;

 AffChXY(0,x,y,'______________',16*or_fond+noir);

 {utilise la page 1 comme tampon}
 AdrTexte :=PtrVRAM(1,x,y+1); p:=WordPtr(AdrTexte);

 coul_fond:=16*coul_clav_b[touche[Do_]];
 AttrTexte:=coul_fond+noir    ; AffCar(car_clav_b_hg);
 coul_txt :=coul_clav_n[touche[Dod]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hg);
 coul_fond:=16*coul_clav_b[touche[Re ]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hd);
 coul_txt :=coul_clav_n[touche[Mib]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hg);
 coul_fond:=16*coul_clav_b[touche[Mi ]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hd);
 AttrTexte:=coul_fond+noir    ; AffCar(car_clav_b_hd);
 coul_fond:=16*coul_clav_b[touche[Fa ]];
 AttrTexte:=coul_fond+noir    ; AffCar(car_clav_b_hg);
 coul_txt :=coul_clav_n[touche[Fad]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hg);
 coul_fond:=16*coul_clav_b[touche[Sol]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hd);
 coul_txt :=coul_clav_n[touche[Sod]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hg);
 coul_fond:=16*coul_clav_b[touche[La ]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hd);
 coul_txt :=coul_clav_n[touche[Sib]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hg);
 coul_fond:=16*coul_clav_b[touche[Si ]];
 AttrTexte:=coul_fond+coul_txt; AffCar(car_clav_n_hd);
 AttrTexte:=coul_fond+noir    ; AffCar(car_clav_b_hd);

 {copie ligne 1 -> ligne 2 en modifiant les caractres des touches noires}
 asm
  les SI,p
  mov BX,2*80
  mov CL,2*7
 @copie_ligne:
  mov AX,ES:[SI]
  cmp AL,car_clav_n_hg
  jne @suite1
  mov AL,car_clav_n_bg
  jmp @copie_car
 @suite1:
  cmp AL,car_clav_n_hd
  jne @copie_car
  mov AL,car_clav_n_bd
 @copie_car:
  mov ES:[SI+BX],AX
  add SI,2
  dec CL
  jnz @copie_ligne
 end;

 AdrTexte :=PtrVRAM(1,x,y+3); p:=WordPtr(AdrTexte);

 AttrTexte:=16*coul_clav_b[touche[Do_]]+noir; AffCh(car_clav_b_hg+car_clav_b_hd);
 AttrTexte:=16*coul_clav_b[touche[Re ]]+noir; AffCh(car_clav_b_hg+car_clav_b_hd);
 AttrTexte:=16*coul_clav_b[touche[Mi ]]+noir; AffCh(car_clav_b_hg+car_clav_b_hd);
 AttrTexte:=16*coul_clav_b[touche[Fa ]]+noir; AffCh(car_clav_b_hg+car_clav_b_hd);
 AttrTexte:=16*coul_clav_b[touche[Sol]]+noir; AffCh(car_clav_b_hg+car_clav_b_hd);
 AttrTexte:=16*coul_clav_b[touche[La ]]+noir; AffCh(car_clav_b_hg+car_clav_b_hd);
 AttrTexte:=16*coul_clav_b[touche[Si ]]+noir; AffCh(car_clav_b_hg+car_clav_b_hd);

 {copie ligne 3 -> ligne 4 en modifiant les caractres des touches blanches}
 asm
  les SI,p
  mov BX,2*80
  mov CL,2*7
 @copie_ligne:
  mov AX,ES:[SI]
  cmp AL,car_clav_b_hg
  jne @suite1
  mov AL,car_clav_b_bg
  jmp @copie_car
 @suite1:
  cmp AL,car_clav_b_hd
  jne @copie_car
  mov AL,car_clav_b_bd
 @copie_car:
  mov ES:[SI+BX],AX
  add SI,2
  dec CL
  jnz @copie_ligne
 end;

 {copie les 4 lignes de la page 1 -> page 0}
 for n:=1 to 4 do Move(PtrVRAM(1,x,y+n)^,PtrVRAM(0,x,y+n)^,2*(2*7));
end; {AffOctaveClavier}

{----------------------------------------------------------}

procedure AffClav; far;

var i,i_instr,n_smp,v,x,y: BYTE;
    coul_fond,coul_note: BYTE;
    coul: type_coul;

    n_debut,n_fin,n_milieu,anc_n_milieu: BYTE;
    o,octave_accord,base_accord,note_basse,nombre_notes_accord,
    note,nombre_notes,numero_accord,a,b: BYTE;
    masque_notes: array[0..Max_Octaves-1] of WORD;
    masque_accord,masque_base,msq: WORD;
    accord_trouve: BOOLEAN;
    nom_accord: STRING[5];

    periode_reelle: WORD;

begin
 with claviers,liste_instr,Info^,MUS do begin
   y:=yh_fen_infos_mus+2;

   for i:=1 to 6 do if i<=nombre_instr then begin
     i_instr:=premier_instr+(i-1);
     n_smp:=num_instr[i_instr]; {n_smp = n d'instrument rel}

     FillChar(masque_notes,sizeof(masque_notes),$000);

     {initialise octave et note relative la plus basse}
     octave_accord:=Max_Octaves-1;
     base_accord:=Si;

     {recherche l'instrument n_smp actif sur les voies}
     for v:=1 to H^.Nb_de_voies_patt do
      if ((VSample[v]+1)=n_smp) and ((VControle[v] and M_Pas_Active)=0) then begin
        periode_reelle:=VPeriode_re[v]; {priode relle de l'instrument actif (tient compte des effets)}
        if (periode_reelle<=PeriodArray[0]) and (periode_reelle>=PeriodArray[Max_Notes-1]) then begin
          periode_reelle:=2*periode_reelle; {*2 car PeriodArray[] est fait comme a!}

          {recherche par dichotomie la note la plus proche}
          n_debut:=0; n_fin:=Max_Notes-1; n_milieu:=0;
          repeat
           anc_n_milieu:=n_milieu;
           n_milieu:=(n_debut+n_fin) div 2;
           if PeriodArray[n_milieu]<=periode_reelle then n_fin:=n_milieu else n_debut:=n_milieu;
          until n_milieu=anc_n_milieu;

          {correction}
          if periode_reelle<(PeriodArray[n_milieu]+PeriodArray[n_milieu+1]) div 2 then inc(n_milieu)
          else
          if periode_reelle>(PeriodArray[n_milieu-1]+PeriodArray[n_milieu]) div 2 then dec(n_milieu);

          o:=(n_milieu div 12)+1; {n octave}
          b:=n_milieu mod 12; {n note}

          masque_notes[o]:=masque_notes[o] or (1 shl b); {active la note}

          if o<=octave_accord then begin {recherche la note relative la plus basse}
            octave_accord:=o;
            if b<base_accord then base_accord:=b;
          end;
        end;
      end;

     note_basse:=base_accord;

     {compose l'accord en ramenant toutes les notes dans une seule octave}
     masque_accord:=$000;
     for o:=0 to Max_Octaves-1 do masque_accord:=masque_accord or masque_notes[o];

     {compte le nombre de notes dans l'accord}
     asm
      mov AX,masque_accord
      xor BL,BL
      mov CL,12

     @compte_bits:
      shr AX,1
      adc BL,0
      dec CL
      jnz @compte_bits

      mov nombre_notes_accord,BL
     end;

     accord_trouve:=FALSE;

     {vrifie qu'un accord peut exister (au moins 3 notes)}
     if nombre_notes_accord>=3 then
      {recherche l'accord avec la note relative la plus basse comme note de base}
      for b:=1 to 12 do begin
        {ajuste le masque accord avec la note de base (par rotation  droite sur 12 bits)
         -> compare avec un accord en Do}
        asm
         mov AX,masque_accord
         mov CH,base_accord
         or  CH,CH
         jz  @fin

        @rotation_masque:
         mov BL,AL
         and BL,00000001b
         ror AX,1
         and AX,0000011111111111b
         shl BL,3
         or  AH,BL
         dec CH
         jnz @rotation_masque

        @fin:
         mov masque_base,AX
        end;

        {commence par rechercher les accords les plus longs}
        if (masque_base and Do_)=Do_ then
         for a:=nbr_accords downto 1 do with Accord[a] do
          if (masque_base and masque)=masque then begin
            {compte le nombre de notes dans l'accord a}
            msq:=masque;
            asm
             mov AX,msq
             xor BL,BL
             mov CL,12

            @compte_bits:
             shr AX,1
             adc BL,0
             dec CL
             jnz @compte_bits

             mov nombre_notes,BL
            end;

            accord_trouve:=(nombre_notes>=nombre_notes_accord);
            numero_accord:=a;
            break;
          end;

        if accord_trouve then break;

        inc(base_accord); {recherche l'accord avec la note de base augmente d'un demi-ton}
        if base_accord>Si then begin inc(octave_accord); base_accord:=Do_; end;
      end;

     coul:=type_coul(i_instr=instr_courant);
     coul_fond:=16*couleur_fond[coul];

     if (etat[n_smp] and smp_joue)=smp_joue then coul_note:=jaune
     else
     if (etat[n_smp] and smp_actif)=smp_actif then coul_note:=gris
     else
     if (etat[n_smp] and smp_dejajoue)=smp_dejajoue then coul_note:=or_sombre
     else coul_note:=couleur_fond[coul];

     AdrTexte :=PtrVRAM(0,xg_fen_infos_mus+2,y+1);
     AttrTexte:=coul_fond+couleur_curs[coul]; AffCh(EntierTexte2(num_instr[premier_instr+(i-1)],2));
     AttrTexte:=coul_fond+couleur_LED_vr[TRUE,(etat[n_smp] and smp_stoppe)=smp_stoppe]; AffCar(car_LED);
     AttrTexte:=coul_fond+coul_note; AffCar('');

     AttrTexte:=16*or_fond+blanc;

     if accord_trouve then begin
      nom_accord:=NomNote[LangueNote,base_accord];
      if nom_accord[Length(nom_accord)] in [' ','-'] then dec(nom_accord[0]);
     end else nom_accord:='';
     AdrTexte :=PtrVRAM(0,xg_fen_infos_mus+2,y+2); AffChE(nom_accord,' ',3); {base accord}

     if accord_trouve then nom_accord:=Accord[numero_accord].nom else nom_accord:='';

     AdrTexte :=PtrVRAM(0,xg_fen_infos_mus+2,y+3); AffChE(nom_accord,' ',5); {nom accord}

     if accord_trouve and (note_basse<>base_accord) then begin
      nom_accord:=NomNote[LangueNote,note_basse];
      if nom_accord[Length(nom_accord)] in [' ','-'] then dec(nom_accord[0]);
     end else nom_accord:='';

     AdrTexte :=PtrVRAM(0,xg_fen_infos_mus+2,y+4); AffChE(nom_accord,' ',3); {note basse}

     x:=xg_fen_infos_mus+7;
     for o:=1 to 5 do begin
       AffOctaveClavier(x,y,masque_notes[premiere_octave+(o-1)]);
       inc(x,2*7);
     end;

     inc(y,5);
   end;

   if nombre_instr>6 then AffAscenseur(premier_instr>1,premier_instr+6-1<nombre_instr);
 end;
end; {AffClav}

{----------------------------------------------------------}

procedure TesteToucheClav(touche: Key); far;
begin
 with claviers do
  case touche of
   Up  : if instr_courant>1 then begin
           dec(instr_courant);
           if instr_courant<premier_instr then dec(premier_instr);
           reaffiche:=reaff_lignes;
         end;

   Down: if instr_courant<nombre_instr then begin
           inc(instr_courant);
           if instr_courant>premier_instr+6-1 then inc(premier_instr);
           reaffiche:=reaff_lignes;
         end;

   PgUp: begin
          if instr_courant>6 then begin
            dec(instr_courant,6);
            if instr_courant<premier_instr then premier_instr:=instr_courant;
          end
          else begin
            instr_courant:=1;
            premier_instr:=1;
          end;
          reaffiche:=reaff_lignes;
         end;

   PgDn: begin
          inc(instr_courant,6);
          if instr_courant>nombre_instr then instr_courant:=nombre_instr;
          if nombre_instr>6 then
           if premier_instr<instr_courant-6+1 then premier_instr:=instr_courant-6+1;
          reaffiche:=reaff_lignes;
         end;

   Home: if instr_courant>1 then begin
           instr_courant:=1;
           premier_instr:=1;
           reaffiche:=reaff_lignes;
         end;

   End_: if instr_courant<nombre_instr then begin
           instr_courant:=nombre_instr;
           if nombre_instr>6 then premier_instr:=nombre_instr-6+1;
           reaffiche:=reaff_lignes;
         end;

   Left : if premiere_octave>0 then begin
            dec(premiere_octave);
            reaffiche:=reaff_titre;
          end;

   Right: if premiere_octave+5-1<Max_Octaves-1 then begin
            inc(premiere_octave);
            reaffiche:=reaff_titre;
          end;

   {active/dsactive l'instrument courant}
   Enter: if ActiveDesactiveInstr(num_instr[instr_courant]) then reaffiche:=reaff_lignes;

   Key('N'): begin
              ChangeLangueNote;
              reaffiche:=reaff_titre;
             end;
  end;
end; {TesteToucheClav}

{----------------------------------------------------------}
{ GESTION DU SPECTRE                                       }
{----------------------------------------------------------}

procedure InitStrucSpect; far;
begin
 with spectre do begin
   mode_spectre:=mode_spectre_;

   {contrle de validit}
   if (mode_spectre=spectre_stereo)
      and not(Variables^.Utilise_Stereo) then mode_spectre:=spectre_mono;

   premiere_frequence:=0;

   amplification:=amplification_;
 end;
end; {InitStrucSpect}

{----------------------------------------------------------}

procedure AffGraduationsHoriz(x,y: BYTE);
var i: BYTE;
begin
 AdrTexte :=PtrVRAM(0,x,y);
 AttrTexte:=16*or_fond+noir;
 for i:=0 to largeur_spectre-1 do AffCar(car_grad_h[(i mod 5)=0]);
end; {AffGraduationsHoriz}

{----------------------------------------------------------}

procedure AffGraduationsVert(x,y,hauteur: BYTE);
var i: BYTE;
    car: CHAR;
begin
 AttrTexte:=16*or_fond+noir;
 for i:=0 to hauteur-1 do begin
   AdrTexte:=PtrVRAM(0,x,y);
   car:=car_grad_v[(i and $03)=0];
   AffCar(car); inc(AdrTexte,largeur_spectre); AffCar(car);
   dec(y);
 end;
end; {AffGraduationsHoriz}

{----------------------------------------------------------}

procedure AffEchelleFreq(x,y: BYTE);
var i: BYTE;
begin
 with spectre do begin
   EffaceZone(0,x-2,y,x+largeur_spectre,y,16*or_clair);

   for i:=0 to (largeur_spectre div 10) do
    AffChXY(0,x-1+10*i,y,ReelTexte(premiere_frequence/1000+0.5*i,1,1),16*or_clair+or_sombre);

   AffChXY(0,x+largeur_spectre-1,y,'kHz',16*or_clair+or_sombre);
 end;
end; {AffEchelleFreq}

{----------------------------------------------------------}

procedure InitAffSpect; far;
var fleches: STRING[2];
begin
 with spectre do begin
   case mode_spectre of
    spectre_mono  : begin
                     {affiche les graduations horizontales}
                     AffGraduationsHoriz(xg_spectre,yb_spectre+1);

                     {affiche les graduations verticales}
                     AffGraduationsVert(xg_spectre-1,yb_spectre,hauteur_spectre_mono);

                     {affiche l'chelle des frquences}
                     AffEchelleFreq(xg_spectre,yb_spectre+2);
                    end;

    spectre_stereo: begin
                     {affiche les symboles HP}
                     AffCarXY(0,x_HP_spectre_g,y_HP_spectre_g,car_HP_g,16*or_fond+or_clair);
                     AffCarXY(0,x_HP_spectre_d,y_HP_spectre_d,car_HP_d,16*or_fond+or_clair);

                     {affiche les graduations horizontales}
                     AffGraduationsHoriz(xg_spectre_g,yb_spectre_g+1);
                     AffGraduationsHoriz(xg_spectre_d,yb_spectre_d+1);

                     {affiche les graduations verticales}
                     AffGraduationsVert(xg_spectre_g-1,yb_spectre_g,hauteur_spectre_stereo);
                     AffGraduationsVert(xg_spectre_d-1,yb_spectre_d,hauteur_spectre_stereo);

                     {affiche l'chelle des frquences}
                     AffEchelleFreq(xg_spectre_d,yb_spectre_d+2);
                    end;
   end;

   if premiere_frequence>=pas_frequence then fleches:='' else fleches:='  ';
   AffChXY(0,xg_fen_infos_mus+1,yb_fen_infos_mus-2,fleches,16*or_fond+blanc);
   if premiere_frequence<bande_spectre_FFT-pas_frequence*((largeur_spectre div 10)+1) then fleches:='' else fleches:='  ';
   AffChXY(0,xd_fen_infos_mus-2,yb_fen_infos_mus-2,fleches,16*or_fond+blanc);
 end;
end; {InitAffSpect}

{----------------------------------------------------------}

procedure AffBarresSpectreMono(x,y: BYTE; spectre: T_FFT_module; decalage: BYTE); assembler;
asm
 push DS

 mov DX,0B800h
 mov ES,DX

 xor DH,DH
 mov DL,largeur_ecran_txt

 xor AH,AH
 mov AL,x
 mov DI,AX
 dec DI
 mov AL,y
 dec AL
 mul DL
 add DI,AX
 shl DI,1            {ES:DI -> B800:2*((x-1)+largeur_ecran_txt*(y-1))}

 shl DX,1            {DX=2*largeur_ecran_txt}

 lds SI,spectre

 xor BH,BH
 mov BL,decalage
 shl BX,1
 add SI,BX           {DS:SI -> spectre[decalage]}

 mov CH,largeur_spectre
@boucle_barres:
  push DI            {sauvegarde DI (=@bas de la barre)}

  mov AL,car_barre_spectre

  mov CL,[SI]
  or  CL,CL
  jz  @barre_gris
  xor BX,BX
 @boucle_barre_couleur:
   mov AH,BYTE(CS:[@couleurs+BX])
   mov WORD(ES:[DI]),AX
   sub DI,DX
   inc BL
   dec CL
  jnz @boucle_barre_couleur

 @barre_gris:
  mov CL,hauteur_spectre_mono
  sub CL,[SI]
  or  CL,CL
  jz  @fin_barre
  mov AH,gris_fonce
 @boucle_barre_gris:
   mov WORD(ES:[DI]),AX
   sub DI,DX
   dec CL
  jnz @boucle_barre_gris

 @fin_barre:
  add SI,2
  pop DI             {rcupre DI}
  add DI,2           {DI=@bas de la barre suivante}
  dec CH
 jnz @boucle_barres

 jmp @fin

@couleurs:
 DB vert,vert,vert,vert,vert,vert,vert,vert
 DB vert,vert,vert,vert,vert,vert,vert                       { 15 }
 DB orange,orange,orange,orange,orange,orange,orange,orange  {  8 }
 DB rouge,rouge,rouge,rouge,rouge,rouge                      {  6 }
                               { TOTAL: hauteur_spectre_mono = 29 }

@fin:
 pop DS
end; {AffBarresSpectreMono}

{----------------------------------------------------------}

procedure AffBarresSpectreStereo(x,y: BYTE; spectre: T_FFT_module; decalage: BYTE); assembler;
asm
 push DS

 mov DX,0B800h
 mov ES,DX

 xor DH,DH
 mov DL,largeur_ecran_txt

 xor AH,AH
 mov AL,x
 mov DI,AX
 dec DI
 mov AL,y
 dec AL
 mul DL
 add DI,AX
 shl DI,1            {ES:DI -> B800:2*((x-1)+largeur_ecran_txt*(y-1))}

 shl DX,1            {DX=2*largeur_ecran_txt}

 lds SI,spectre

 xor BH,BH
 mov BL,decalage
 shl BX,1
 add SI,BX           {DS:SI -> spectre[decalage]}

 mov CH,largeur_spectre
@boucle_barres:
  push DI            {sauvegarde DI (=@bas de la barre)}

  mov AL,car_barre_spectre

  mov CL,[SI]
  or  CL,CL
  jz  @barre_gris
  xor BX,BX
 @boucle_barre_couleur:
   mov AH,BYTE(CS:[@couleurs+BX])
   mov WORD(ES:[DI]),AX
   sub DI,DX
   inc BL
   dec CL
  jnz @boucle_barre_couleur

 @barre_gris:
  mov CL,hauteur_spectre_stereo
  sub CL,[SI]
  or  CL,CL
  jz  @fin_barre
  mov AH,gris_fonce
 @boucle_barre_gris:
   mov WORD(ES:[DI]),AX
   sub DI,DX
   dec CL
  jnz @boucle_barre_gris

 @fin_barre:
  add SI,2
  pop DI             {rcupre DI}
  add DI,2           {DI=@bas de la barre suivante}
  dec CH
 jnz @boucle_barres

 jmp @fin

@couleurs:
 DB vert,vert,vert,vert,vert,vert,vert  { 7 }
 DB orange,orange,orange,orange         { 4 }
 DB rouge,rouge,rouge                   { 3 }
       { TOTAL: hauteur_spectre_stereo = 14 }

@fin:
 pop DS
end; {AffBarresSpectreStereo}

{----------------------------------------------------------}

procedure AffSpect; far;

var onde: record case BYTE of
           1: (mono: array[0..taille_tampon_FFT-1] of SHORTINT);
           2: (gauche,droite: array[0..taille_tampon_FFT-1] of SHORTINT);
          end;

    tampon_FFT: record case BYTE of
                 1: (mono: T_FFT);
                 2: (gauche,droite: T_FFT);
                end;

    i,position: WORD;

begin
 with spectre,tampon_FFT,Variables^,Info^ do
  case mode_spectre of
   spectre_mono  :
     begin
      {chantillonne le tampon DMA  taille_tampon_FFT points}
      EchantTampon(Mem[Seg_Table_Volume:Pointeur_Tampon_Joue],
                   onde.mono,Taille_Tampon,taille_tampon_FFT);

      {convertit en 8 bits signs}
      ConvSigne(onde.mono,taille_tampon_FFT);

      {convertit 8 bits -> 16 bits en amplifiant l'onde}
      Extension16(onde.mono,mono.points,taille_tampon_FFT,amplification);

      {calcule la FFT}
      FFT(mode_FFT,mono,log2_taille_tampon_FFT);

      {normalise et crte les modules}
      for i:=0 to (taille_tampon_FFT div 2)-1 do
       mono.module[i]:=MinEntier((mono.module[i]*hauteur_spectre_mono) div valeur_max,hauteur_spectre_mono);

      {affiche les barres de spectre}
      AffBarresSpectreMono(xg_spectre,yb_spectre,mono.module,premiere_frequence div intervalle_freq_FFT);
     end;

   spectre_stereo:
     begin
      {chantillonne le tampon DMA  taille_tampon_FFT points}
      EchantTamponStereo(Mem[Seg_Table_Volume:Pointeur_Tampon_Joue],
                         onde.gauche,onde.droite,Taille_Tampon,taille_tampon_FFT);

      {convertit en 8 bits signs}
      ConvSigne(onde.gauche,taille_tampon_FFT);
      ConvSigne(onde.droite,taille_tampon_FFT);

      {convertit 8 bits -> 16 bits en amplifiant l'onde}
      Extension16(onde.gauche,gauche.points,taille_tampon_FFT,amplification);
      Extension16(onde.droite,droite.points,taille_tampon_FFT,amplification);

      {calcule les FFT gauche & droite}
      FFT(mode_FFT,gauche,log2_taille_tampon_FFT);
      FFT(mode_FFT,droite,log2_taille_tampon_FFT);

      {normalise et crte les modules}
      for i:=0 to (taille_tampon_FFT div 2)-1 do begin
        gauche.module[i]:=MinEntier((gauche.module[i]*hauteur_spectre_stereo) div valeur_max,hauteur_spectre_stereo);
        droite.module[i]:=MinEntier((droite.module[i]*hauteur_spectre_stereo) div valeur_max,hauteur_spectre_stereo);
      end;

      {affiche les barres de spectre}
      AffBarresSpectreStereo(xg_spectre_g,yb_spectre_g,gauche.module,premiere_frequence div intervalle_freq_FFT);
      AffBarresSpectreStereo(xg_spectre_d,yb_spectre_d,droite.module,premiere_frequence div intervalle_freq_FFT);
     end;
  end;
end; {AffSpect}

{----------------------------------------------------------}

procedure TesteToucheSpect(touche: Key); far;
begin
 with spectre do
  case touche of
   Left : if premiere_frequence>=pas_frequence then begin
            dec(premiere_frequence,pas_frequence);
            reaffiche:=reaff_titre;
          end;

   Right: if premiere_frequence<bande_spectre_FFT-pas_frequence*((largeur_spectre div 10)+1) then begin
            inc(premiere_frequence,pas_frequence);
            reaffiche:=reaff_titre;
          end;

   PgUp,Home: begin
               premiere_frequence:=0;
               reaffiche:=reaff_titre;
              end;

   PgDn,End_: begin
               while premiere_frequence<bande_spectre_FFT-pas_frequence*((largeur_spectre div 10)+1) do
                inc(premiere_frequence,pas_frequence);
               reaffiche:=reaff_titre;
              end;

   Up  : if amplification<ampl_spect_max then inc(amplification);

   Down: if amplification>ampl_spect_min then dec(amplification);

   Tab: begin
         case mode_spectre of
          spectre_mono  : if Variables^.Utilise_Stereo then mode_spectre:=spectre_stereo;
          spectre_stereo: mode_spectre:=spectre_mono;
         end;
         reaffiche:=reaff_fen;
        end;
  end;
end; {TesteToucheSpect}

{----------------------------------------------------------}
{ GESTION GENERALE DE L'AFFICHAGE                          }
{----------------------------------------------------------}

CONST {groupe de procdures pour la gestion de la fentre infrieure}
      FenInf: array[aff_aide..aff_spect] of record
                InitStruc  : procedure; {initialisation des structures d'affichage}
                InitAff    : procedure; {initialisation de l'affichage (excut une seule fois aprs changement d'affichage)}
                Aff        : procedure; {affichage principal (excut priodiquement)}
                TesteTouche: procedure(touche: Key); {gestion du clavier}
              end

            =((InitStruc:InitStrucAide ; InitAff:InitAffAide ; Aff:AffAide ; TesteTouche:TesteToucheAide ),
              (InitStruc:InitStrucInstr; InitAff:InitAffInstr; Aff:AffInstr; TesteTouche:TesteToucheInstr),
              (InitStruc:InitStrucVoies; InitAff:InitAffVoies; Aff:AffVoies; TesteTouche:TesteToucheVoies),
              (InitStruc:InitStrucPart ; InitAff:InitAffPart ; Aff:AffPart ; TesteTouche:TesteTouchePart ),
              (InitStruc:InitStrucClav ; InitAff:InitAffClav ; Aff:AffClav ; TesteTouche:TesteToucheClav ),
              (InitStruc:InitStrucSpect; InitAff:InitAffSpect; Aff:AffSpect; TesteTouche:TesteToucheSpect));

{----------------------------------------------------------}

procedure InitAffMus(reaffiche: BYTE);

var volume: WORD;
    type_tracker: STRING[27];

begin
 if reaffiche>=reaff_tout then begin
   Clignotement(clignote);
   AfficheLignesHautBas;

   AttendVBL;

   {fentre contrle musique (partie suprieure de l'cran...)}
   FenetreOmbre(xg_fen_contr_mus,yh_fen_contr_mus,xd_fen_contr_mus,yb_fen_contr_mus,encadre);
   LigneHorizontale(0,1,yh_fen_contr_mus+ 4,xg_fen_contr_mus+2,xg_fen_contr_mus+35,16*or_fond+or_sombre);
   LigneHorizontale(0,1,yh_fen_contr_mus+ 7,xg_fen_contr_mus+2,xg_fen_contr_mus+35,16*or_fond+or_sombre);
   LigneVerticale  (0,1,xg_fen_contr_mus+35,yh_fen_contr_mus+2,yh_fen_contr_mus+11,16*or_fond+or_sombre);
   LigneVerticale  (0,1,xg_fen_contr_mus+12,yh_fen_contr_mus+8,yh_fen_contr_mus+11,16*or_fond+or_sombre);

   {intersections caractres semi-graphiques}
   AffChXY(0,xg_fen_contr_mus+12,yh_fen_contr_mus+7,'',16*or_fond+or_sombre);

   {module: nom, taille, volume, titre}
   with Module[indice_module] do begin
     if (Etat and vol_par_defaut)=vol_par_defaut then volume:=Volume_par_defaut
                                                 else volume:=Volume_reel;
     AffChXY(0,x_txt_module,y_txt_module,txt_module+':',16*or_fond+or_clair);
     AffChXY(0,x_nom_module,y_nom_module,XString(Module[indice_module].Nom,' ',12),16*or_fond+gris_fonce);
     AffChXY(0,x_taille_module,y_taille_module,EntierTexte2(Module[indice_module].Taille,4)+'k',16*or_fond+gris_fonce);
     AffChXY(0,x_vol_module,y_vol_module,EntierTexte2(volume div 5,3)+'%',16*or_fond+gris_fonce);
     AffChXY(0,x_titre_module,y_titre_module,XString(MUS.H^.Titre,' ',32),16*or_fond+blanc);
   end;

   {type module, nombre de voies et nombre d'instruments}
   with MUS.H^ do begin
     AffChXY(0,x_txt_type,y_txt_type,txt_type+':',16*or_fond+or_clair);
     if Type_module=T_MOD then type_tracker:=TypeModuleMOD[Type_Module_Etendu]
                          else type_tracker:=TypeModule[Type_Module];
     AffChXY(0,x_type_module,y_type_module,XString(type_tracker,' ',27),16*or_fond+gris_fonce);
     AffChXY(0,x_voies_instr,y_voies_instr,EntierTexte2(Nb_de_voies,2)+' '+txt_voies[Langue]+' - '+
                                           EntierTexte2(Nb_Total_Samples,2)+' '+txt_instr[Langue]+'s ('+
                                           EntierTexte2(Nb_Samples,2)+')',16*or_fond+gris_fonce);
   end;

   {mmoire libre: DOS,EMS,GUS}
   AffChXY(0,x_txt_mem_libre,y_txt_mem_libre,txt_mem_libre[Langue]+':',16*or_fond+or_clair);
   AffChXY2(0,x_txt_mem_DOS,y_txt_mem_DOS,'DOS^ '+EntierTexte2(MemoireDOSLibre shr 6,5)+'k',
            16*or_fond+gris_clair,16*or_fond+gris_fonce);
   AffChXY2(0,x_txt_mem_EMS,y_txt_mem_EMS,'EMS^ '+EntierTexte2(Nb_pages_EMS_Libres*16,5)+'k',
            16*or_fond+gris_clair,16*or_fond+gris_fonce);
   AffChXY2(0,x_txt_mem_GUS,y_txt_mem_GUS,'GUS^ '+EntierTexte2(Taille_GUS,5)+'k',
            16*or_fond+gris_clair,16*or_fond+gris_fonce);

   {programme: nom, LED, titre, nmodule, longueur, avancement (curseur et %)}
   with programme_courant do begin
     AffChXY(0,x_txt_programme,y_txt_programme,txt_prog[Langue]+':',16*or_fond+or_clair);
     AffChXY(0,x_nom_programme,y_nom_programme,XString(Nom,' ',8),16*or_fond+gris_fonce);
     AffCarXY(0,x_LED_programme,y_LED_programme,car_LED,16*or_fond+couleur_LED_vr[actif,modifie]);
     AffChXY(0,x_titre_programme,y_titre_programme,XString(Titre,' ',22),16*or_fond+blanc);
     if actif then begin
       AffChXY(0,x_numero_module,y_numero_module,EntierTexte2(Numero_module,3)+'/'+EntierTexte2(Nombre_modules,3),
               16*or_fond+gris_fonce);
       AfficheBarreCurseur(xg_avanc_prog_curs,y_avanc_prog_curs,larg_curs_avanc_prog,'',Nombre_modules,Numero_module);
       if Nombre_modules<>0 then
        AffChXY(0,x_avanc_prog,y_avanc_prog,EntierTexte2((100*LONGINT(Numero_module)) div Nombre_modules,3)+'%',
                16*or_fond+gris_fonce);
     end;
   end;

   {textes position,pattern,note,tempo,BPM,volume global}
   AffChXY(0,x_txt_position,y_txt_position,txt_position,16*or_fond+or_clair);
   AffChXY(0,x_txt_pattern,y_txt_pattern,txt_pattern,16*or_fond+or_clair);
   AffChXY(0,x_txt_note,y_txt_note,txt_note,16*or_fond+or_clair);
   AffChXY(0,x_txt_tempo,y_txt_tempo,txt_tempo,16*or_fond+or_clair);
   AffChXY(0,x_txt_BPM2,y_txt_BPM2,txt_BPM,16*or_fond+or_clair);
   AffChXY(0,x_txt_gvol,y_txt_gvol,txt_gvol,16*or_fond+or_clair);

   {texte barre curseur volume}
   AffChXY2(0,x_txt_volume,y_txt_volume,txt_volume,16*or_fond+gris_fonce,16*or_fond+or_clair);

   {fentre afficheurs chrono et symboles}
   AffChXY(0,xg_fen_afficheurs,yh_fen_afficheurs-1,RepeteCar('',larg_fen_afficheurs),16*or_fond+noir);
   EffaceZone(0,xg_fen_afficheurs,yh_fen_afficheurs,xd_fen_afficheurs,yb_fen_afficheurs,16*noir);
   LigneHorizontale(0,1,yh_fen_afficheurs+1,xg_fen_afficheurs,xd_fen_afficheurs,16*noir+or_fond);

   {efface le fond de la ligne suprieure}
   EffaceZone(0,xg_fen_infos_mus,yh_fen_infos_mus,xd_fen_infos_mus,yh_fen_infos_mus,16*noir);
 end;

 AttendVBL;

 {fentre infrieure: efface fond}
 if reaffiche>=reaff_fen then
  FenetreOmbre(xg_fen_infos_mus,yh_fen_infos_mus,xd_fen_infos_mus,yb_fen_infos_mus,normale);

 {fentre infrieure: affiche titre}
 if reaffiche>=reaff_titre then FenInf[type_aff].InitAff;

end; {InitAffMus}

{----------------------------------------------------------}

procedure GestionAffMus;

var touche: Key;

    i,nombre_patterns: WORD;
    v: BYTE;

    {onde pour oscillo son (canal de sortie) ou instrument}
    onde: record case BYTE of
           1: (mono:          array[0..largeur_osc_mono  -1] of SHORTINT);
           2: (gauche,droite: array[0..largeur_osc_stereo-1] of SHORTINT);
           3: (smp: array[0..(largeur_osc shl zoom_smp_max)-1] of SHORTINT);
          end;

    {indicateur de fin de musique: TRUE -> le module s'est termin sans appui de touche}
    fin_normale: BOOLEAN;


procedure AffOscillo;

var s: POINTER;
    v,n_curs,coul_note: BYTE;
    long_onde,p: WORD;

    curs_osc: array[0..Max_Voies] of T_CursOsc;
              {0: position du bouclage, 1..Max_Voies: offset courant sur la voie}

begin
 case liste_instr.mode_oscillo of
  oscillo_son: with Variables^,Info^ do begin
                 curs_osc[0].type_:=curs_aucun;

                 if Utilise_Stereo then begin
                   {chantillonne le tampon DMA en stro}
                   EchantTamponStereo(Mem[Seg_Table_Volume:Pointeur_Tampon_Joue],onde.gauche,onde.droite,
                                      Taille_Tampon,largeur_osc_stereo);

                   AfficheOscillo(x_oscillo_g,y_oscillo_g,largeur_osc_car_stereo,onde.gauche,table_car[1],curs_osc);
                   AfficheOscillo(x_oscillo_d,y_oscillo_d,largeur_osc_car_stereo,onde.droite,
                                  table_car[(nombre_car_limite shr 1)+1],curs_osc);

                   AffCarXY(0,x_HP_oscillo_g,y_HP_oscillo_g,car_HP[FALSE],16*or_fond+or_clair);
                   AffCarXY(0,x_HP_oscillo_d,y_HP_oscillo_d,car_HP[TRUE ],16*or_fond+or_clair);
                 end
                 else begin
                   {chantillonne le tampon DMA en mono}
                   EchantTampon(Mem[Seg_Table_Volume:Pointeur_Tampon_Joue],onde.mono,Taille_Tampon,largeur_osc_mono);
                   AfficheOscillo(x_oscillo,y_oscillo,largeur_osc_car_mono,onde.mono,table_car[1],curs_osc);
                 end;
               end;

  oscillo_instr: with liste_instr,onde,Info^,MUS.T^.PtrSamples[courant]^ do begin
                   long_onde:=largeur_osc shl zoom; {longueur totale de l'onde chantillonne}

                   if (((Info and M_Adlib)=0) and
                       ((etat[courant] and smp_charge)=smp_charge)) then begin {l'instrument existe en mmoire}
                     {teste s'il faut r-chantillonner l'onde de l'instrument courant}
                     if echant_onde then begin
                       {dtermine @ instrument courant (!! mapping EMS !!)}
                       s:=AdrSample(courant);

                       if Len>long_onde then EchantTampon(s^,smp,Len,long_onde)
                                        else begin FillChar(smp,long_onde,0); Move(s^,smp,Len); end;

                       {convertit ventuellement en 8 bits signs}
                       if (Info and M_NSigne)=0 then ConvSigne(smp,long_onde);

                       echant_onde:=FALSE;
                     end;

                     {initialise le curseur de bouclage}
                     with curs_osc[0] do
                      if Rep_Len<>0 then begin
                        type_:=curs_pointille;

                        if Len>long_onde then begin
                          {calcule la position du bouclage dans l'onde}
                          p:=((long_onde-1)*LONGINT(Rep)) div (Len-1);
                          if (p>=debut_onde) and
                             (p<=debut_onde+largeur_osc-1) then position:=p-debut_onde
                                                           else type_:=curs_vide; {non visible}
                        end
                        else position:=Rep;
                      end
                      else type_:=curs_vide; {pas de bouclage}

                     n_curs:=1; {n du curseur offset courant dans l'onde}

                     {recherche l'instrument courant actif sur les voies}
                     for v:=1 to MUS.H^.Nb_de_voies_patt do
                      if ((VSample[v]+1)=courant) and ((VControle[v] and M_Pas_Active)=0) then
                       with curs_osc[n_curs] do begin
                         type_:=curs_continu;

                         if Len>long_onde then begin
                           {calcule la position de l'offset courant dans l'onde}
                           p:=((long_onde-1)*LONGINT(VOffset_actuel[v])) div (Len-1);
                           if (p>=debut_onde) and
                              (p<=debut_onde+largeur_osc-1) then position:=p-debut_onde
                                                            else type_:=curs_vide; {non visible}
                         end
                         else position:=VOffset_actuel[v];

                         inc(n_curs);
                       end;

                     curs_osc[n_curs].type_:=curs_dernier;
                   end
                   else begin {l'instrument courant n'existe pas ou n'est pas charg en mmoire}
                     FillChar(smp,long_onde,0);
                     curs_osc[0].type_:=curs_aucun;
                   end;

                   AfficheOscillo(x_oscillo,y_oscillo,largeur_osc_car,smp[debut_onde],table_car[1],curs_osc);

                   {affiche les flches}
                   AdrTexte :=PtrVRAM(0,x_oscillo-1,y_oscillo+(hauteur_osc_car div 2));
                   AttrTexte:=16*or_fond+orange;
                   if debut_onde>0 then AffCar('') else AffCar(' ');
                   inc(AdrTexte,largeur_osc_car);
                   if debut_onde+largeur_osc<(largeur_osc shl zoom) then AffCar('') else AffCar(' ');

                   {affiche le facteur de zoom}
                   AdrTexte :=PtrVRAM(0,x_oscillo-2,y_oscillo+hauteur_osc_car-1);
                   AttrTexte:=16*or_fond+gris;
                   if Len>long_onde then AffCh(CopyR('x'+EntierTexte2(1 shl zoom,1),2)) else AffCh('  ');

                   {affiche n instrument courant, LED et tat}
                   if (etat[courant] and smp_joue)=smp_joue then coul_note:=jaune
                   else
                   if (etat[courant] and smp_actif)=smp_actif then coul_note:=gris
                   else
                   if (etat[courant] and smp_dejajoue)=smp_dejajoue then coul_note:=or_sombre
                   else coul_note:=or_fond;

                   AdrTexte :=PtrVRAM(0,x_oscillo-2,y_oscillo);
                   with claviers do
                    AttrTexte:=16*or_fond+couleur_curs[type_coul(
                                (type_aff=aff_instr) or
                               ((type_aff=aff_clav ) and (num_instr[instr_courant]=courant)))];
                   AffCh(EntierTexte2(courant,2));
                   AdrTexte :=PtrVRAM(0,x_oscillo-2,y_oscillo+1);
                   AttrTexte:=16*or_fond+couleur_LED_vr[(etat[courant] and smp_charge)=smp_charge,
                                                        (etat[courant] and smp_stoppe)=smp_stoppe]; AffCar(car_LED);
                   AttrTexte:=16*or_fond+coul_note; AffCar('');
                 end;
 end;
end; {AffOscillo}


procedure EffaceOscillo;
begin
 case liste_instr.mode_oscillo of
  oscillo_son: if Variables^.Utilise_Stereo then begin
                 EffaceZone(0,x_oscillo_g,y_oscillo_g,
                              x_oscillo_g+largeur_osc_car_stereo-1,y_oscillo_g+hauteur_osc_car-1,16*noir+vert);
                 EffaceZone(0,x_oscillo_d,y_oscillo_d,
                              x_oscillo_d+largeur_osc_car_stereo-1,y_oscillo_d+hauteur_osc_car-1,16*noir+vert);
               end
               else
                 EffaceZone(0,x_oscillo,y_oscillo,
                              x_oscillo+largeur_osc_car_mono-1,y_oscillo+hauteur_osc_car-1,16*noir+vert);

  oscillo_instr: EffaceZone(0,x_oscillo,y_oscillo,
                              x_oscillo+largeur_osc_car-1,y_oscillo+hauteur_osc_car-1,16*noir+vert);
 end;
end; {EffaceOscillo}


begin  {GestionAffMus}
 if mode_interactif then CacheCursSouris;

 {initialisations donnes du module}
 nombre_patterns:=0;
 for i:=1 to Max_Pattern do if MUS.T^.Pattern[i]<>0 then inc(nombre_patterns);

 {initialisations structures d'affichage}
 for i:=aff_aide to aff_spect do FenInf[i].InitStruc;

 if type_aff=aff_aide then type_aff:=anc_type_aff;

 {initialise les volumes}
 Volume_Total:=0; Changer_Volume_Total; {maximum}
 Mute:=FALSE;
 Fading.actif:=FALSE;

 fin_normale:=TRUE;

 {initialisations affichage}
 InitAffMus(reaff_tout);
 reaffiche:=reaff_lignes;

 with Variables^,Info^,MUS do
  repeat
   if type_aff<>aff_aide then anc_type_aff:=type_aff;

   if reaffiche>=reaff_titre then InitAffMus(reaffiche);

   if not(Pause) then begin
     {attente du 50 de seconde}
     Test_Periodique:=0;
     repeat until (Test_Periodique=1) or Sortie_Stoppee;

     with Fading do
      if actif then begin
        {diminue progressivement le volume principal}
        Volume_Total:=MinEntier(Volume_Total+vitesse,64);
        Changer_Volume_Total;
        if Volume_Total=64 then Stopper_Sortie:=TRUE; {force la fin de la musique}
      end;

     {position,pattern,note,tempo,BPM,volume global}
     AffChXY2(0,x_position,y_position,EntierTexte2(Pos_Sequence,3)+'^/^'+
                                      EntierTexte2(H^.Sequence_Len,3),
                                      16*or_fond+orange,16*or_fond+gris_fonce);

     AffChXY2(0,x_pattern,y_pattern,EntierTexte2(Num_Pattern,3)+'^/^'+
                                    EntierTexte2(nombre_patterns,3),
                                    16*or_fond+orange,16*or_fond+gris_fonce);

     AffChXY2(0,x_note,y_note,EntierTexte2(Pos_Note+1,3)+'^/^'+
                              EntierTexte2(T^.Taille_Pattern[Num_Pattern+1],3),
                              16*or_fond+orange,16*or_fond+gris_fonce);

     AffChXY(0,x_tempo,y_tempo,EntierTexte2(Tempo,2+et_zeros),16*or_fond+orange);
     AffChXY(0,x_BPM  ,y_BPM  ,EntierTexte2(BPM,3+et_zeros),16*or_fond+orange);
     AffChXY(0,x_gvol ,y_gvol ,EntierTexte2(Volume_General,2+et_zeros),16*or_fond+orange);

     {avancement module}
     AffChXY(0,x_avanc_module,y_avanc_module,
             EntierTexte2((LONGINT(100)*(Chrono_50+50*(Chrono_sec+60*Chrono_min)))
                                    div (Duree_50 +50*(Duree_sec +60*Duree_min )),3)+'%',16*noir+bleu);
   end;

   AffChrono;    {chrono/compte  rebours}
   AffOscillo;   {oscillo canaux/instrument courant}
   AffFonctions; {fonctions et symboles}

   {barre curseur volume}
   AfficheBarreCurseur(xg_curs_volume,y_curs_volume,larg_curs_volume,car_barre_curs,64,64-Volume_Total);

   MAJ_EtatSmp;   {mise  jour de l'tat des instruments (actif,inactif,...)}
   MAJ_EtatVoies; {mise  jour de l'tat des voies (panning,pitch,...)}

   FenInf[type_aff].Aff;
   reaffiche:=reaff_rien;

   if IntroScan and (Chrono_sec>=duree_IntroScan) then
    with Fading do begin actif:=TRUE; vitesse:=2; end;

   if KeyPressed then begin
     touche:=UpCase(ReadKey);
     case touche of
      {touches gnrales}

      {dplacements dans le module courant: avance/retour}
      CtrlLeft : if not(Pause or IntroScan) then Position_Precedente;
      CtrlRight: if not(Pause or IntroScan) then Position_Suivante;

      {aide}
      F1: begin
           if type_aff=aff_aide then type_aff:=anc_type_aff
                                else type_aff:=aff_aide;
           reaffiche:=reaff_fen;
          end;

      {change fentre infrieure}
      F2..F6: if type_aff<>aff_instr+(touche-F2) then begin
                type_aff:=aff_instr+(touche-F2);
                reaffiche:=reaff_fen;
              end;

      {chrono <-> compte  rebours}
      Key('C'): Chrono:=not(Chrono);

      {DOS shell}
      Key('D'): if not(Pause or Fading.actif) then begin
                  Shell;
                  InitModeTexte(clignote);
                  reaffiche:=reaff_tout;
                end;

      {active/dsactive interpolation}
      Key('I'): Utiliser_Interpolation:=not(Utiliser_Interpolation);

      {active/dsactive Jump}
      Key('J'): Jump:=not(Jump);

      {active/dsactive Mute}
      Key('M'): if not(Fading.actif) then begin
                  Mute:=not(Mute);
                  v:=64-Volume_Total;
                  if Mute then v:=v div 2 else v:=MinEntier(2*v,64);
                  Volume_Total:=64-v;
                  Changer_Volume_Total;
                end;

      {active/dsactive Pause}
      Key('P'): if not(Fading.actif) then Pause_Module;

      {active/dsactive repeat}
      Key('R'): Boucler_MOD:=not(Boucler_MOD);

      {dsactive IntroScan}
      Key('S'): if IntroScan and not(Fading.actif) then IntroScan:=FALSE;

      {baisse le volume principal (augmente Volume_Total)}
      Key('-'): if not(Fading.actif) and (Volume_Total<64) then begin
                  Mute:=FALSE;
                  inc(Volume_Total);
                  Changer_Volume_Total;
                end;

      {augmente le volume principal (diminue Volume_Total)}
      Key('+'): if not(Fading.actif) and (Volume_Total>0) then begin
                  Mute:=FALSE;
                  dec(Volume_Total);
                  Changer_Volume_Total;
                end;

      {active/dsactive voie 1..10}
      Key('0')..Key('9'): begin
                           if touche<>Key('0') then i:=touche-BYTE('0') else i:=10;
                           if Voie_Active(i) then Desactiver_Voie(i) else Activer_Voie(i);
                           reaffiche:=reaff_titre;
                          end;

      {dplacement dans le programme courant: module prcdent}
      CtrlPgUp: if not(Pause or Fading.actif) then
                 with programme_courant do
                  if actif and (Numero_module>1) then begin
                    dec(Numero_module);
                    fin_normale:=FALSE;
                    sens:=-1;

                    with Fading do begin actif:=TRUE; vitesse:=8; end;
                  end;

      {dplacement dans le programme courant: module suivant}
      CtrlPgDn: if not(Pause or Fading.actif) then
                 with programme_courant do
                  if actif and (Numero_module<Nombre_modules) then begin
                    inc(Numero_module);
                    fin_normale:=FALSE;
                    sens:=+1;

                    with Fading do begin actif:=TRUE; vitesse:=8; end;
                  end;

      {dplacement dans le programme courant: module prcdent ou suivant}
      Space: if not(Pause or Fading.actif) then
              with programme_courant do
               if actif and (Numero_module+sens>=1) and
                            (Numero_module+sens<=Nombre_modules) then begin
                 inc(Numero_module,sens);
                 fin_normale:=FALSE;

                 with Fading do begin actif:=TRUE; vitesse:=8; end;
               end;

      {force la fin de la lecture}
      Esc: if not(Pause or Fading.actif) then begin
             lecture_module:=FALSE;
             fin_normale:=FALSE;

             with Fading do begin actif:=TRUE; vitesse:=6; end;
           end;

      {touches particulires  l'cran courant}
      else FenInf[type_aff].TesteTouche(touche);
     end;
   end;
  until Sortie_Stoppee; { attente de la fin de la musique }

 if lecture_module and fin_normale then ModuleSuivant; {<- peut modifier lecture_module}

 if not(lecture_module) then programme_courant.actif:=FALSE;

 {met  jour les paramtres de configuration}
 with liste_voies do type_barres_:=type_barres;
 with spectre do
  begin mode_spectre_:=mode_spectre; amplification_:=amplification; end;

 EffaceOscillo;
 EffacePage(2,16*noir); {le tampon d'affichage de la partition peut tre visible dans l'conomiseur}
 fonte_ModM.Redef(0); {certains caractres utiles peuvent avoir t redfinis par l'oscillo}

 if mode_interactif then AffCursSouris;
end; {GestionAffMus}

{=========================================================}

END.
