{ͻ}
{                       --== TIMERS.PAS ==--                       }
{                                                                      }
{ Unit pour grer plusieurs timers 'virtuels'  frquence relle.     }
{                                                                      }
{ AUTEUR: Fabrice Couteau                                              }
{Ķ}
{ Compatibilit: 286           Dbut d'criture 31/10/95 }
{                              Dernires modifications 26/07/98 }
{ͼ}

UNIT Timers;

{$A+,G+,Q-,R-,S-}

INTERFACE

CONST {tats pour Timer.Status}
      empty=$00;
      used =$01;
      run  =$02;
      busy =$04;

TYPE ProcFAR=procedure;

     Timer=object
            procedure Set_(frequence_: REAL; routine_: ProcFAR);
            procedure Start(reset: BOOLEAN);
            procedure Stop;
            procedure Remove;
            function  Status: BYTE;
            function  Ticks: LONGINT;
           private
            handle: BYTE; {lien avec TimerData[]}
           end;

{================================================}

IMPLEMENTATION

USES Util;

CONST freq_8253=1193180;
      freq_18_2=freq_8253/65536; { 18.206}

      base_compteur=256; {prcision des calculs=1/base_compteur}

      aucun=0;
      nombre_timers_max=10;

VAR ancINT08: POINTER;
    ratio_18_2,compteur_18_2: LONGINT;

    TimerData: array[1..nombre_timers_max] of
                record
                 frequence: REAL;
                 ratio,compteur,appels: LONGINT;
                 Routine: ProcFAR;
                 etat: BYTE;
                end;

    ancExitProc: POINTER;

{================================================}

procedure INT08; interrupt;
var t: BYTE;
begin
 for t:=1 to nombre_timers_max do with TimerData[t] do
  if etat=used+run then begin
   inc(compteur,base_compteur);
   if compteur>=ratio then begin
    dec(compteur,ratio); inc(appels);
    etat:=etat or busy; Routine; etat:=etat and not(busy);
   end;
  end;

 inc(compteur_18_2,base_compteur);
 if compteur_18_2>=ratio_18_2 then begin
  dec(compteur_18_2,ratio_18_2);
  asm pushf; call ancINT08 end;
 end
 else Port[$20]:=$20;
end; {INT08}

{------------------------------------------------}

procedure FreqTimer0(freq: REAL);
{change la frquence du timer 0 du 8253 (freq_18_2  freq  freq_8253)}
var valeur_8253: WORD;
begin
 valeur_8253:=Round(freq_8253/freq);
 asm
  cli
  mov AL,00110110b
  out 43h,AL
  mov AX,valeur_8253
  out 40h,AL
  mov AL,AH
  out 40h,AL
  sti
 end;
end; {FreqTimer0}

{------------------------------------------------}

procedure RestaureTimer_18_2;
begin
 SwapIntVec($08,ancINT08,ancINT08);
 FreqTimer0(freq_18_2);
end; {RestaureTimer_18_2}

{------------------------------------------------}

function TimerMax: BYTE;
{renvoie le numro du timer de frquence la plus leve, 0 si aucun utilis}
var t: BYTE;
    freq_max: REAL;
begin
 TimerMax:=aucun; freq_max:=0;
 for t:=1 to nombre_timers_max do with TimerData[t] do
  if (etat<>empty) and (frequence>freq_max) then
   begin freq_max:=frequence; TimerMax:=t end;
end; {TimerMax}

{------------------------------------------------}

procedure AjusteTimers(freq_max: REAL);
{calcule le ratio des timers par rapport  freq_max}
var t: BYTE;
    anc_ratio: LONGINT;
begin
 for t:=1 to nombre_timers_max do with TimerData[t] do
  if etat<>empty then begin
   anc_ratio:=ratio;
   ratio:=Round(base_compteur*freq_max/frequence);
   if anc_ratio=0 then compteur:=0
                  else compteur:=Round(compteur/anc_ratio*ratio);
  end;
end; {AjusteTimers}

{------------------------------------------------}

procedure Synchro_18_2;
{calcule le ratio du timer  18.2 Hz par rapport au plus rapide}
var anc_ratio_18_2: LONGINT;
begin
 with TimerData[TimerMax] do begin
  anc_ratio_18_2:=ratio_18_2;
  if frequence>freq_18_2 then begin
   ratio_18_2:=Round(base_compteur*frequence/freq_18_2);
   FreqTimer0(frequence);
  end
  else begin
   ratio_18_2:=base_compteur;
   FreqTimer0(freq_18_2);
  end;
  if anc_ratio_18_2=0 then compteur_18_2:=0
                      else compteur_18_2:=Round(compteur_18_2/anc_ratio_18_2*ratio_18_2);
 end;
end; {Synchro_18_2}

{------------------------------------------------}

procedure Timer.Set_(frequence_: REAL; routine_: ProcFAR);
{installe une routine qui sera appele  la frquence indique (0 < freq  freq_8253)}
var t,timer_max: BYTE;
    freq_max: REAL;
begin
 if (frequence_<=0) or (frequence_>freq_8253) then exit;

 handle:=aucun;
 for t:=1 to nombre_timers_max do with TimerData[t] do
  if etat=empty then begin
   handle:=t;
   timer_max:=TimerMax;
   frequence:=frequence_;
   compteur:=0;
   Routine:=routine_;
   etat:=used;

   if timer_max=aucun then begin
    if frequence<freq_18_2 then ratio:=Round(base_compteur*freq_18_2/frequence)
                           else ratio:=base_compteur;
    SwapIntVec($08,ancINT08,@INT08);
   end
   else begin
    freq_max:=TimerData[timer_max].frequence;
    if frequence>=freq_18_2 then begin
     if frequence>freq_max then AjusteTimers(frequence)
                           else ratio:=Round(base_compteur*freq_max/frequence);
    end
    else begin
     if freq_max>freq_18_2 then ratio:=Round(base_compteur*freq_max/frequence)
                           else AjusteTimers(freq_18_2);
    end;
   end;

   Synchro_18_2;
   exit;
  end;
end; {Timer.Set_}

{------------------------------------------------}

procedure Timer.Start(reset: BOOLEAN);
{dmarre le timer; si reset=TRUE, la variable 'appels' est remise  0;
 mettre FALSE pour reprendre aprs une pause}
begin
 if handle in [1..nombre_timers_max] then with TimerData[handle] do
  if etat<>empty then begin
   if reset then appels:=0;
   etat:=etat or run;
  end;
end; {Timer.Start}

{------------------------------------------------}

procedure Timer.Stop;
{arrte momentanment le timer}
begin
 if handle in [1..nombre_timers_max] then with TimerData[handle] do
  etat:=etat and not(run);
end; {Timer.Stop}

{------------------------------------------------}

procedure Timer.Remove;
{retire dfinitivement le timer; resynchronise les autres}
var ajuste_timers: BOOLEAN;
    timer_max: BYTE;
begin
 if handle in [1..nombre_timers_max] then with TimerData[handle] do
  if etat<>empty then begin
   ajuste_timers:=(handle=TimerMax) and (frequence>freq_18_2);
   etat:=empty;
   handle:=aucun;
   timer_max:=TimerMax;
   if timer_max=aucun then RestaureTimer_18_2
   else
    if ajuste_timers then begin
     with TimerData[timer_max] do
      if frequence>freq_18_2 then AjusteTimers(frequence) else AjusteTimers(freq_18_2);
     Synchro_18_2;
    end;
  end;
end; {Timer.Remove}

{------------------------------------------------}

function Timer.Status: BYTE;
{renvoie l'tat du timer}
begin
 Status:=empty;
 if handle in [1..nombre_timers_max] then with TimerData[handle] do
  Status:=etat;
end; {Timer.Status}

{------------------------------------------------}

function Timer.Ticks;
{renvoie le nombre d'appels  la routine depuis le dernier Start(TRUE)}
begin
 Ticks:=0;
 if handle in [1..nombre_timers_max] then with TimerData[handle] do
  if etat<>empty then Ticks:=appels;
end; {Timer.Ticks}

{------------------------------------------------}

procedure InitTimers;
var t: BYTE;
begin
 ancINT08:=LongPtr(MemL[0:4*$08]);
 for t:=1 to nombre_timers_max do with TimerData[t] do
  begin etat:=empty; appels:=0 end;
end; {InitTimers}

{------------------------------------------------}

{$F+}
procedure ExitTimers;
begin
 ExitProc:=ancExitProc;
 RestaureTimer_18_2;
end;
{$F-}

{================================================}

{initialisation de l'Unit}
begin
 InitTimers;
 ancExitProc:=ExitProc;
 ExitProc:=@ExitTimers; {installe routine de sortie}
END.
